(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/demo.js                                                      //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
allowDemo = Meteor.settings && Meteor.settings["public"] && Meteor.settings["public"].allowDemoAccounts;
                                                                       //
var DEMO_EXPIRATION_MS = 60 * 60 * 1000;                               // 20
var DEMO_GRACE_MS = 10 * 60 * 1000; // time between expiration and deletion
                                                                       //
if (Meteor.isServer) {                                                 // 23
  Accounts.validateLoginAttempt(function (attempt) {                   // 24
    // Enforce expiration times for demo accounts.                     //
                                                                       //
    if (attempt.user && attempt.user.expires) {                        // 27
      (function () {                                                   //
        var expireIn = attempt.user.expires.getTime() - Date.now() + DEMO_GRACE_MS;
        if (expireIn < 0) {                                            // 29
          throw new Meteor.Error(403, "This demo account has expired.");
        }                                                              //
                                                                       //
        if (!allowDemo) {                                              // 33
          // An expiration time implies that this is a demo account, but those have been disabled.
          // Perhaps the config was just changed.                      //
          throw new Meteor.Error(403, "Demo accounts have been disabled.");
        }                                                              //
                                                                       //
        // Force connection close when account expires, so that the client reconnects and
        // re-authenticates, which fails.                              //
        var connection = attempt.connection;                           // 41
        var handle = Meteor.setTimeout(function () {                   // 42
          connection.close();                                          // 42
        }, expireIn);                                                  //
                                                                       //
        connection.onClose(function () {                               // 44
          Meteor.clearTimeout(handle);                                 // 44
        });                                                            //
      })();                                                            //
    }                                                                  //
                                                                       //
    return true;                                                       // 47
  });                                                                  //
                                                                       //
  function cleanupExpiredUsers() {                                     // 50
    // Delete expired demo accounts and all their grains.              //
                                                                       //
    var now = new Date(Date.now() - DEMO_GRACE_MS);                    // 53
    Meteor.users.find({ expires: { $lt: now } }, { fields: { _id: 1, loginIdentities: 1, lastActive: 1, appDemoId: 1 } }).forEach(function (user) {
      Grains.find({ userId: user._id }, { fields: { _id: 1, lastUsed: 1, appId: 1 } }).forEach(function (grain) {
        console.log("delete grain: " + grain._id);                     // 59
        globalDb.removeApiTokens({ grainId: grain._id });              // 60
        Grains.remove(grain._id);                                      // 61
        if (grain.lastUsed) {                                          // 62
          DeleteStats.insert({ type: "demoGrain", lastActive: grain.lastUsed, appId: grain.appId });
        }                                                              //
                                                                       //
        globalBackend.deleteGrain(grain._id, user._id);                // 66
      });                                                              //
                                                                       //
      console.log("delete user: " + user._id);                         // 69
      // We intentionally do not do `ApiTokens.remove({accountId: user._id})`, because some such
      // tokens might still play an active role in the sharing graph.  //
      Contacts.remove({ ownerId: user._id });                          // 72
      UserActions.remove({ userId: user._id });                        // 73
      Notifications.remove({ userId: user._id });                      // 74
      Meteor.users.remove(user._id);                                   // 75
      waitPromise(globalBackend.cap().deleteUser(user._id));           // 76
      if (user.loginIdentities && user.lastActive) {                   // 77
        // When deleting a user, we can specify it as a "normal" user  //
        // (type: user) or as a user who started out by using the app  //
        // demo feature (type: appDemoUser).                           //
        var deleteStatsType = "demoUser";                              // 81
        var isAppDemoUser = !!user.appDemoId;                          // 82
        if (isAppDemoUser) {                                           // 83
          deleteStatsType = "appDemoUser";                             // 84
        }                                                              //
                                                                       //
        // Intentionally record deleted users at time of deletion to avoid miscounting users that
        // were demoing just before the day rolled over.               //
        DeleteStats.insert({ type: deleteStatsType, lastActive: new Date(), appId: user.appDemoId });
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  if (allowDemo) {                                                     // 94
    Meteor.methods({                                                   // 95
      createDemoUser: function (displayName, appDemoId) {              // 96
        // This is a login method that creates a new temporary user    //
        // every time it is used.                                      //
        //                                                             //
        // appDemoId is important for stats; see cleanupExpiredUsers().
        check(displayName, String);                                    // 101
        check(appDemoId, Match.OneOf(undefined, null, String));        // 102
                                                                       //
        // Create the new user.                                        //
        var newUser = { expires: new Date(Date.now() + DEMO_EXPIRATION_MS) };
        if (appDemoId) {                                               // 106
          newUser.appDemoId = appDemoId;                               // 107
        }                                                              //
                                                                       //
        var userId = Accounts.insertUserDoc({ profile: { name: displayName } }, newUser);
                                                                       //
        // Log them in on this connection.                             //
        return Accounts._loginMethod(this, "createDemoUser", arguments, "demo", function () {
          return { userId: userId };                                   // 114
        });                                                            //
      },                                                               //
                                                                       //
      testExpireDemo: function () {                                    // 117
        if (!isDemoUser()) throw new Meteor.Error(403, "not a demo user");
                                                                       //
        var newExpires = new Date(Date.now() + 15000);                 // 120
        if (Meteor.user().expires.getTime() < newExpires.getTime()) {  // 121
          throw new Meteor.Error(403, "can't exend demo");             // 122
        }                                                              //
                                                                       //
        Meteor.users.update(this.userId, { $set: { expires: newExpires } });
      }                                                                //
    });                                                                //
                                                                       //
    // If demo mode is enabled, we permit the client to subscribe to   //
    // information about an app by appId. If this were available in    //
    // non-demo mode, then anonymous users could effectively ask the   //
    // server which apps are installed.                                //
    Meteor.publish("appInfo", function (appId) {                       // 133
      // This publishes info about an app, including the latest        //
      // version of it. Once you log in, it also publishes your        //
      // list of UserActions.                                          //
      check(appId, String);                                            // 137
                                                                       //
      var packageCursor = Packages.find({ appId: appId }, { sort: { "manifest.appVersion": -1 } });
                                                                       //
      var pkg = packageCursor.fetch()[0];                              // 143
                                                                       //
      // This allows us to avoid creating duplicate UserActions.       //
      if (this.userId) {                                               // 146
        return [packageCursor, UserActions.find({ userId: this.userId, appId: appId })];
      }                                                                //
                                                                       //
      return packageCursor;                                            // 153
    });                                                                //
                                                                       //
    SandstormDb.periodicCleanup(DEMO_EXPIRATION_MS, cleanupExpiredUsers);
  } else {                                                             //
    // Just run once, in case the config just changed from allowing demos to prohibiting them.
    Meteor.setTimeout(cleanupExpiredUsers, DEMO_EXPIRATION_MS + DEMO_GRACE_MS);
  }                                                                    //
}                                                                      //
                                                                       //
if (Meteor.isClient && allowDemo) {                                    // 163
  Meteor.loginWithDemo = function (options, callback) {                // 164
    Router.go("demo");                                                 // 165
    callback();                                                        // 166
  };                                                                   //
  // Note: We intentionally don't register the demo service with Accounts.registerService(); we
  //   don't want it to appear in the sign-in drop-down.               //
                                                                       //
  window.testExpireDemo = function () {                                // 171
    Meteor.call("testExpireDemo");                                     // 172
  };                                                                   //
                                                                       //
  Template.demo.events({                                               // 175
    "click button.start": function (event) {                           // 176
      var displayName = "Demo User";                                   // 177
                                                                       //
      var userCallbackFunction = function (err) {                      // 179
        if (err) {                                                     // 180
          window.alert(err);                                           // 181
        } else {                                                       //
          Router.go("root");                                           // 183
        }                                                              //
      };                                                               //
                                                                       //
      if (isSignedUpOrDemo()) {                                        // 187
        userCallbackFunction();                                        // 188
      } else {                                                         //
        Accounts.callLoginMethod({                                     // 190
          methodName: "createDemoUser",                                // 191
          methodArguments: ["Demo User", null],                        // 192
          userCallback: userCallbackFunction                           // 193
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.appdemo.events({                                            // 199
    "click button.start": function (event) {                           // 200
      // When clicking on the createDemoUser button on the app demo,   //
      // we want to:                                                   //
      //                                                               //
      // 1. Create the Demo User if they are not logged in.            //
      //                                                               //
      // 2. Log the user in as this Demo User if we created it.        //
      //                                                               //
      // 3. Install the chosen app.                                    //
      //                                                               //
      // 4. Create a new grain with this app.                          //
      //                                                               //
      // 5. Take them into this grain.                                 //
                                                                       //
      // calculate the appId                                           //
                                                                       //
      // We copy the appId into the scope so the autorun function can access it.
      var appId = this.appId;                                          // 217
      var signingIn = false;                                           // 218
                                                                       //
      // TODO(cleanup): Is there a better way to do this? Before multi-identity, we could use the
      //   `userCallback` option of `Accounts.callLoginMethod()`, but now that doesn't work because
      //   it fires too early.                                         //
      //                                                               //
      // Note that we don't use Template.instance().autorun() because the template gets destroyed
      // and recreated during account creation.                        //
      var done = false;                                                // 226
      var handle = Tracker.autorun(function () {                       // 227
        if (done) return;                                              // 228
                                                                       //
        if (!isSignedUpOrDemo()) {                                     // 230
          if (!signingIn) {                                            // 231
            signingIn = true;                                          // 232
            // 1. Create the Demo User & 2. Log the user in as this Demo User.
            Accounts.callLoginMethod({                                 // 234
              methodName: "createDemoUser",                            // 235
              methodArguments: ["Demo User", appId]                    // 236
            });                                                        //
          } else {                                                     //
            return;                                                    // 239
          }                                                            //
        } else {                                                       //
          // `handle` may not have been assigned yet if this is the first run of the callback.
          // But we definitely don't want the callback to run again. So we have to resort to
          // setting a flag that disables subsequent runs, and then deferring the actual stop. Ick.
          done = true;                                                 // 245
          Meteor.defer(function () {                                   // 246
            handle.stop();                                             // 246
          });                                                          //
                                                                       //
          // First, find the package ID, since that is what            //
          // addUserActions takes. Choose the package ID with          //
          // highest version number.                                   //
          var packageId = Packages.findOne({ appId: appId }, { sort: { "manifest.appVersion": -1 } })._id;
                                                                       //
          // 3. Install this app for the user, if needed.              //
          if (UserActions.find({ appId: appId, userId: Meteor.userId() }).count() == 0) {
            globalDb.addUserActions(packageId);                        // 258
          }                                                            //
                                                                       //
          // 4. Create new grain and 5. browse to it.                  //
          launchAndEnterGrainByPackageId(packageId);                   // 262
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
Router.map(function () {                                               // 269
  this.route("demo", {                                                 // 270
    path: "/demo",                                                     // 271
    waitOn: function () {                                              // 272
      return Meteor.subscribe("credentials");                          // 273
    },                                                                 //
                                                                       //
    data: function () {                                                // 276
      return {                                                         // 277
        allowDemo: allowDemo,                                          // 278
        pageTitle: "Demo",                                             // 279
        isDemoUser: isDemoUser()                                       // 280
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("demoRestart", {                                          // 285
    path: "/demo-restart",                                             // 286
    waitOn: function () {                                              // 287
      return Meteor.subscribe("credentials");                          // 288
    },                                                                 //
                                                                       //
    data: function () {                                                // 291
      Meteor.logout();                                                 // 292
      Router.go("demo");                                               // 293
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
Router.map(function () {                                               // 298
  this.route("appdemo", {                                              // 299
    path: "/appdemo/:appId",                                           // 300
    waitOn: function () {                                              // 301
      return Meteor.subscribe("appInfo", this.params.appId);           // 302
    },                                                                 //
                                                                       //
    data: function () {                                                // 305
      // find the newest (highest version, so "first" when sorting by  //
      // inverse order) matching package.                              //
      var thisPackage = Packages.findOne({                             // 308
        appId: this.params.appId                                       // 309
      }, {                                                             //
        sort: { "manifest.appVersion": -1 }                            // 311
      });                                                              //
                                                                       //
      // In the case that the app requested is not present, we show    //
      // this string as the app name.                                  //
      var appName = "missing package";                                 // 316
                                                                       //
      if (thisPackage) {                                               // 318
        appName = SandstormDb.appNameFromPackage(thisPackage);         // 319
      }                                                                //
                                                                       //
      return {                                                         // 322
        allowDemo: allowDemo,                                          // 323
        // For appdemo, we always allow you to start the demo, because
        // the this refers to the app demo, and if a visitor clicks    //
        // visits /appdemo/:appId once and creates a Demo User         //
        // account, and then clicks a different /appdemo/:appId URL to
        // demo a different app, we want them to experience the joy of
        // trying the second app.                                      //
        shouldShowStartDemo: true,                                     // 330
        appName: appName,                                              // 331
        pageTitle: appName + " Demo on Sandstorm",                     // 332
        appId: this.params.appId,                                      // 333
        isDemoUser: isDemoUser()                                       // 334
      };                                                               //
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=demo.js.map
