(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/shell.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements the common shell components such as the top bar.
// It also covers the root page.                                       //
                                                                       //
var getNamesFromIdentityIds = function (identityIds) {                 // 20
  check(identityIds, [String]);                                        // 21
  if (identityIds.length === 0) {                                      // 22
    return [];                                                         // 23
  }                                                                    //
                                                                       //
  var identities = Meteor.users.find({                                 // 26
    _id: { $in: identityIds }                                          // 27
  });                                                                  //
  return identities.map(function (identity) {                          // 29
    return { name: identity.profile.name };                            // 30
  });                                                                  //
};                                                                     //
                                                                       //
browseHome = function () {                                             // 34
  Router.go("root");                                                   // 35
};                                                                     //
                                                                       //
getOrigin = function () {                                              // 38
  return document.location.protocol + "//" + document.location.host;   // 39
};                                                                     //
                                                                       //
if (Meteor.isClient) {                                                 // 42
  // Subscribe to basic grain information first and foremost, since    //
  // without it we might e.g. redirect to the wrong place on login.    //
  globalSubs = [Meteor.subscribe("grainsMenu"), Meteor.subscribe("userPackages"), Meteor.subscribe("devPackages"), Meteor.subscribe("credentials"), Meteor.subscribe("accountIdentities")];
                                                                       //
  Tracker.autorun(function () {                                        // 53
    var me = Meteor.user();                                            // 54
    if (me) {                                                          // 55
      if (me.profile) {                                                // 56
        Meteor.subscribe("identityProfile", me._id);                   // 57
      }                                                                //
                                                                       //
      if (me.loginIdentities) {                                        // 60
        me.loginIdentities.forEach(function (identity) {               // 61
          Meteor.subscribe("identityProfile", identity.id);            // 62
        });                                                            //
      }                                                                //
                                                                       //
      if (me.nonloginIdentities) {                                     // 66
        me.nonloginIdentities.forEach(function (identity) {            // 67
          Meteor.subscribe("identityProfile", identity.id);            // 68
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  // export: called by sandstorm-accounts-ui/login_buttons.js          //
  logoutSandstorm = function () {                                      // 75
    Meteor.logout(function () {                                        // 76
      sessionStorage.removeItem("linkingIdentityLoginToken");          // 77
      Accounts._loginButtonsSession.closeDropdown();                   // 78
      globalTopbar.closePopup();                                       // 79
      var openGrains = globalGrains.get();                             // 80
      openGrains.forEach(function (grain) {                            // 81
        grain.destroy();                                               // 82
      });                                                              //
                                                                       //
      globalGrains.set([]);                                            // 85
      Router.go("root");                                               // 86
    });                                                                //
  };                                                                   //
}                                                                      //
                                                                       //
if (Meteor.isServer) {                                                 // 91
  // Set up browser policy.                                            //
  //                                                                   //
  // Note that by default (when the browser-policy package is added), Content-Security-Policy will
  // already be set to same-origin-only for everything except XHR and WebSocket. Eval is disabled,
  // but inline script are enabled.                                    //
  //                                                                   //
  // TODO(security): Consider disallowing inline scripts. Currently this forces Meteor to do an
  //   extra round trip on startup to fetch server settings. That seems like something that could
  //   be fixed in Meteor (e.g. embed the settings as JSON text rather than script). Startup time
  //   is incredibly important, and Meteor's templating system (which we use to render all our
  //   HTML) already does a good job of protecting us from script injection, so right now I think
  //   we should favor avoiding the round trip.                        //
  //                                                                   //
  // TODO(someday): Detect when an app tries to navigate off-site using CSP's violation reporting
  //   feature. Ask the user whether they want to open the link in a new tab. This is an annoying
  //   prompt, but if we just open the tab directly then apps can trivially leak by opening tabs
  //   with encoded URLs to an evil server. Although, this attack would be very detectable to the
  //   user, so maybe it's not a big deal...                           //
                                                                       //
  BrowserPolicy.framing.disallow(); // Disallow framing of the UI.     // 111
  BrowserPolicy.content.allowFrameOrigin(getWildcardOrigin());         // 112
                                                                       //
  // Allow anything to be loaded from the static asset host.           //
  var Url = Npm.require("url");                                        // 115
  var staticAssetHost = Url.parse(process.env.ROOT_URL).protocol + "//" + globalDb.makeWildcardHost("static");
  BrowserPolicy.content.allowImageOrigin(staticAssetHost);             // 118
  BrowserPolicy.content.allowScriptOrigin(staticAssetHost);            // 119
  BrowserPolicy.content.allowFontOrigin(staticAssetHost);              // 120
  BrowserPolicy.content.allowConnectOrigin(staticAssetHost);           // 121
                                                                       //
  Meteor.publish("grainsMenu", function () {                           // 123
    var _this = this;                                                  //
                                                                       //
    if (this.userId) {                                                 // 124
      if (Meteor.settings["public"].quotaEnabled) {                    // 125
        (function () {                                                 //
          // Hack: Fire off an asynchronous update to the user's storage usage whenever they open the
          //   front page.                                             //
          // TODO(someday): Implement the ability to reactively subscribe to storage usage from the
          //   back-end?                                               //
          var userId = _this.userId;                                   // 130
          globalBackend.cap().getUserStorageUsage(userId).then(function (results) {
            inMeteor(function () {                                     // 132
              Meteor.users.update(userId, { $set: { storageUsage: parseInt(results.size) } });
            });                                                        //
          })["catch"](function (err) {                                 //
            if (err.kjType !== "unimplemented") {                      // 136
              console.error(err.stack);                                // 137
            }                                                          //
          });                                                          //
        })();                                                          //
      }                                                                //
                                                                       //
      var identityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
      return [UserActions.find({ userId: this.userId }), Grains.find({ userId: this.userId }), ApiTokens.find({ "owner.user.identityId": { $in: identityIds } })];
    } else {                                                           //
      return [];                                                       // 149
    }                                                                  //
  });                                                                  //
                                                                       //
  Meteor.publish("sessions", function (sessionId) {                    // 153
    // sessionId itself should be secret enough, but they are also not meant to be shared, so as
    // a backup we only publish the session to its owner. Note that `userId` can be null if the
    // user is not logged in or is using incognito mode.               //
    check(sessionId, String);                                          // 157
    return Sessions.find({ _id: sessionId, $or: [{ userId: this.userId }, { userId: null }] });
  });                                                                  //
                                                                       //
  Meteor.publish("devPackages", function () {                          // 161
    return DevPackages.find();                                         // 162
  });                                                                  //
                                                                       //
  Meteor.publish("notifications", function () {                        // 165
    return Notifications.find({ userId: this.userId }, { fields: { timestamp: 1, text: 1, grainId: 1, userId: 1, isUnread: 1, appUpdates: 1,
        admin: 1, referral: 1, mailingListBonus: 1 } });               // 168
  });                                                                  //
                                                                       //
  Meteor.publish("notificationGrains", function (notificationIds) {    // 171
    // Since publishes can't be reactive, we leave it to the client to subscribe to both
    // "notifications" and "notificationGrains" reactively.            //
    check(notificationIds, [String]);                                  // 174
    var notifications = Notifications.find({                           // 175
      _id: { $in: notificationIds },                                   // 176
      userId: this.userId                                              // 177
    }, {                                                               //
      fields: { grainId: 1 }                                           // 179
    });                                                                //
                                                                       //
    var grainIds = notifications.map(function (row) {                  // 182
      return row.grainId;                                              // 183
    });                                                                //
                                                                       //
    return Grains.find({ _id: { $in: grainIds } }, { fields: { title: 1 } });
  });                                                                  //
                                                                       //
  Meteor.publish("hasUsers", function () {                             // 189
    var _this2 = this;                                                 //
                                                                       //
    // Publish pseudo-collection which tells the client if there are any users at all.
    //                                                                 //
    // TODO(cleanup):  This seems overcomplicated.  Does Meteor have a better way?
    var cursor = Meteor.users.find();                                  // 193
    if (cursor.count() > 0) {                                          // 194
      this.added("hasUsers", "hasUsers", { hasUsers: true });          // 195
    } else {                                                           //
      (function () {                                                   //
        var handle = cursor.observeChanges({                           // 197
          added: function (id) {                                       // 198
            _this2.added("hasUsers", "hasUsers", { hasUsers: true });  // 199
            handle.stop();                                             // 200
            handle = null;                                             // 201
          }                                                            //
        });                                                            //
        _this2.onStop(function () {                                    // 204
          if (handle) handle.stop();                                   // 205
        });                                                            //
      })();                                                            //
    }                                                                  //
                                                                       //
    this.ready();                                                      // 209
  });                                                                  //
                                                                       //
  Meteor.publish("referralInfoPseudo", function () {                   // 212
    var _this3 = this;                                                 //
                                                                       //
    // This publishes a pseudo-collection called referralInfo whose documents have the following
    // form:                                                           //
    //                                                                 //
    // - id: (String) same as the User._id of an identity this user has referred
    // - name: (String) the profile.name from that identity            //
    // - completed: (Boolean) if this referral is complete             //
                                                                       //
    //  If the user is not logged in, then we have no referralInfo.    //
    if (!this.userId) {                                                // 221
      return [];                                                       // 222
    }                                                                  //
                                                                       //
    // Implementation note:                                            //
    //                                                                 //
    // This pseudo-collection is populated very differently for (1) the completed: false case versus
    // the (2) completed: true case.                                   //
                                                                       //
    // Case 1. Publish information about not-yet-complete referrals.   //
    var notCompletedReferralIdentitiesCursor = Meteor.users.find({     // 231
      referredBy: this.userId,                                         // 232
      "profile.name": { $exists: true }                                // 233
    }, {                                                               //
      fields: {                                                        // 235
        _id: 1,                                                        // 236
        referredBy: 1,                                                 // 237
        "profile.name": 1                                              // 238
      }                                                                //
    });                                                                //
    var notCompletedReferralIdentitiesHandle = notCompletedReferralIdentitiesCursor.observeChanges({
      // The added function gets called with the id of Bob when Alice refers Bob.
      added: function (id, fields) {                                   // 243
        _this3.added("referralInfo", id, { name: fields.profile.name, completed: false });
      },                                                               //
      // The removed function gets called when Bob is no longer an uncompleted referral.  Note that
      // this will get more complicated once we support sending completed referrals to the client.
      removed: function (id) {                                         // 248
        _this3.removed("referralInfo", id);                            // 249
      },                                                               //
      // The modified function gets called when Bob's profile.name changed.
      modified: function (id, fields) {                                // 252
        _this3.modified("referralInfo", id, { name: fields.profile.name, completed: false });
      }                                                                //
    });                                                                //
                                                                       //
    // Case 2. Handle completed referrals.                             //
    //                                                                 //
    // - Do a query for the current list of completed identities.      //
    //                                                                 //
    // - Every time we see a new such identity, we create a query that watches that one identity in
    //   case its profile.name changes.                                //
    //                                                                 //
    // - Also watch the first query, since the list of completed identities might change.
    var handleForProfileNameByIdentityId = {};                         // 265
    var stopWatchingAllIdentities = function () {                      // 266
      Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {
        stopWatchingIdentity(identityId);                              // 268
      });                                                              //
    };                                                                 //
                                                                       //
    var stopWatchingIdentity = function (identityId) {                 // 272
      var handleForProfileName = handleForProfileNameByIdentityId[identityId];
      if (handleForProfileName) {                                      // 274
        _this3.removed("referralInfo", identityId);                    // 275
        handleForProfileName.stop();                                   // 276
        // delete is safe because we iterate across `Object.keys()` which returns a copy.
        delete handleForProfileNameByIdentityId[identityId];           // 278
      }                                                                //
    };                                                                 //
                                                                       //
    var watchIdentityAndPublishReferralSuccess = function (identityId) {
      var handleForProfileName = handleForProfileNameByIdentityId[identityId];
      if (handleForProfileName) {                                      // 284
        return;                                                        // 285
      }                                                                //
                                                                       //
      handleForProfileName = Meteor.users.find({                       // 288
        _id: identityId                                                // 289
      }, {                                                             //
        fields: {                                                      // 291
          "profile.name": 1                                            // 292
        }                                                              //
      }).observeChanges({                                              //
        added: function (id, fields) {                                 // 295
          _this3.added("referralInfo", id, { name: fields.profile.name, completed: true });
        },                                                             //
                                                                       //
        changed: function (id, fields) {                               // 299
          _this3.changed("referralInfo", id, { name: fields.profile.name, completed: true });
        },                                                             //
                                                                       //
        removed: function (id) {                                       // 303
          stopWatchingIdentity(id);                                    // 304
        }                                                              //
      });                                                              //
                                                                       //
      handleForProfileNameByIdentityId[identityId] = handleForProfileName;
    };                                                                 //
                                                                       //
    var completedIdentityIdsHandle = Meteor.users.find({               // 311
      _id: this.userId,                                                // 312
      referredIdentityIds: { $exists: true }                           // 313
    }, {                                                               //
      fields: {                                                        // 315
        referredIdentityIds: true                                      // 316
      }                                                                //
    }).observeChanges({                                                //
      // `added` gets called when a user gets their first completed referral.
      added: function (id, fields) {                                   // 320
        for (var i = 0; i < fields.referredIdentityIds.length; i++) {  // 321
          // Unconditionally mark these as successful referrals and start watching.
          watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);
        }                                                              //
      },                                                               //
      // `changed` gets called when a user adds/removes referredIdentityIds, usually when a
      // referral becomes complete.                                    //
      changed: function (id, fields) {                                 // 329
        // Two major tasks.                                            //
        //                                                             //
        // 1. Look for identityIds to unsubscribe from & send removed notices to the client.
        //                                                             //
        // 2. Look for identityIds to subscribe to.                    //
                                                                       //
        // Task 1. Unsubscribe where needed.                           //
        var referredIdentityIdsAsObject = {};                          // 337
        fields.referredIdentityIds.forEach(function (i) {              // 338
          referredIdentityIdsAsObject[i] = true;                       // 338
        });                                                            //
                                                                       //
        Object.keys(handleForProfileNameByIdentityId).forEach(function (identityId) {
          // If the handle doesn't show up in the new list of referredIdentityIds, then remove
          // info from the client & stop it on the server & make it null.
          var handleForProfileName = handleForProfileNameByIdentityId[identityId];
          if (referredIdentityIdsAsObject.hasOwnProperty(identityId)) {
            stopWatchingIdentity(identityId);                          // 345
          }                                                            //
        });                                                            //
                                                                       //
        // Task 2. Subscribe where needed.                             //
        for (var i = 0; i < fields.referredIdentityIds.length; i++) {  // 350
          // The watch... function will avoid double-creating subscriptions, so this is safe.
          watchIdentityAndPublishReferralSuccess(fields.referredIdentityIds[i]);
        }                                                              //
      },                                                               //
      // `removed` gets called when a User suddenly has no referredIdentityIds.
      removed: function () {                                           // 356
        // Remove all data from client; stop all handles.              //
        stopWatchingAllIdentities();                                   // 358
      }                                                                //
    });                                                                //
                                                                       //
    // With cases 1 and 2 handled, register a cleanup function, then declare victory.
    this.onStop(function () {                                          // 363
      stopWatchingAllIdentities();                                     // 364
      notCompletedReferralIdentitiesHandle.stop();                     // 365
      completedIdentityIdsHandle.stop();                               // 366
    });                                                                //
                                                                       //
    this.ready();                                                      // 369
  });                                                                  //
                                                                       //
  Meteor.publish("backers", function () {                              // 372
    var _this4 = this;                                                 //
                                                                       //
    var backers = Assets.getText("backers.txt");                       // 373
    var anonCount = 0;                                                 // 374
    var counter = 0;                                                   // 375
                                                                       //
    backers.split("\n").forEach(function (name) {                      // 377
      name = name.trim();                                              // 378
      if (name === "") {                                               // 379
        ++anonCount;                                                   // 380
      } else {                                                         //
        _this4.added("backers", counter++, { name: name });            // 382
      }                                                                //
    });                                                                //
                                                                       //
    // Text file ends in \n but that shouldn't count.                  //
    --anonCount;                                                       // 387
                                                                       //
    this.added("backers", "anonymous", { count: anonCount - 1 });      // 389
                                                                       //
    this.ready();                                                      // 391
  });                                                                  //
}                                                                      //
                                                                       //
var makeAccountSettingsUi = function () {                              // 395
  return new SandstormAccountSettingsUi(globalTopbar, globalDb, window.location.protocol + "//" + makeWildcardHost("static"));
};                                                                     //
                                                                       //
var MONTHS = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
                                                                       //
var formatInCountdown = function (template, countdownDatetime) {       // 402
  var diff = countdownDatetime.getTime() - Date.now();                 // 403
                                                                       //
  var units = {                                                        // 405
    day: 86400000,                                                     // 406
    hour: 3600000,                                                     // 407
    minute: 60000,                                                     // 408
    second: 1000                                                       // 409
  };                                                                   //
                                                                       //
  for (var unit in babelHelpers.sanitizeForInObject(units)) {          // 412
    // If it's more than one full unit away, then we'll print in terms of this unit. This does
    // mean that we write e.g. "1 minute" for the whole range between 2 minutes and 1 minute, but
    // whatever, this is typical of these sorts of displays.           //
    if (diff >= units[unit]) {                                         // 416
      var count = Math.floor(diff / units[unit]);                      // 417
      // Update next 1ms after the point where `count` would change.   //
      setTopBarTimeout(template, diff - count * units[unit] + 1);      // 419
      return {                                                         // 420
        text: "in " + count + " " + unit + (count > 1 ? "s" : ""),     // 421
        className: "countdown-" + unit                                 // 422
      };                                                               //
    }                                                                  //
  }                                                                    //
                                                                       //
  // We're within a second of the countdown, or past it.               //
  if (diff < -3600000) {                                               // 428
    // Notification appears stale.                                     //
    return null;                                                       // 430
  } else {                                                             //
    setTopBarTimeout(template, diff + 3600001);                        // 432
    return { text: "any moment", className: "countdown-now" };         // 433
  }                                                                    //
};                                                                     //
                                                                       //
var formatAccountExpires = function () {                               // 437
  var expires = Meteor.user().expires;                                 // 438
  return expires && expires.toLocaleTimeString() || null;              // 439
};                                                                     //
                                                                       //
var formatAccountExpiresIn = function (template, currentDatetime) {    // 442
  // TODO(someday): formatInCountdown will set the interval to match account expiration time, and
  // completely overwrite the previous interval for $IN_COUNTDOWN      //
  var user = Meteor.user() || {};                                      // 445
  var expires = user.expires || null;                                  // 446
  if (!expires) {                                                      // 447
    return null;                                                       // 448
  } else {                                                             //
    return formatInCountdown(template, expires, currentDatetime);      // 450
  }                                                                    //
};                                                                     //
                                                                       //
var setTopBarTimeout = function (template, delay) {                    // 454
  Meteor.clearTimeout(template.timeout);                               // 455
  template.timeout = Meteor.setTimeout(function () {                   // 456
    template.timer.changed();                                          // 457
  }, delay);                                                           //
                                                                       //
  // Make sure we re-run when the timeout triggers.                    //
  template.timer.depend();                                             // 461
};                                                                     //
                                                                       //
var determineAppName = function (grainId) {                            // 464
  // Returns:                                                          //
  //                                                                   //
  // - The current app title, if we can determine it, or               //
  //                                                                   //
  // - The empty string "", if we can't determine the current app title.
  var params = "";                                                     // 470
                                                                       //
  // Try our hardest to find the package's name, falling back on the default if needed.
  if (grainId) {                                                       // 473
    var grain = Grains.findOne({ _id: grainId });                      // 474
    if (grain && grain.packageId) {                                    // 475
      var thisPackage = Packages.findOne({ _id: grain.packageId });    // 476
      if (thisPackage) {                                               // 477
        params = SandstormDb.appNameFromPackage(thisPackage);          // 478
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return params;                                                       // 483
};                                                                     //
                                                                       //
var billingPromptState = new ReactiveVar(null);                        // 486
                                                                       //
var showBillingPrompt = function (reason, next) {                      // 488
  billingPromptState.set({                                             // 489
    reason: reason,                                                    // 490
    db: globalDb,                                                      // 491
    topbar: globalTopbar,                                              // 492
    accountsUi: globalAccountsUi,                                      // 493
    onComplete: function () {                                          // 494
      billingPromptState.set(null);                                    // 495
      if (next) next();                                                // 496
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var ifQuotaAvailable = function (next) {                               // 501
  var reason = isUserOverQuota(Meteor.user());                         // 502
  if (reason) {                                                        // 503
    if (window.BlackrockPayments) {                                    // 504
      showBillingPrompt(reason, function () {                          // 505
        // If the user successfully raised their quota, continue the operation.
        if (!isUserOverQuota(Meteor.user())) {                         // 507
          next();                                                      // 508
        }                                                              //
      });                                                              //
    } else {                                                           //
      alert("You are out of storage space. Please delete some things and try again.");
    }                                                                  //
  } else {                                                             //
    next();                                                            // 515
  }                                                                    //
};                                                                     //
                                                                       //
var ifPlanAllowsCustomApps = function (next) {                         // 519
  if (globalDb.isDemoUser() || globalDb.isUninvitedFreeUser()) {       // 520
    if (window.BlackrockPayments) {                                    // 521
      showBillingPrompt("customApp", function () {                     // 522
        // If the user successfully chose a plan, continue the operation.
        if (!globalDb.isDemoUser() && !globalDb.isUninvitedFreeUser()) {
          next();                                                      // 525
        }                                                              //
      });                                                              //
    } else {                                                           //
      alert("Sorry, demo users cannot upload custom apps.");           // 529
    }                                                                  //
  } else {                                                             //
    next();                                                            // 532
  }                                                                    //
};                                                                     //
                                                                       //
var isDemoExpired = function () {                                      // 536
  var user = Meteor.user();                                            // 537
  if (!user) return false;                                             // 538
  var expires = user.expires;                                          // 539
  if (!expires) return false;                                          // 540
  expires = expires.getTime() - Date.now();                            // 541
  if (expires <= 0) return true;                                       // 542
  var comp = Tracker.currentComputation;                               // 543
  if (expires && comp) {                                               // 544
    Meteor.setTimeout(comp.invalidate.bind(comp), expires);            // 545
  }                                                                    //
                                                                       //
  return false;                                                        // 548
};                                                                     //
                                                                       //
// export: this is also used by grain.js                               //
makeDateString = function (date) {                                     // 552
  if (!date) {                                                         // 553
    return "";                                                         // 554
  }                                                                    //
                                                                       //
  var result = undefined;                                              // 557
                                                                       //
  var now = new Date();                                                // 559
  var diff = now.valueOf() - date.valueOf();                           // 560
                                                                       //
  if (diff < 86400000 && now.getDate() === date.getDate()) {           // 562
    result = date.toLocaleTimeString();                                // 563
  } else {                                                             //
    result = MONTHS[date.getMonth()] + " " + date.getDate() + " ";     // 565
                                                                       //
    if (now.getFullYear() !== date.getFullYear()) {                    // 567
      result = date.getFullYear() + " " + result;                      // 568
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 572
};                                                                     //
                                                                       //
// export: used in sandstorm-ui-grainlist                              //
prettySize = function (size) {                                         // 576
  var suffix = "B";                                                    // 577
  if (size >= 1000000000) {                                            // 578
    size = size / 1000000000;                                          // 579
    suffix = "GB";                                                     // 580
  } else if (size >= 1000000) {                                        //
    size = size / 1000000;                                             // 582
    suffix = "MB";                                                     // 583
  } else if (size >= 1000) {                                           //
    size = size / 1000;                                                // 585
    suffix = "kB";                                                     // 586
  }                                                                    //
                                                                       //
  return size.toPrecision(3) + suffix;                                 // 589
};                                                                     //
                                                                       //
// export: used in shared/demo.js                                      //
launchAndEnterGrainByPackageId = function (packageId) {                // 593
  var action = UserActions.findOne({ packageId: packageId });          // 594
  if (!action) {                                                       // 595
    alert("Somehow, you seem to have attempted to launch a package you have not installed.");
    return;                                                            // 597
  } else {                                                             //
    launchAndEnterGrainByActionId(action._id, null, null);             // 599
  }                                                                    //
};                                                                     //
                                                                       //
// export: used in sandstorm-ui-app-details                            //
launchAndEnterGrainByActionId = function (actionId, devPackageId, devIndex) {
  // Note that this takes a devPackageId and a devIndex as well. If provided,
  // they override the actionId.                                       //
  var packageId = undefined;                                           // 607
  var command = undefined;                                             // 608
  var appTitle = undefined;                                            // 609
  var nounPhrase = undefined;                                          // 610
  if (devPackageId) {                                                  // 611
    var devPackage = DevPackages.findOne(devPackageId);                // 612
    if (!devPackage) {                                                 // 613
      console.error("no such dev package: ", devPackageId);            // 614
      return;                                                          // 615
    }                                                                  //
                                                                       //
    var devAction = devPackage.manifest.actions[devIndex];             // 618
    packageId = devPackageId;                                          // 619
    command = devAction.command;                                       // 620
    appTitle = SandstormDb.appNameFromPackage(devPackage);             // 621
    nounPhrase = SandstormDb.nounPhraseForActionAndAppTitle(devAction, appTitle);
  } else {                                                             //
    var action = UserActions.findOne(actionId);                        // 624
    if (!action) {                                                     // 625
      console.error("no such action:", actionId);                      // 626
      return;                                                          // 627
    }                                                                  //
                                                                       //
    packageId = action.packageId;                                      // 630
    var pkg = Packages.findOne(packageId);                             // 631
    command = action.command;                                          // 632
    appTitle = SandstormDb.appNameFromPackage(pkg);                    // 633
    nounPhrase = SandstormDb.nounPhraseForActionAndAppTitle(action, appTitle);
  }                                                                    //
                                                                       //
  var title = "Untitled " + appTitle + " " + nounPhrase;               // 637
                                                                       //
  var identityId = Accounts.getCurrentIdentityId();                    // 639
                                                                       //
  // We need to ask the server to start a new grain, then browse to it.
  Meteor.call("newGrain", packageId, command, title, identityId, function (error, grainId) {
    if (error) {                                                       // 643
      console.error(error);                                            // 644
      alert(error.message);                                            // 645
    } else {                                                           //
      Router.go("grain", { grainId: grainId });                        // 647
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// export global - used in grain.js                                    //
globalQuotaEnforcer = {                                                // 653
  ifQuotaAvailable: ifQuotaAvailable,                                  // 654
  ifPlanAllowsCustomApps: ifPlanAllowsCustomApps                       // 655
};                                                                     //
                                                                       //
if (Meteor.isClient) {                                                 // 658
  HasUsers = new Mongo.Collection("hasUsers"); // dummy collection defined above
  Backers = new Mongo.Collection("backers"); // pseudo-collection defined above
  ReferralInfo = new Meteor.Collection("referralInfo"); // pseudo-collection
                                                                       //
  if (Meteor.settings["public"].quotaEnabled) {                        // 663
    window.testDisableQuotaClientSide = function () {                  // 664
      Meteor.settings["public"].quotaEnabled = false;                  // 665
    };                                                                 //
  }                                                                    //
                                                                       //
  Router.onRun(function () {                                           // 669
    // Close menus and popups any time we navigate.                    //
    globalTopbar.reset();                                              // 671
    this.next();                                                       // 672
  });                                                                  //
                                                                       //
  Template.layout.events({                                             // 675
    "click a": function (event) {                                      // 676
      // Close menus if a navigation link is clicked. Usually the Router.onRun(), above, will also
      // execute, but it will not in the case where the link points to the current page, yet we'd
      // really still like for the menus to close in such cases.       //
      if (!event.isDefaultPrevented()) {                               // 680
        globalTopbar.reset();                                          // 681
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.about.helpers({                                             // 686
    setDocumentTitle: function () {                                    // 687
      document.title = "About · " + globalDb.getServerTitle();         // 688
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.referrals.helpers({                                         // 692
    setDocumentTitle: function () {                                    // 693
      document.title = "Referral Program · " + globalDb.getServerTitle();
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.body.onRendered(function () {                               // 698
    // If we're on iOS, set a class name on <body> so we can use CSS styles to work around mobile
    // Safari's ridiculous iframe rendering behavior.                  //
    //                                                                 //
    // Note that this can't be done as a template helper because the <body> tag cannot have
    // attributes determined by helpers. This appears to be a Meteor bug related to the fact that
    // the <body> tag is not inside a <template>, but rather is itself equivalent to a <template>.
    if (/iPad|iPhone|iPod/.test(navigator.userAgent)) {                // 705
      document.body.className = "ios";                                 // 706
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.layout.onCreated(function () {                              // 710
    this.timer = new Tracker.Dependency();                             // 711
    var resizeTracker = this.resizeTracker = new Tracker.Dependency();
    this.resizeFunc = function () {                                    // 713
      resizeTracker.changed();                                         // 714
    };                                                                 //
                                                                       //
    window.addEventListener("resize", this.resizeFunc, false);         // 717
  });                                                                  //
                                                                       //
  Template.layout.onDestroyed(function () {                            // 720
    Meteor.clearTimeout(this.timeout);                                 // 721
    window.removeEventListener("resize", this.resizeFunc, false);      // 722
  });                                                                  //
                                                                       //
  Template.referrals.helpers({                                         // 725
    isPaid: function () {                                              // 726
      return Meteor.user() && Meteor.user().plan && Meteor.user().plan !== "free";
    },                                                                 //
                                                                       //
    notYetCompleteReferralNames: function () {                         // 730
      return ReferralInfo.find({ completed: false });                  // 731
    },                                                                 //
                                                                       //
    completeReferralNames: function () {                               // 734
      return ReferralInfo.find({ completed: true });                   // 735
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.layout.helpers({                                            // 739
    adminAlertIsTooLarge: function () {                                // 740
      Template.instance().resizeTracker.depend();                      // 741
      var setting = Settings.findOne({ _id: "adminAlert" });           // 742
      if (!setting || !setting.value) {                                // 743
        return false;                                                  // 744
      }                                                                //
      // TODO(someday): 10 and 850 are just magic values that estimate the size of the font and the
      // number of pixels everything else in the topbar respectively. This should really be
      // calculated based on actual sizes of things in the topbar.     //
      return window.innerWidth - setting.value.length * 10 < 850;      // 749
    },                                                                 //
                                                                       //
    adminAlert: function () {                                          // 752
      var setting = Settings.findOne({ _id: "adminAlert" });           // 753
      if (!setting || !setting.value) {                                // 754
        return null;                                                   // 755
      }                                                                //
                                                                       //
      var text = setting.value;                                        // 758
                                                                       //
      var alertTimeSetting = Settings.findOne({ _id: "adminAlertTime" });
      var alertTime = alertTimeSetting && alertTimeSetting.value;      // 761
                                                                       //
      var alertUrlSetting = Settings.findOne({ _id: "adminAlertUrl" });
      var alertUrl = alertUrlSetting ? alertUrlSetting.value.trim() : null;
      if (!alertUrl) alertUrl = null;                                  // 765
                                                                       //
      var template = Template.instance();                              // 767
      var param = undefined;                                           // 768
      var className = undefined;                                       // 769
      if (text.indexOf("$TIME") !== -1) {                              // 770
        if (!alertTime) return null;                                   // 771
        text = text.replace("$TIME", alertTime.toLocaleTimeString());  // 772
      }                                                                //
                                                                       //
      if (text.indexOf("$DATE") !== -1) {                              // 775
        if (!alertTime) return null;                                   // 776
        text = text.replace("$DATE", alertTime.toLocaleDateString());  // 777
      }                                                                //
                                                                       //
      if (text.indexOf("$IN_COUNTDOWN") !== -1) {                      // 780
        if (!alertTime) return null;                                   // 781
        param = formatInCountdown(template, alertTime);                // 782
        if (!param) return null;                                       // 783
        text = text.replace("$IN_COUNTDOWN", param.text);              // 784
        className = param.className;                                   // 785
      }                                                                //
                                                                       //
      if (text.indexOf("$ACCOUNT_EXPIRES_IN") !== -1) {                // 788
        param = formatAccountExpiresIn(template);                      // 789
        if (!param) return null;                                       // 790
        text = text.replace("$ACCOUNT_EXPIRES_IN", param.text);        // 791
        className = param.className;                                   // 792
      }                                                                //
                                                                       //
      if (text.indexOf("$ACCOUNT_EXPIRES") !== -1) {                   // 795
        param = formatAccountExpires();                                // 796
        if (!param) return null;                                       // 797
        text = text.replace("$ACCOUNT_EXPIRES", param);                // 798
      }                                                                //
                                                                       //
      if (text.indexOf("$APPNAME") !== -1) {                           // 801
        text = text.replace("$APPNAME", determineAppName(this.grainId));
      }                                                                //
                                                                       //
      if (alertUrl && alertUrl.indexOf("$APPNAME") !== -1) {           // 805
        alertUrl = alertUrl.replace("$APPNAME", determineAppName(this.grainId));
      }                                                                //
                                                                       //
      return {                                                         // 809
        text: text,                                                    // 810
        className: className,                                          // 811
        alertUrl: alertUrl                                             // 812
      };                                                               //
    },                                                                 //
                                                                       //
    billingPromptState: function () {                                  // 816
      return billingPromptState.get();                                 // 817
    },                                                                 //
                                                                       //
    demoExpired: isDemoExpired,                                        // 820
    canUpgradeDemo: function () {                                      // 821
      return Meteor.settings["public"].allowUninvited;                 // 822
    },                                                                 //
                                                                       //
    globalAccountsUi: function () {                                    // 825
      return globalAccountsUi;                                         // 826
    },                                                                 //
                                                                       //
    identityUser: function () {                                        // 829
      var user = Meteor.user();                                        // 830
      return user && user.profile;                                     // 831
    },                                                                 //
                                                                       //
    showAccountButtons: function () {                                  // 834
      return Meteor.user() && !Meteor.loggingIn() && !isDemoUser();    // 835
    },                                                                 //
                                                                       //
    accountButtonsData: function () {                                  // 838
      return { isAdmin: globalDb.isAdmin() };                          // 839
    },                                                                 //
                                                                       //
    firstLogin: function () {                                          // 842
      return credentialsSubscription.ready() && !isDemoUser() && !Meteor.loggingIn() && Meteor.user() && !Meteor.user().hasCompletedSignup;
    },                                                                 //
                                                                       //
    accountSettingsUi: function () {                                   // 847
      return makeAccountSettingsUi();                                  // 848
    },                                                                 //
                                                                       //
    firstTimeBillingPromptState: function () {                         // 851
      // Should we show the first-time billing plan selector?          //
                                                                       //
      // Don't show if billing is not enabled.                         //
      if (!window.BlackrockPayments) return;                           // 855
                                                                       //
      var user = Meteor.user();                                        // 857
                                                                       //
      // Don't show if not logged in.                                  //
      if (!user) return;                                               // 860
                                                                       //
      // Don't show if not in the experiment.                          //
      if (!user.experiments || user.experiments.firstTimeBillingPrompt !== "test") return;
                                                                       //
      // Don't show if the user has selected a plan already.           //
      if (user.plan && !Session.get("firstTimeBillingPromptOpen")) return;
                                                                       //
      // Only show to account users (not identities).                  //
      if (!user.loginIdentities) return;                               // 869
                                                                       //
      // Don't show to demo users.                                     //
      if (user.expires) return;                                        // 872
                                                                       //
      // Don't show when viewing another user's grain. We don't want to scare people away from
      // logging in to collaborate.                                    //
      var route = Router.current().route.getName();                    // 876
      if (route === "shared") return;                                  // 877
      if (route === "grain") {                                         // 878
        if (_.some(globalGrains.get(), function (grain) {              // 879
          return grain.isActive() && !grain.isOwner();                 // 880
        })) {                                                          //
                                                                       //
          return;                                                      // 883
        }                                                              //
      }                                                                //
                                                                       //
      // Don't show if user is trying to use a signup key. We're probably going to assign them
      // a plan shortly.                                               //
      if (route === "signup") return;                                  // 889
                                                                       //
      // Don't let the plan chooser disappear instantly once user.plan is set.
      Session.set("firstTimeBillingPromptOpen", true);                 // 892
                                                                       //
      // OK, show it.                                                  //
      return {                                                         // 895
        db: globalDb,                                                  // 896
        topbar: globalTopbar,                                          // 897
        accountsUi: globalAccountsUi,                                  // 898
        onComplete: function () {                                      // 899
          Session.set("firstTimeBillingPromptOpen", false);            // 900
        }                                                              //
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.layout.events({                                             // 906
    "click #demo-expired .logout": function (event) {                  // 907
      logoutSandstorm();                                               // 908
    }                                                                  //
  });                                                                  //
                                                                       //
  credentialsSubscription = Meteor.subscribe("credentials");           // 912
                                                                       //
  Template.registerHelper("dateString", makeDateString);               // 914
  Template.registerHelper("hideNavbar", function () {                  // 915
    // Hide navbar if user is not logged in, since they can't go anywhere with it.
    return !Meteor.userId() || isDemoExpired();                        // 917
  });                                                                  //
                                                                       //
  Template.registerHelper("shrinkNavbar", function () {                // 920
    // Shrink the navbar if the user clicked the button to do so.      //
    return Session.get("shrink-navbar");                               // 922
  });                                                                  //
                                                                       //
  Template.registerHelper("quotaEnabled", function () {                // 925
    return Meteor.settings["public"].quotaEnabled;                     // 926
  });                                                                  //
                                                                       //
  Template.root.helpers({                                              // 929
    storageUsage: function () {                                        // 930
      return Meteor.userId() ? prettySize(Meteor.user().storageUsage || 0) : undefined;
    },                                                                 //
                                                                       //
    storageQuota: function () {                                        // 934
      var plan = globalDb.getMyPlan();                                 // 935
      return plan ? prettySize(plan.storage) : undefined;              // 936
    },                                                                 //
                                                                       //
    overQuota: function () {                                           // 939
      return !window.BlackrockPayments && isUserOverQuota(Meteor.user());
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.root.events({                                               // 944
    "click .uninstall-app-button": function (event) {                  // 945
      // TODO(cleanup): This event handler is no longer used, but the new UI does not yet implement
      //   uninstall. Leave this code here for reference until it does.
      var appId = event.currentTarget.getAttribute("data-appid");      // 948
      if (window.confirm("Really uninstall this app?")) {              // 949
        UserActions.find({ appId: appId, userId: Meteor.userId() }).forEach(function (action) {
          UserActions.remove(action._id);                              // 951
        });                                                            //
                                                                       //
        Meteor.call("deleteUnusedPackages", appId);                    // 954
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationsPopup.helpers({                                // 959
    notifications: function () {                                       // 960
      Meteor.call("readAllNotifications");                             // 961
      return Notifications.find({ userId: Meteor.userId() }, { sort: { timestamp: -1 } }).map(function (row) {
        var grain = Grains.findOne({ _id: row.grainId });              // 963
        if (grain) {                                                   // 964
          row.grainTitle = grain.title;                                // 965
        }                                                              //
                                                                       //
        return row;                                                    // 968
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notifications.helpers({                                     // 973
    notificationCount: function () {                                   // 974
      return Notifications.find({ userId: Meteor.userId(), isUnread: true }).count();
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationsPopup.events({                                 // 979
    "click #notification-dropdown": function (event) {                 // 980
      return false;                                                    // 981
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationItem.helpers({                                  // 985
    isAppUpdates: function () {                                        // 986
      return !!this.appUpdates;                                        // 987
    },                                                                 //
                                                                       //
    notificationTitle: function () {                                   // 990
      if (this.admin) {                                                // 991
        return "Notification from System";                             // 992
      } else if (this.appUpdates) {                                    //
        return "App updates are available";                            // 994
      } else if (this.referral || this.mailingListBonus) {             //
        return false;                                                  // 996
      }                                                                //
                                                                       //
      return this.grainTitle + " is backgrounded";                     // 999
    },                                                                 //
                                                                       //
    titleHelperText: function () {                                     // 1002
      if (this.admin) {                                                // 1003
        return "Dismiss this system notification";                     // 1004
      } else if (this.referral) {                                      //
        return "Dismiss this referral notification";                   // 1006
      } else {                                                         //
        return "Stop the background app";                              // 1008
      }                                                                //
    },                                                                 //
                                                                       //
    dismissText: function () {                                         // 1012
      if (this.admin && this.admin.type === "reportStats") {           // 1013
        return false;                                                  // 1014
      } else if (this.referral) {                                      //
        return "Dismiss";                                              // 1016
      }                                                                //
                                                                       //
      return "Cancel";                                                 // 1019
    },                                                                 //
                                                                       //
    adminLink: function () {                                           // 1022
      return this.admin && this.admin.action;                          // 1023
    },                                                                 //
                                                                       //
    appUpdatesList: function () {                                      // 1026
      return _.values(this.appUpdates);                                // 1027
    },                                                                 //
                                                                       //
    paidUser: function () {                                            // 1030
      var plan = Meteor.user().plan;                                   // 1031
      return plan && plan !== "free";                                  // 1032
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.notificationItem.events({                                   // 1036
    "click .cancel-notification": function (event) {                   // 1037
      Meteor.call("dismissNotification", this._id);                    // 1038
      return false;                                                    // 1039
    },                                                                 //
                                                                       //
    "click .accept-notification": function (event) {                   // 1042
      var _this5 = this;                                               //
                                                                       //
      if (this.appUpdates) {                                           // 1043
        Meteor.call("updateApps", this.appUpdates, function (err) {    // 1044
          // TODO(someday): if (err)                                   //
          Meteor.call("dismissNotification", _this5._id);              // 1046
        });                                                            //
      }                                                                //
                                                                       //
      return false;                                                    // 1050
    },                                                                 //
                                                                       //
    "click .dismiss-notification": function (event) {                  // 1053
      Meteor.call("dismissNotification", this._id);                    // 1054
    }                                                                  //
  });                                                                  //
                                                                       //
  Meteor.startup(function () {                                         // 1058
    // Tell app authors how to run JS in the context of the grain-frame.
    console.log("%cApp authors: To understand the grain-frame in Sandstorm and how to find " + "logs and perform troubleshooting, see: " + "\n- https://docs.sandstorm.io/en/latest/developing/path/ " + "\n- https://docs.sandstorm.io/en/latest/using/top-bar/ " + "\n- https://docs.sandstorm.io/en/latest/developing/troubleshooting/ " + "\n" + "\nWhen debugging, make sure you execute Javascript " + "in the context of the 'grain-frame' IFRAME. References: " + "\n- https://stackoverflow.com/questions/3275816/debugging-iframes-with-chrome-developer-tools " + "\n- https://developer.mozilla.org/en-US/docs/Tools/Working_with_iframes " + "\n" + "\nWe can also provide personal assistance! Get in touch: https://sandstorm.io/community", "font-size: large; background-color: yellow;");
                                                                       //
    Meteor.subscribe("notifications");                                 // 1075
                                                                       //
    Meteor.autorun(function () {                                       // 1077
      Meteor.subscribe("notificationGrains", Notifications.find().map(function (row) {
        return row._id;                                                // 1080
      }));                                                             //
    });                                                                //
  });                                                                  //
                                                                       //
  Meteor.methods({                                                     // 1086
    dismissNotification: function (notificationId) {                   // 1087
      // Client-side simulation of dismissNotification.                //
      Notifications.remove({ _id: notificationId });                   // 1089
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
Router.configure({                                                     // 1094
  layoutTemplate: "layout",                                            // 1095
  notFoundTemplate: "notFound",                                        // 1096
  loadingTemplate: "loading"                                           // 1097
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 1100
  Router.onBeforeAction("loading");                                    // 1101
}                                                                      //
                                                                       //
function getBuildInfo() {                                              // 1104
  var build = Meteor.settings && Meteor.settings["public"] && Meteor.settings["public"].build;
  var isNumber = typeof build === "number";                            // 1106
  if (!build) {                                                        // 1107
    build = "(unknown)";                                               // 1108
  } else if (isNumber) {                                               //
    build = String(Math.floor(build / 1000)) + "." + String(build % 1000);
  }                                                                    //
                                                                       //
  return {                                                             // 1113
    build: build,                                                      // 1114
    isUnofficial: !isNumber                                            // 1115
  };                                                                   //
}                                                                      //
                                                                       //
var promptForFile = function (input, callback) {                       // 1119
  // TODO(cleanup): Share code with "upload picture" and other upload buttons.
  function listener(e) {                                               // 1121
    input.removeEventListener("change", listener);                     // 1122
    callback(e.currentTarget.files[0]);                                // 1123
  }                                                                    //
                                                                       //
  input.addEventListener("change", listener);                          // 1126
  input.click();                                                       // 1127
};                                                                     //
                                                                       //
var startUpload = function (file, endpoint, onComplete) {              // 1130
  // TODO(cleanup): Use Meteor's HTTP, although this may require sending them a PR to support
  //   progress callbacks (and officially document that binary input is accepted).
                                                                       //
  Session.set("uploadStatus", "Uploading");                            // 1134
                                                                       //
  var xhr = new XMLHttpRequest();                                      // 1136
                                                                       //
  xhr.onreadystatechange = function () {                               // 1138
    if (xhr.readyState == 4) {                                         // 1139
      if (xhr.status == 200) {                                         // 1140
        Session.set("uploadProgress", 0);                              // 1141
        onComplete(xhr.responseText);                                  // 1142
      } else {                                                         //
        Session.set("uploadError", {                                   // 1144
          status: xhr.status,                                          // 1145
          statusText: xhr.statusText,                                  // 1146
          response: xhr.responseText                                   // 1147
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  if (xhr.upload) {                                                    // 1153
    xhr.upload.addEventListener("progress", function (progressEvent) {
      Session.set("uploadProgress", Math.floor(progressEvent.loaded / progressEvent.total * 100));
    });                                                                //
  }                                                                    //
                                                                       //
  xhr.open("POST", endpoint, true);                                    // 1160
  xhr.send(file);                                                      // 1161
                                                                       //
  Router.go("uploadStatus");                                           // 1163
};                                                                     //
                                                                       //
restoreBackup = function (file) {                                      // 1166
  // This function is global so tests can call it                      //
  startUpload(file, "/uploadBackup", function (response) {             // 1168
    Session.set("uploadStatus", "Unpacking");                          // 1169
    var identityId = Accounts.getCurrentIdentityId();                  // 1170
    Meteor.call("restoreGrain", response, identityId, function (err, grainId) {
      if (err) {                                                       // 1172
        console.log(err);                                              // 1173
        Session.set("uploadStatus", undefined);                        // 1174
        Session.set("uploadError", {                                   // 1175
          status: "",                                                  // 1176
          statusText: err.message                                      // 1177
        });                                                            //
      } else {                                                         //
        Router.go("grain", { grainId: grainId });                      // 1180
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promptRestoreBackup = function (input) {                               // 1186
  promptForFile(input, restoreBackup);                                 // 1187
};                                                                     //
                                                                       //
uploadApp = function (file) {                                          // 1190
  // This function is global so tests can call it                      //
  Meteor.call("newUploadToken", function (err, token) {                // 1192
    if (err) {                                                         // 1193
      console.error(err);                                              // 1194
      alert(err.message);                                              // 1195
    } else {                                                           //
      startUpload(file, "/upload/" + token, function (response) {      // 1197
        Session.set("uploadStatus", undefined);                        // 1198
        Router.go("install", { packageId: response });                 // 1199
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
promptUploadApp = function (input) {                                   // 1205
  promptForFile(input, uploadApp);                                     // 1206
};                                                                     //
                                                                       //
Router.map(function () {                                               // 1209
  this.route("root", {                                                 // 1210
    path: "/",                                                         // 1211
    waitOn: function () {                                              // 1212
      return [Meteor.subscribe("hasUsers"), Meteor.subscribe("grainsMenu")];
    },                                                                 //
                                                                       //
    data: function () {                                                // 1219
      // If the user is logged-in, and can create new grains, and      //
      // has no grains yet, then send them to "new".                   //
      if (this.ready() && Meteor.userId() && !Meteor.loggingIn()) {    // 1222
        if (globalDb.currentUserGrains({}, {}).count() === 0 && globalDb.currentUserApiTokens().count() === 0) {
          Router.go("apps", {}, { replaceState: true });               // 1225
        } else {                                                       //
          Router.go("grains", {}, { replaceState: true });             // 1227
        }                                                              //
      }                                                                //
                                                                       //
      return {                                                         // 1231
        needsAdminTokenLogin: this.ready() && !HasUsers.findOne("hasUsers") && !globalDb.allowDevAccounts(),
        build: getBuildInfo().build,                                   // 1233
        splashUrl: (Settings.findOne("splashUrl") || {}).value         // 1234
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("linkHandler", {                                          // 1239
    path: "/link-handler/:url",                                        // 1240
                                                                       //
    data: function () {                                                // 1242
      var url = this.params.url;                                       // 1243
      if (url.lastIndexOf("web+sandstorm:", 0) === 0) {                // 1244
        url = url.slice("web+sandstorm:".length);                      // 1245
      }                                                                //
      // TODO(cleanup):  Didn't use Router.go() because the url may contain a query term.
      document.location = "/install/" + url;                           // 1248
      return {};                                                       // 1249
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("about", {                                                // 1253
    path: "/about",                                                    // 1254
    data: function () {                                                // 1255
      var result = getBuildInfo();                                     // 1256
                                                                       //
      var backers = Session.get("backers");                            // 1258
      if (backers) {                                                   // 1259
        result.backers = backers.names;                                // 1260
        result.anonCount = backers.anonCount;                          // 1261
      } else {                                                         //
        HTTP.get("/sandstorm-backers.txt", function (err, response) {  // 1263
          var names = response.content.split("\n").sort(function (a, b) {
            return a.toLowerCase().localeCompare(b.toLowerCase());     // 1265
          });                                                          //
                                                                       //
          var anonCount = 0;                                           // 1268
          while (anonCount < names.length && names[anonCount] === "") {
            ++anonCount;                                               // 1270
          }                                                            //
                                                                       //
          names = names.slice(anonCount);                              // 1273
                                                                       //
          // File ends in trailing newline, but that last blank line does not represent an
          // anonymous contributor.                                    //
          --anonCount;                                                 // 1277
                                                                       //
          Session.set("backers", {                                     // 1279
            names: names,                                              // 1280
            anonCount: anonCount                                       // 1281
          });                                                          //
        });                                                            //
      }                                                                //
                                                                       //
      result.termsUrl = globalDb.getSetting("termsUrl");               // 1286
      result.privacyUrl = globalDb.getSetting("privacyUrl");           // 1287
                                                                       //
      return result;                                                   // 1289
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("uploadStatus", {                                         // 1293
    path: "/upload",                                                   // 1294
                                                                       //
    waitOn: function () {                                              // 1296
      return Meteor.subscribe("credentials");                          // 1297
    },                                                                 //
                                                                       //
    data: function () {                                                // 1300
      return {                                                         // 1301
        progress: Session.get("uploadProgress"),                       // 1302
        status: Session.get("uploadStatus"),                           // 1303
        error: Session.get("uploadError")                              // 1304
      };                                                               //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("referrals", {                                            // 1309
    path: "/referrals",                                                // 1310
                                                                       //
    waitOn: function () {                                              // 1312
      return Meteor.subscribe("referralInfoPseudo");                   // 1313
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("account", {                                              // 1317
    path: "/account",                                                  // 1318
                                                                       //
    data: function () {                                                // 1320
      // Don't allow logged-out or demo users to visit the accounts page. There should be no way
      // for them to get there except for typing the URL manually. In theory showing the accounts
      // page to a demo user could make some sense for editing their profile, but we really do not
      // want them signing up for subscription plans!                  //
      if (!Meteor.user() && !Meteor.loggingIn() || globalDb.isDemoUser()) {
        Router.go("root", {}, { replaceState: true });                 // 1326
      } else {                                                         //
        return makeAccountSettingsUi();                                // 1328
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("accountUsage", {                                         // 1333
    path: "/account/usage"                                             // 1334
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=shell.js.map
