(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var ECMAScript = Package.ecmascript.ECMAScript;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;
var AccountsServer = Package['accounts-base'].AccountsServer;
var HTTP = Package.http.HTTP;
var HTTPInternals = Package.http.HTTPInternals;
var babelHelpers = Package['babel-runtime'].babelHelpers;
var Symbol = Package['ecmascript-runtime'].Symbol;
var Map = Package['ecmascript-runtime'].Map;
var Set = Package['ecmascript-runtime'].Set;
var Promise = Package.promise.Promise;

/* Package-scope variables */
var Packages, DevPackages, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, ApiHosts, Notifications, StatsTokens, Misc, Settings, Migrations, StaticAssets, AssetUploadTokens, Plans, AppIndex, KeybaseProfiles, isDemoUser, isSignedUp, isSignedUpOrDemo, successfulReferralsCount, getUserQuota, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, sendReferralProgramNotification, roleAssignmentPattern, SandstormDb, appName, addStaticAsset, userPictureUrl, fetchPicture;

(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/db.js                                                                                        //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                                 //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                                              //
// All rights reserved.                                                                                               //
//                                                                                                                    //
// Licensed under the Apache License, Version 2.0 (the "License");                                                    //
// you may not use this file except in compliance with the License.                                                   //
// You may obtain a copy of the License at                                                                            //
//                                                                                                                    //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                       //
//                                                                                                                    //
// Unless required by applicable law or agreed to in writing, software                                                //
// distributed under the License is distributed on an "AS IS" BASIS,                                                  //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                           //
// See the License for the specific language governing permissions and                                                //
// limitations under the License.                                                                                     //
                                                                                                                      //
// This file defines the database schema.                                                                             //
                                                                                                                      //
// Useful for debugging: Set the env variable LOG_MONGO_QUERIES to have the server write every                        //
// query it makes, so you can see if it's doing queries too often, etc.                                               //
if (Meteor.isServer && process.env.LOG_MONGO_QUERIES) {                                                               // 21
  (function () {                                                                                                      //
    var oldFind = Mongo.Collection.prototype.find;                                                                    // 22
    Mongo.Collection.prototype.find = function () {                                                                   // 23
      console.log(this._prefix, arguments);                                                                           // 24
      return oldFind.apply(this, arguments);                                                                          // 25
    };                                                                                                                //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
// Helper so that we don't have to if (Meteor.isServer) before declaring indexes.                                     //
if (Meteor.isServer) {                                                                                                // 30
  Mongo.Collection.prototype.ensureIndexOnServer = Mongo.Collection.prototype._ensureIndex;                           // 31
} else {                                                                                                              //
  Mongo.Collection.prototype.ensureIndexOnServer = function () {};                                                    // 33
}                                                                                                                     //
                                                                                                                      //
// TODO(soon): Systematically go through this file and add ensureIndexOnServer() as needed.                           //
                                                                                                                      //
// Users = new Mongo.Collection("users");                                                                             //
// The users collection is special and can be accessed through `Meteor.users`.                                        //
// See https://docs.meteor.com/#/full/meteor_users.                                                                   //
//                                                                                                                    //
// There are two distinct types of entries in the users collection: identities and accounts. An                       //
// identity contains personal profile information and typically includes some intrinsic method for                    //
// authenticating as the owner of that information.                                                                   //
//                                                                                                                    //
// An account is an owner of app actions, grains, contacts, notifications, and payment info.                          //
// Each account can have multiple identities linked to it. To log in as an account you must first                     //
// authenticate as one of its linked identities.                                                                      //
//                                                                                                                    //
// Every user contains the following fields:                                                                          //
//   _id: Unique string ID. For accounts, this is random. For identities, this is the globally                        //
//        stable SHA-256 ID of this identity, hex-encoded.                                                            //
//   createdAt: Date when this entry was added to the collection.                                                     //
//   lastActive: Date of the user's most recent interaction with this Sandstorm server.                               //
//   services: Object containing login data used by Meteor authentication services.                                   //
//   expires: Date when this user should be deleted. Only present for demo users.                                     //
//   upgradedFromDemo: If present, the date when this user was upgraded from being a demo user.                       //
//                     TODO(cleanup): Unlike other dates in our database, this is stored as a number                  //
//                     rather than as a Date object. We should fix that.                                              //
//   appDemoId: If present and non-null, then the user is a demo user who arrived via an /appdemo/                    //
//              link. This field contains the app ID of the app that the user started out demoing.                    //
//              Unlike the `expires` field, this field is not cleared when the user upgrades from                     //
//              being a demo user.                                                                                    //
//                                                                                                                    //
// Identity users additionally contain the following fields:                                                          //
//   profile: Object containing the data that will be shared with users and grains that come into                     //
//            contact with this identity. Includes the following fields:                                              //
//       service: String containing the name of this identity's authentication method.                                //
//       name: String containing the chosen display name of the identity.                                             //
//       handle: String containing the identity's preferred handle.                                                   //
//       picture: _id into the StaticAssets table for the identity's picture. If not present,                         //
//                an identicon will be used.                                                                          //
//       pronoun: One of "male", "female", "neutral", or "robot".                                                     //
//   unverifiedEmail: If present, a string containing an email address specified by the user.                         //
//   referredBy: ID of the Account that referred this Identity.                                                       //
//                                                                                                                    //
// Account users additionally contain the following fields:                                                           //
//   loginIdentities: Array of identity objects, each of which may include the following fields.                      //
//       id: The globally-stable SHA-256 ID of this identity, hex-encoded.                                            //
//   nonloginIdentities: Array of identity objects, of the same form as `loginIdentities`. We use                     //
//                       a separate array here so that we can use a Mongo index to enforce the                        //
//                       invariant that an identity only be a login identity for a single account.                    //
//   primaryEmail: String containing this account's primary email address. Must be a verified adress                  //
//                 of one of this account's linked identities. Call SandstormDb.getUserEmails()                       //
//                 to do this checking automatically.                                                                 //
//   isAdmin: Boolean indicating whether this account is allowed to access the Sandstorm admin panel.                 //
//   signupKey: If this is an invited user, then this field contains their signup key.                                //
//   signupNote: If the user was invited through a link, then this field contains the note that the                   //
//               inviter admin attached to the key.                                                                   //
//   signupEmail: If the user was invited by email, then this field contains the email address that                   //
//                the invite was sent to.                                                                             //
//   hasCompletedSignup: True if this account has confirmed its profile and agreed to this server's                   //
//                       terms of service.                                                                            //
//   plan: _id of an entry in the Plans table which determines the user's quota.                                      //
//   planBonus: {storage, compute, grains} bonus amounts to add to the user's plan. The payments                      //
//              module writes data here; we merely read it. Missing fields should be treated as                       //
//              zeroes. Does not yet include referral bonus, which is calculated separately.                          //
//              TODO(cleanup): Use for referral bonus too.                                                            //
//   storageUsage: Number of bytes this user is currently storing.                                                    //
//   payments: Object defined by payments module, if loaded.                                                          //
//   dailySentMailCount: Number of emails sent by this user today; used to limit spam.                                //
//   accessRequests: Object containing the following fields; used to limit spam.                                      //
//       count: Number of "request access" emails during sent during the current interval.                            //
//       resetAt: Date when the count should be reset.                                                                //
//   referredByComplete: ID of the Account that referred this Account. If this is set, we                             //
//                        stop writing new referredBy values onto Identities for this account.                        //
//   referredCompleteDate: The Date at which the completed referral occurred.                                         //
//   referredIdentityIds: List of Identity IDs that this Account has referred. This is used for                       //
//                        reliably determining which Identity's names are safe to display.                            //
//   experiments: Object where each field is an experiment that the user is in, and each value                        //
//           is the parameters for that experiment. Typically, the value simply names which                           //
//           experiment group which the user is in, where "control" is one group. If an experiment                    //
//           is not listed, then the user should not be considered at all for the purpose of that                     //
//           experiment. Each experiment may define a point in time where users not already in the                    //
//           experiment may be added to it and assigned to a group (for example, at user creation                     //
//           time). Current experiments:                                                                              //
//       firstTimeBillingPrompt: Value is "control" or "test". Users are assigned to groups at                        //
//               account creation on servers where billing is enabled (i.e. Oasis). Users in the                      //
//               test group will see a plan selection dialog and asked to make an explitic choice                     //
//               (possibly "free") before they can create grains (but not when opening someone                        //
//               else's shared grain). The goal of the experiment is to determine whether this                        //
//               prompt scares users away -- and also whether it increases paid signups.                              //
//   stashedOldUser: A complete copy of this user from before the accounts/identities migration.                      //
//                   TODO(cleanup): Delete this field once we're sure it's safe to do so.                             //
                                                                                                                      //
Meteor.users.ensureIndexOnServer("services.google.email", { sparse: 1 });                                             // 126
Meteor.users.ensureIndexOnServer("services.github.emails.email", { sparse: 1 });                                      // 127
Meteor.users.ensureIndexOnServer("services.email.email", { unique: 1, sparse: 1 });                                   // 128
Meteor.users.ensureIndexOnServer("loginIdentities.id", { unique: 1, sparse: 1 });                                     // 129
Meteor.users.ensureIndexOnServer("nonloginIdentities.id", { sparse: 1 });                                             // 130
Meteor.users.ensureIndexOnServer("services.google.id", { unique: 1, sparse: 1 });                                     // 131
Meteor.users.ensureIndexOnServer("services.github.id", { unique: 1, sparse: 1 });                                     // 132
                                                                                                                      //
// TODO(cleanup): This index is obsolete; delete it.                                                                  //
Meteor.users.ensureIndexOnServer("identities.id", { unique: 1, sparse: 1 });                                          // 135
                                                                                                                      //
Packages = new Mongo.Collection("packages");                                                                          // 137
// Packages which are installed or downloading.                                                                       //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                                   //
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"                  //
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),                               //
//       >1 = download byte count.                                                                                    //
//   error:  If status is "failed", error message string.                                                             //
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in package.capnp.                         //
//   appId:  If status is "ready", the application ID string.  Packages representing different                        //
//       versions of the same app have the same appId.  The spk tool defines the app ID format                        //
//       and can cryptographically verify that a package belongs to a particular app ID.                              //
//   shouldCleanup:  If true, a reference to this package was recently dropped, and the package                       //
//       collector should at some point check whether there are any other references and, if not,                     //
//       delete the package.                                                                                          //
//   url:  When status is "download", the URL from which the SPK can be obtained, if provided.                        //
//   isAutoUpdated: This package was downloaded as part of an auto-update. We shouldn't clean it up                   //
//     even if it has no users.                                                                                       //
//   authorPgpKeyFingerprint: Verified PGP key fingerprint (SHA-1, hex, all-caps) of the app                          //
//     packager.                                                                                                      //
                                                                                                                      //
DevPackages = new Mongo.Collection("devpackages");                                                                    // 159
// List of packages currently made available via the dev tools running on the local machine.                          //
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool                    //
// on the local machine to publish an under-development app to this server. That should only ever                     //
// happen on developers' desktop machines.                                                                            //
//                                                                                                                    //
// While a dev package is published, it automatically appears as installed by every user of the                       //
// server, and it overrides all packages with the same application ID. If any instances of those                      //
// packages are currently open, they are killed and reset on publish.                                                 //
//                                                                                                                    //
// When the dev tool disconnects, the package is automatically unpublished, and any open instances                    //
// are again killed and refreshed.                                                                                    //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  The package ID string (as with Packages._id).                                                              //
//   appId: The app ID this package is intended to override (as with Packages.appId).                                 //
//   timestamp:  Time when the package was last updated. If this changes while the package is                         //
//     published, all running instances are reset. This is used e.g. to reset the app each time                       //
//     changes are made to the source code.                                                                           //
//   manifest:  The app's manifest, as with Packages.manifest.                                                        //
                                                                                                                      //
UserActions = new Mongo.Collection("userActions");                                                                    // 180
// List of actions that each user has installed which create new grains.  Each app may install                        //
// some number of actions (usually, one).                                                                             //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   userId:  User who has installed this action.                                                                     //
//   packageId:  Package used to run this action.                                                                     //
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                               //
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle; denormalized so                                //
//       that clients can access it without subscribing to the Packages collection.                                   //
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                           //
//       searchability.                                                                                               //
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"                              //
//   title: JSON-encoded LocalizedText title for this action, e.g.                                                    //
//       `{defaultText: "New Spreadsheet"}`.                                                                          //
//   nounPhrase: JSON-encoded LocalizedText describing what is created when this action is run.                       //
//   command:  Manifest.Command to run this action (see package.capnp).                                               //
                                                                                                                      //
Grains = new Mongo.Collection("grains");                                                                              // 199
// Grains belonging to users.                                                                                         //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   packageId:  _id of the package of which this grain is an instance.                                               //
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.                               //
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for                           //
//       searchability.                                                                                               //
//   userId: The _id of the account that owns this grain.                                                             //
//   identityId: The identity with which the owning account prefers to open this grain.                               //
//   title:  Human-readable string title, as chosen by the user.                                                      //
//   lastUsed:  Date when the grain was last used by a user.                                                          //
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                                   //
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent                   //
//                   time a session to this grain was opened.                                                         //
//                                                                                                                    //
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and                    //
// web publishing functionality without powerbox support; they will be replaced once the powerbox                     //
// is implemented.                                                                                                    //
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and                    //
//       web publishing. This field is initialized when first requested by the app.                                   //
                                                                                                                      //
RoleAssignments = new Mongo.Collection("roleAssignments");                                                            // 222
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges                   //
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens                            //
// collection.                                                                                                        //
                                                                                                                      //
Contacts = new Mongo.Collection("contacts");                                                                          // 227
// Edges in the social graph.                                                                                         //
//                                                                                                                    //
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob                        //
// will show up in her user-picker UI for actions like share-by-identity.                                             //
//                                                                                                                    //
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of                       //
// Bob's.                                                                                                             //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: random                                                                                                      //
//   ownerId: The accountId of the user account who owns this contact.                                                //
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies                       //
//            the contact to the owner.                                                                               //
//   created: Date when this contact was created.                                                                     //
//   identityId: The `_id` of the user whose contact info this contains.                                              //
                                                                                                                      //
Sessions = new Mongo.Collection("sessions");                                                                          // 244
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.                   //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   grainId:  _id of the grain to which this session is connected.                                                   //
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the                    //
//       '*' in WILDCARD_HOST.                                                                                        //
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some                       //
//       period.                                                                                                      //
//   userId:  User ID of the user who owns this session.                                                              //
//   identityId:  Identity ID of the user who owns this session.                                                      //
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens                     //
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),                        //
//       anonymous users can get access without an API token and so neither userId nor hashedToken                    //
//       are present.                                                                                                 //
//   powerboxView: If present, this is a view that should be presented as part of a powerbox                          //
//       interaction.                                                                                                 //
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                                  //
//   viewInfo: The UiView.ViewInfo corresponding to the underlying UiSession. This isn't populated                    //
//       until newSession is called on the UiView.                                                                    //
//   permissions: The permissions for the current identity on this UiView. This isn't populated                       //
//       until newSession is called on the UiView.                                                                    //
//   hasLoaded: Marked as true by the proxy when the underlying UiSession has responded to its first                  //
//       request                                                                                                      //
                                                                                                                      //
SignupKeys = new Mongo.Collection("signupKeys");                                                                      // 270
// Invite keys which may be used by users to get access to Sandstorm.                                                 //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:  random                                                                                                     //
//   used:  Boolean indicating whether this key has already been consumed.                                            //
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.                         //
//   email: If this key was sent as an email invite, the email address to which it was sent.                          //
                                                                                                                      //
ActivityStats = new Mongo.Collection("activityStats");                                                                // 279
// Contains usage statistics taken on a regular interval. Each entry is a data point.                                 //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   timestamp: Date when measurements were taken.                                                                    //
//   daily: Contains stats counts pertaining to the last day before the sample time.                                  //
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.                          //
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.                          //
//                                                                                                                    //
// Each of daily, weekly, and monthly contains:                                                                       //
//   activeUsers: The number of unique users who have used a grain on the server in the time                          //
//       interval. Only counts logged-in users.                                                                       //
//   demoUsers: Demo users.                                                                                           //
//   appDemoUsers: Users that came in through "app demo".                                                             //
//   activeGrains: The number of unique grains that have been used in the time interval.                              //
//   apps: An object indexed by app ID recording, for each app:                                                       //
//       owners: Number of unique owners of this app (counting only grains that still exist).                         //
//       sharedUsers: Number of users who have accessed other people's grains of this app (counting                   //
//         only grains that still exist).                                                                             //
//       grains: Number of active grains of this app (that still exist).                                              //
//       deleted: Number of non-demo grains of this app that were deleted.                                            //
//       demoed: Number of demo grains created and expired.                                                           //
//       appDemoUsers: Number of app demos initiated with this app.                                                   //
                                                                                                                      //
DeleteStats = new Mongo.Collection("deleteStats");                                                                    // 303
// Contains records of objects that were deleted, for stat-keeping purposes.                                          //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   type: "grain" or "user" or "demoGrain" or "demoUser" or "appDemoUser"                                            //
//   lastActive: Date of the user's or grain's last activity.                                                         //
//   appId: For type = "grain", the app ID of the grain. For type = "appDemoUser", the app ID they                    //
//     arrived to demo. For others, undefined.                                                                        //
                                                                                                                      //
FileTokens = new Mongo.Collection("fileTokens");                                                                      // 312
// Tokens corresponding to backup files that are currently stored on the server. A user receives                      //
// a token when they create a backup file (either by uploading it, or by backing up one of their                      //
// grains) and may use the token to read the file (either to download it, or to restore a new                         //
// grain from it).                                                                                                    //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The unguessable token string.                                                                         //
//   name:      Suggested filename.                                                                                   //
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.                       //
                                                                                                                      //
ApiTokens = new Mongo.Collection("apiTokens");                                                                        // 323
// Access tokens for APIs exported by apps.                                                                           //
//                                                                                                                    //
// Originally API tokens were only used by external users through the HTTP API endpoint. However,                     //
// now they are also used to implement SturdyRefs, not just held by external users, but also when                     //
// an app holds a SturdyRef to another app within the same server. See the various `save()`,                          //
// `restore()`, and `drop()` methods in `grain.capnp` (on `SandstormApi`, `AppPersistent`, and                        //
// `MainView`) -- the fields of type `Data` are API tokens.                                                           //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       A SHA-256 hash of the token, base64-encoded.                                                          //
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)                     //
//   identityId: For UiView capabilities, this is the identity for which the view is attenuated.                      //
//              That is, the UiView's newSession() method will intersect the requested permissions                    //
//              with this identity's permissions before forwarding on to the underlying app. If                       //
//              `identityId` is not present, then no identity attenuation is applied, i.e. this is                    //
//              a raw UiView as implemented by the app. (The `roleAssignment` field, below, may                       //
//              still apply. For non-UiView capabilities, `identityId` is never present. Note that                    //
//              this is NOT the identity against which the `requiredPermissions` parameter of                         //
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerIdentity`.                  //
//   accountId: For tokens where `identityId` is set, the `_id` (in the Users table) of the account                   //
//              that created the token.                                                                               //
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded                        //
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These                   //
//              permissions will be intersected with those held by `identityId` when the view is                      //
//              opened.                                                                                               //
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as                   //
//              anonymous rather than as directly associated with `identityId`. This has no effect                    //
//              on the permissions granted.                                                                           //
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by                     //
//              the app or its supervisor (whereas without this it strictly represents UiView).                       //
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).                     //
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is                            //
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.                              //
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,                   //
//              not a particular grain. (`grainId` and `identityId` are not set.) This is an object                   //
//              containing exactly one of the following fields:                                                       //
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                                   //
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the                          //
//                           `Notifications` collection.                                                              //
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this                     //
//                  will be moved out of the frontend and into the backend, but we'll migrate the                     //
//                  database when that happens. This field contains the boolean true to signify that                  //
//                  it has been set.                                                                                  //
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                                    //
//   parentToken: If present, then this token represents exactly the capability represented by                        //
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated                   //
//              by `roleAssignment` (if present). To facilitate permissions computations, if the                      //
//              capability is a UiView, then `grainId` is set to the backing grain, `identityId`                      //
//              is set to the identity that shared the view, and `accountId` is set to the account                    //
//              that shared the view. Neither `objectId` nor `frontendRef` is present when                            //
//              `parentToken` is present.                                                                             //
//   petname:   Human-readable label for this access token, useful for identifying tokens for                         //
//              revocation. This should be displayed when visualizing incoming capabilities to                        //
//              the grain identified by `grainId`.                                                                    //
//   created:   Date when this token was created.                                                                     //
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may                   //
//              become un-revoked in the future.                                                                      //
//   expires:   Optional expiration Date. If undefined, the token does not expire.                                    //
//   owner:     A `ApiTokenOwner` (defined in `supervisor.capnp`, stored as a JSON object)                            //
//              as passed to the `save()` call that created this token. If not present, treat                         //
//              as `webkey` (the default for `ApiTokenOwner`).                                                        //
//   expiresIfUnused:                                                                                                 //
//              Optional Date after which the token, if it has not been used yet, expires.                            //
//              This field should be cleared on a token's first use.                                                  //
//   requirements: List of conditions which must hold for this token to be considered valid.                          //
//              Semantically, this list specifies the powers which were *used* to originally                          //
//              create the token. If any condition in the list becomes untrue, then the token must                    //
//              be considered revoked, and all live refs and sturdy refs obtained transitively                        //
//              through it must also become revoked. Each item is the JSON serialization of the                       //
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                                        //
//   hasApiHost: If true, there is an entry in ApiHosts for this token, which will need to be                         //
//              cleaned up when the token is.                                                                         //
//                                                                                                                    //
// It is important to note that a token's owner and provider are independent from each other. To                      //
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:                   //
//                                                                                                                    //
// struct ApiToken {                                                                                                  //
//   owner :ApiTokenOwner;                                                                                            //
//   provider :union {                                                                                                //
//     grain :group {                                                                                                 //
//       grainId :Text;                                                                                               //
//       union {                                                                                                      //
//         uiView :group {                                                                                            //
//           identityId :Text;                                                                                        //
//           roleAssignment :RoleAssignment;                                                                          //
//           forSharing :Bool;                                                                                        //
//         }                                                                                                          //
//         objectId :SupervisorObjectId;                                                                              //
//       }                                                                                                            //
//     }                                                                                                              //
//     frontendRef :union {                                                                                           //
//        notificationHandle :Text;                                                                                   //
//        ipNetwork :Bool;                                                                                            //
//        ipInterface :Bool;                                                                                          //
//     }                                                                                                              //
//     child :group {                                                                                                 //
//       parentToken :Text;                                                                                           //
//       union {                                                                                                      //
//         uiView :group {                                                                                            //
//           grainId :Text;                                                                                           //
//           identityId :Text;                                                                                        //
//           roleAssignment :RoleAssignment = (allAccess = ());                                                       //
//         }                                                                                                          //
//         other :Void;                                                                                               //
//       }                                                                                                            //
//     }                                                                                                              //
//   }                                                                                                                //
//   requirements: List(Supervisor.MembraneRequirement);                                                              //
//   ...                                                                                                              //
// }                                                                                                                  //
                                                                                                                      //
ApiHosts = new Mongo.Collection("apiHosts");                                                                          // 435
// Allows defining some limited static behavior for an API host when accessed unauthenticated. This                   //
// mainly exists to allow backwards-compatibility with client applications that expect to be able                     //
// to probe an API host without authentication to determine capabilities such as DAV protocols                        //
// supported, before authenticating to perform real requests. An app can specify these properties                     //
// when creating an offerTemplate.                                                                                    //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:          apiHostIdHashForToken() of the corresponding API token.                                            //
//   hash2:        hash(hash(token)), aka hash(ApiToken._id). Used to allow ApiHosts to be cleaned                    //
//                 up when ApiTokens are deleted.                                                                     //
//   options:      Specifies how to respond to unauthenticated OPTIONS requests on this host.                         //
//                 This is an object containing fields:                                                               //
//     dav:        List of strings specifying DAV header `compliance-class`es, e.g. "1" or                            //
//                 "calendar-access". https://tools.ietf.org/html/rfc4918#section-10.1                                //
//   resources:    Object mapping URL paths (including initial '/') to static HTTP responses to                       //
//                 give when those paths are accessed unauthenticated. Due to Mongo disliking '.'                     //
//                 and '$' in keys, these characters must be escaped as '\uFF0E' and '\uFF04'                         //
//                 (see SandstormDb.escapeMongoKey). Each value in this map is an object with                         //
//                 fields:                                                                                            //
//     type:       Content-Type.                                                                                      //
//     language:   Content-Language.                                                                                  //
//     encoding:   Content-Encoding.                                                                                  //
//     body:       Entity-body as a string or buffer.                                                                 //
                                                                                                                      //
Notifications = new Mongo.Collection("notifications");                                                                // 460
// Notifications for a user.                                                                                          //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:          random                                                                                             //
//   grainId:      The grain originating this notification, if any.                                                   //
//   userId:       The user receiving the notification.                                                               //
//   text:         The JSON-ified LocalizedText to display in the notification.                                       //
//   isUnread:     Boolean indicating if this notification is unread.                                                 //
//   timestamp:    Date when this notification was last updated                                                       //
//   ongoing:      If present, this is an ongoing notification, and this field contains an                            //
//                 ApiToken referencing the `OngoingNotification` capability.                                         //
//   admin:        If present, this is a notification intended for an admin.                                          //
//     action:     If present, this is a (string) link that the notification should direct the                        //
//                 admin to.                                                                                          //
//     type:       The type of notification (currently only "reportStats").                                           //
//   appUpdates:   If present, this is an app update notification. It is an object with the appIds                    //
//                 as keys.                                                                                           //
//     $appId:     The appId that has an outstanding update.                                                          //
//       packageId: The packageId that it will update to.                                                             //
//       name: The name of the app. (appTitle from package.manifest)                                                  //
//       version: The app's version number. (appVersion from package.manifest)                                        //
//       marketingVersion: String marketing version of this app. (appMarketingVersion from package.manifest)          //
//   referral:     If this boolean field is true, then treat this notification as a referral                          //
//                 notification. This causes text to be ignored, since we need custom logic.                          //
//   mailingListBonus: Like `referral`, but notify the user about the mailing list bonus. This is                     //
//                 a one-time notification only to Oasis users who existed when the bonus program                     //
//                 was implemented.                                                                                   //
                                                                                                                      //
StatsTokens = new Mongo.Collection("statsTokens");                                                                    // 489
// Access tokens for the Stats collection                                                                             //
//                                                                                                                    //
// These tokens are used for accessing the ActivityStats collection remotely                                          //
// (ie. from a dashboard webapp)                                                                                      //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                                 //
                                                                                                                      //
Misc = new Mongo.Collection("misc");                                                                                  // 498
// Miscellaneous configuration and other settings                                                                     //
//                                                                                                                    //
// This table is currently only used for persisting BASE_URL from one session to the next,                            //
// but in general any miscellaneous settings should go in here                                                        //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The name of the setting. eg. "BASE_URL"                                                               //
//   value:     The value of the setting.                                                                             //
                                                                                                                      //
Settings = new Mongo.Collection("settings");                                                                          // 508
// Settings for this Sandstorm instance go here. They are configured through the adminSettings                        //
// route. This collection differs from misc in that any admin user can update it through the admin                    //
// interface.                                                                                                         //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       The name of the setting. eg. "MAIL_URL"                                                               //
//   value:     The value of the setting.                                                                             //
//   automaticallyReset: Sometimes the server needs to automatically reset a setting. When it does                    //
//                       so, it will also write an object to this field indicating why the reset was                  //
//                       needed. That object can have the following variants:                                         //
//       baseUrlChangedFrom: The reset was due to BASE_URL changing. This field contains a string                     //
//                           with the old BASE_URL.                                                                   //
//                                                                                                                    //
//   potentially other fields that are unique to the setting                                                          //
                                                                                                                      //
Migrations = new Mongo.Collection("migrations");                                                                      // 524
// This table tracks which migrations we have applied to this instance.                                               //
// It contains a single entry:                                                                                        //
//   _id:       "migrations_applied"                                                                                  //
//   value:     The number of migrations this instance has successfully completed.                                    //
                                                                                                                      //
StaticAssets = new Mongo.Collection("staticAssets");                                                                  // 530
// Collection of static assets served up from the Sandstorm server's "static" host. We only                           //
// support relatively small assets: under 1MB each.                                                                   //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       Random ID; will be used in the URL.                                                                   //
//   hash:      A base64-encoded SHA-256 hash of the data, used to de-dupe.                                           //
//   mimeType:  MIME type of the asset, suitable for Content-Type header.                                             //
//   encoding:  Either "gzip" or not present, suitable for Content-Encoding header.                                   //
//   content:   The asset content (byte buffer).                                                                      //
//   refcount:  Number of places where this asset's ID appears in the database. Since Mongo doesn't                   //
//       have transactions, this needs to bias towards over-counting; a backup GC could be used                       //
//       to catch leaked assets, although it's probably not a big deal in practice.                                   //
                                                                                                                      //
AssetUploadTokens = new Mongo.Collection("assetUploadTokens");                                                        // 544
// Collection of tokens representing a single-use permission to upload an asset, such as a new                        //
// profile picture.                                                                                                   //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id:       Random ID.                                                                                            //
//   purpose:   Contains one of the following, indicating how the asset is to be used:                                //
//       profilePicture: Indicates that the upload is a new profile picture. Contains fields:                         //
//           userId: User whose picture shall be replaced.                                                            //
//           identityId: Which of the user's identities shall be updated.                                             //
//   expires:   Time when this token will go away if unused.                                                          //
                                                                                                                      //
Plans = new Mongo.Collection("plans");                                                                                // 556
// Subscription plans, which determine quota.                                                                         //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: Plan ID, usually a short string like "free", "standard", "large", "mega", ...                               //
//   storage: Number of bytes this user is allowed to store.                                                          //
//   compute: Number of kilobyte-RAM-seconds this user is allowed to consume.                                         //
//   computeLabel: Label to display to the user describing this plan's compute units.                                 //
//   grains: Total number of grains this user can create (often `Infinity`).                                          //
//   price: Price per month in US cents.                                                                              //
//   hidden: If true, a user cannot switch to this plan, but some users may be on it and are                          //
//       allowed to switch away.                                                                                      //
//   title: Title from display purposes. If missing, default to capitalizing _id.                                     //
                                                                                                                      //
AppIndex = new Mongo.Collection("appIndex");                                                                          // 570
// A mirror of the data from the App Market index                                                                     //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: the appId of the app                                                                                        //
//  The rest of the fields are defined in src/sandstorm/app-index/app-index.capnp:AppIndexForMarket                   //
                                                                                                                      //
KeybaseProfiles = new Mongo.Collection("keybaseProfiles");                                                            // 577
// Cache of Keybase profile information. The profile for a user is re-fetched every time a package                    //
// by that user is installed, as well as if the keybase profile is requested and not already                          //
// present for some reason.                                                                                           //
//                                                                                                                    //
// Each contains:                                                                                                     //
//   _id: PGP key fingerprint (SHA-1, hex, all-caps)                                                                  //
//   displayName: Display name from Keybase. (NOT VERIFIED AT ALL.)                                                   //
//   handle: Keybase handle.                                                                                          //
//   proofs: The "proofs_summary.all" array from the Keybase lookup. See the non-existent Keybase                     //
//     docs for details. We also add a boolean "status" field to each proof indicating whether                        //
//     we have directly verified the proof ourselves. Its values may be "unverified" (Keybase                         //
//     returned this but we haven't checked it directly), "verified" (we verified the proof and it                    //
//     is valid), "invalid" (we checked the proof and it was definitely bogus), or "checking" (the                    //
//     server is currently actively checking this proof). Note that if a check fails due to network                   //
//     errors, the status goes back to "unverified".                                                                  //
//                                                                                                                    //
//     WARNING: Currently verification is NOT IMPLEMENTED, so all proofs will be "unverified"                         //
//       for now and we just trust Keybase.                                                                           //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 597
  Meteor.publish("credentials", function () {                                                                         // 598
    // Data needed for isSignedUp() and isAdmin() to work.                                                            //
                                                                                                                      //
    if (this.userId) {                                                                                                // 601
      return [Meteor.users.find({ _id: this.userId }, { fields: { signupKey: 1, isAdmin: 1, expires: 1, storageUsage: 1,
          plan: 1, planBonus: 1, hasCompletedSignup: 1, experiments: 1,                                               // 605
          referredIdentityIds: 1 } }), Plans.find()];                                                                 // 606
    } else {                                                                                                          //
      return [];                                                                                                      // 610
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
isDemoUser = function () {                                                                                            // 615
  // Returns true if this is a demo user.                                                                             //
                                                                                                                      //
  var user = Meteor.user();                                                                                           // 618
  if (user && user.expires) {                                                                                         // 619
    return true;                                                                                                      // 620
  } else {                                                                                                            //
    return false;                                                                                                     // 622
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
isSignedUp = function () {                                                                                            // 626
  // Returns true if the user has presented an invite key.                                                            //
                                                                                                                      //
  var user = Meteor.user();                                                                                           // 629
                                                                                                                      //
  if (!user) return false; // not signed in                                                                           // 631
                                                                                                                      //
  if (!user.loginIdentities) return false; // not an account                                                          // 633
                                                                                                                      //
  if (user.expires) return false; // demo user.                                                                       // 635
                                                                                                                      //
  if (Meteor.settings["public"].allowUninvited) return true; // all accounts qualify                                  // 637
                                                                                                                      //
  if (user.signupKey) return true; // user is invited                                                                 // 639
                                                                                                                      //
  return false;                                                                                                       // 641
};                                                                                                                    //
                                                                                                                      //
isSignedUpOrDemo = function () {                                                                                      // 644
  var user = Meteor.user();                                                                                           // 645
                                                                                                                      //
  if (!user) return false; // not signed in                                                                           // 647
                                                                                                                      //
  if (!user.loginIdentities) return false; // not an account                                                          // 649
                                                                                                                      //
  if (user.expires) return true; // demo user.                                                                        // 651
                                                                                                                      //
  if (Meteor.settings["public"].allowUninvited) return true; // all accounts qualify                                  // 653
                                                                                                                      //
  if (user.signupKey) return true; // user is invited                                                                 // 655
                                                                                                                      //
  return false;                                                                                                       // 657
};                                                                                                                    //
                                                                                                                      //
var countReferrals = function (user) {                                                                                // 660
  var referredIdentityIds = user.referredIdentityIds;                                                                 // 661
  return referredIdentityIds && referredIdentityIds.length || 0;                                                      // 662
};                                                                                                                    //
                                                                                                                      //
var calculateReferralBonus = function (user) {                                                                        // 665
  // This function returns an object of the form:                                                                     //
  //                                                                                                                  //
  // - {grains: 0, storage: 0}                                                                                        //
  //                                                                                                                  //
  // which are extra resources this account gets as part of participating in the referral                             //
  // program. (Storage is measured in bytes, as usual for plans.)                                                     //
                                                                                                                      //
  // TODO(cleanup): Consider moving referral bonus logic into Oasis payments module (since it's                       //
  //   payments-specific) and aggregating into `planBonus`.                                                           //
                                                                                                                      //
  // Authorization note: Only call this if accountId is the current user!                                             //
  var isPaid = user.plan && user.plan !== "free";                                                                     // 677
                                                                                                                      //
  successfulReferralsCount = countReferrals(user);                                                                    // 679
  if (isPaid) {                                                                                                       // 680
    var maxPaidStorageBonus = 30 * 1e9;                                                                               // 681
    return { grains: 0,                                                                                               // 682
      storage: Math.min(successfulReferralsCount * 2 * 1e9, maxPaidStorageBonus) };                                   // 683
  } else {                                                                                                            //
    var maxFreeStorageBonus = 2 * 1e9;                                                                                // 687
    var bonus = {                                                                                                     // 688
      storage: Math.min(successfulReferralsCount * 50 * 1e6, maxFreeStorageBonus)                                     // 689
    };                                                                                                                //
    if (successfulReferralsCount > 0) {                                                                               // 693
      bonus.grains = Infinity;                                                                                        // 694
    } else {                                                                                                          //
      bonus.grains = 0;                                                                                               // 696
    }                                                                                                                 //
                                                                                                                      //
    return bonus;                                                                                                     // 699
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
getUserQuota = function (user) {                                                                                      // 703
  var plan = Plans.findOne(user.plan || "free");                                                                      // 704
  var referralBonus = calculateReferralBonus(user);                                                                   // 705
  var bonus = user.planBonus || {};                                                                                   // 706
  var userQuota = {                                                                                                   // 707
    storage: plan.storage + referralBonus.storage + (bonus.storage || 0),                                             // 708
    grains: plan.grains + referralBonus.grains + (bonus.grains || 0),                                                 // 709
    compute: plan.compute + (bonus.compute || 0)                                                                      // 710
  };                                                                                                                  //
  return userQuota;                                                                                                   // 712
};                                                                                                                    //
                                                                                                                      //
isUserOverQuota = function (user) {                                                                                   // 715
  // Return false if user has quota space remaining, true if it is full. When this returns true,                      //
  // we will not allow the user to create new grains, though they may be able to open existing ones                   //
  // which may still increase their storage usage.                                                                    //
  //                                                                                                                  //
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                                 //
                                                                                                                      //
  if (!Meteor.settings["public"].quotaEnabled || user.isAdmin) return false;                                          // 722
                                                                                                                      //
  var plan = getUserQuota(user);                                                                                      // 724
  if (plan.grains < Infinity) {                                                                                       // 725
    var count = Grains.find({ userId: user._id }, { fields: {}, limit: plan.grains }).count();                        // 726
    if (count >= plan.grains) return "outOfGrains";                                                                   // 727
  }                                                                                                                   //
                                                                                                                      //
  return plan && user.storageUsage && user.storageUsage >= plan.storage && "outOfStorage";                            // 730
};                                                                                                                    //
                                                                                                                      //
isUserExcessivelyOverQuota = function (user) {                                                                        // 733
  // Return true if user is so far over quota that we should prevent their existing grains from                       //
  // running at all.                                                                                                  //
  //                                                                                                                  //
  // (Actually returns a string which can be fed into `billingPrompt` as the reason.)                                 //
                                                                                                                      //
  if (!Meteor.settings["public"].quotaEnabled || user.isAdmin) return false;                                          // 739
                                                                                                                      //
  var quota = getUserQuota(user);                                                                                     // 741
                                                                                                                      //
  // quota.grains = Infinity means unlimited grains. IEEE754 defines Infinity == Infinity.                            //
  if (quota.grains < Infinity) {                                                                                      // 744
    var count = Grains.find({ userId: user._id }, { fields: {}, limit: quota.grains * 2 }).count();                   // 745
    if (count >= quota.grains * 2) return "outOfGrains";                                                              // 746
  }                                                                                                                   //
                                                                                                                      //
  return quota && user.storageUsage && user.storageUsage >= quota.storage * 1.2 && "outOfStorage";                    // 749
};                                                                                                                    //
                                                                                                                      //
isAdmin = function () {                                                                                               // 752
  // Returns true if the user is the administrator.                                                                   //
                                                                                                                      //
  var user = Meteor.user();                                                                                           // 755
  if (user && user.isAdmin) {                                                                                         // 756
    return true;                                                                                                      // 757
  } else {                                                                                                            //
    return false;                                                                                                     // 759
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
isAdminById = function (id) {                                                                                         // 763
  // Returns true if the user's id is the administrator.                                                              //
                                                                                                                      //
  var user = Meteor.users.findOne({ _id: id }, { fields: { isAdmin: 1 } });                                           // 766
  if (user && user.isAdmin) {                                                                                         // 767
    return true;                                                                                                      // 768
  } else {                                                                                                            //
    return false;                                                                                                     // 770
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
findAdminUserForToken = function (token) {                                                                            // 774
  if (!token.requirements) {                                                                                          // 775
    return;                                                                                                           // 776
  }                                                                                                                   //
                                                                                                                      //
  var requirements = token.requirements.filter(function (requirement) {                                               // 779
    return "userIsAdmin" in requirement;                                                                              // 780
  });                                                                                                                 //
                                                                                                                      //
  if (requirements.length > 1) {                                                                                      // 783
    return;                                                                                                           // 784
  }                                                                                                                   //
                                                                                                                      //
  if (requirements.length === 0) {                                                                                    // 787
    return;                                                                                                           // 788
  }                                                                                                                   //
                                                                                                                      //
  return requirements[0].userIsAdmin;                                                                                 // 791
};                                                                                                                    //
                                                                                                                      //
var wildcardHost = Meteor.settings["public"].wildcardHost.toLowerCase().split("*");                                   // 794
                                                                                                                      //
if (wildcardHost.length != 2) {                                                                                       // 796
  throw new Error("Wildcard host must contain exactly one asterisk.");                                                // 797
}                                                                                                                     //
                                                                                                                      //
matchWildcardHost = function (host) {                                                                                 // 800
  // See if the hostname is a member of our wildcard. If so, extract the ID.                                          //
                                                                                                                      //
  // We remove everything after the first ":" character so that our                                                   //
  // comparison logic ignores port numbers.                                                                           //
  var prefix = wildcardHost[0];                                                                                       // 805
  var suffix = wildcardHost[1].split(":")[0];                                                                         // 806
  var hostSansPort = host.split(":")[0];                                                                              // 807
                                                                                                                      //
  if (hostSansPort.lastIndexOf(prefix, 0) >= 0 && hostSansPort.indexOf(suffix, -suffix.length) >= 0 && hostSansPort.length >= prefix.length + suffix.length) {
    var id = hostSansPort.slice(prefix.length, -suffix.length);                                                       // 812
    if (id.match(/^[-a-z0-9]*$/)) {                                                                                   // 813
      return id;                                                                                                      // 814
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  return null;                                                                                                        // 818
};                                                                                                                    //
                                                                                                                      //
makeWildcardHost = function (id) {                                                                                    // 821
  return wildcardHost[0] + id + wildcardHost[1];                                                                      // 822
};                                                                                                                    //
                                                                                                                      //
var isApiHostId = function (hostId) {                                                                                 // 825
  if (hostId) {                                                                                                       // 826
    var split = hostId.split("-");                                                                                    // 827
    if (split[0] === "api") return split[1] || "*";                                                                   // 828
  }                                                                                                                   //
                                                                                                                      //
  return false;                                                                                                       // 831
};                                                                                                                    //
                                                                                                                      //
var isTokenSpecificHostId = function (hostId) {                                                                       // 834
  return hostId.lastIndexOf("api-", 0) === 0;                                                                         // 835
};                                                                                                                    //
                                                                                                                      //
var apiHostIdHashForToken = undefined;                                                                                // 838
if (Meteor.isServer) {                                                                                                // 839
  (function () {                                                                                                      //
    var Crypto = Npm.require("crypto");                                                                               // 840
    apiHostIdHashForToken = function (token) {                                                                        // 841
      // Given an API token, compute the host ID that must be used when requesting this token.                        //
                                                                                                                      //
      // We add a leading 'x' to the hash so that knowing the hostname alone is not sufficient to                     //
      // find the corresponding API token in the ApiTokens table (whose _id values are also hashes                    //
      // of tokens). This doesn't technically add any security, but helps prove that we don't have                    //
      // any bugs which would allow someone who knows only the hostname to access the app API.                        //
      return Crypto.createHash("sha256").update("x" + token).digest("hex").slice(0, 32);                              // 848
    };                                                                                                                //
  })();                                                                                                               //
} else {                                                                                                              //
  apiHostIdHashForToken = function (token) {                                                                          // 851
    // Given an API token, compute the host ID that must be used when requesting this token.                          //
                                                                                                                      //
    // We add a leading 'x' to the hash so that knowing the hostname alone is not sufficient to                       //
    // find the corresponding API token in the ApiTokens table (whose _id values are also hashes                      //
    // of tokens). This doesn't technically add any security, but helps prove that we don't have                      //
    // any bugs which would allow someone who knows only the hostname to access the app API.                          //
    return SHA256("x" + token).slice(0, 32);                                                                          // 858
  };                                                                                                                  //
}                                                                                                                     //
                                                                                                                      //
var apiHostIdForToken = function (token) {                                                                            // 862
  return "api-" + apiHostIdHashForToken(token);                                                                       // 863
};                                                                                                                    //
                                                                                                                      //
var makeApiHost = function (token) {                                                                                  // 866
  return makeWildcardHost(apiHostIdForToken(token));                                                                  // 867
};                                                                                                                    //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 870
  (function () {                                                                                                      //
    var Url = Npm.require("url");                                                                                     // 871
    getWildcardOrigin = function () {                                                                                 // 872
      // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're                     //
      // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),                   //
      // an astrisk is only allowed as the first character and must be followed by a period. So we need               //
      // "*.example.com" instead -- which matches more than we actually want, but is the best we can                  //
      // really do. We also add the protocol to the front (again, that's what CSP wants).                             //
                                                                                                                      //
      // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.                //
      var protocol = Url.parse(process.env.ROOT_URL).protocol;                                                        // 880
                                                                                                                      //
      var dotPos = wildcardHost[1].indexOf(".");                                                                      // 882
      if (dotPos < 0) {                                                                                               // 883
        return protocol + "//*";                                                                                      // 884
      } else {                                                                                                        //
        return protocol + "//*" + wildcardHost[1].slice(dotPos);                                                      // 886
      }                                                                                                               //
    };                                                                                                                //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
allowDevAccounts = function () {                                                                                      // 891
  var setting = Settings.findOne({ _id: "devAccounts" });                                                             // 892
  if (setting) {                                                                                                      // 893
    return setting.value;                                                                                             // 894
  } else {                                                                                                            //
    return Meteor.settings && Meteor.settings["public"] && Meteor.settings["public"].allowDevAccounts;                // 896
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
sendReferralProgramNotification = function (userId) {                                                                 // 901
  Notifications.upsert({                                                                                              // 902
    userId: userId,                                                                                                   // 903
    referral: true                                                                                                    // 904
  }, {                                                                                                                //
    userId: userId,                                                                                                   // 906
    referral: true,                                                                                                   // 907
    timestamp: new Date(),                                                                                            // 908
    isUnread: true                                                                                                    // 909
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
roleAssignmentPattern = {                                                                                             // 913
  none: Match.Optional(null),                                                                                         // 914
  allAccess: Match.Optional(null),                                                                                    // 915
  roleId: Match.Optional(Match.Integer),                                                                              // 916
  addPermissions: Match.Optional([Boolean]),                                                                          // 917
  removePermissions: Match.Optional([Boolean])                                                                        // 918
};                                                                                                                    //
                                                                                                                      //
SandstormDb = function () {                                                                                           // 921
  this.collections = {                                                                                                // 922
    // Direct access to underlying collections. DEPRECATED.                                                           //
    //                                                                                                                //
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove                     //
    //   direct access to the collections.                                                                            //
                                                                                                                      //
    packages: Packages,                                                                                               // 928
    devPackages: DevPackages,                                                                                         // 929
    userActions: UserActions,                                                                                         // 930
    grains: Grains,                                                                                                   // 931
    contacts: Contacts,                                                                                               // 932
    sessions: Sessions,                                                                                               // 933
    signupKeys: SignupKeys,                                                                                           // 934
    activityStats: ActivityStats,                                                                                     // 935
    deleteStats: DeleteStats,                                                                                         // 936
    fileTokens: FileTokens,                                                                                           // 937
    apiTokens: ApiTokens,                                                                                             // 938
    apiHosts: ApiHosts,                                                                                               // 939
    notifications: Notifications,                                                                                     // 940
    statsTokens: StatsTokens,                                                                                         // 941
    misc: Misc,                                                                                                       // 942
    settings: Settings,                                                                                               // 943
    appIndex: AppIndex,                                                                                               // 944
    keybaseProfiles: KeybaseProfiles                                                                                  // 945
                                                                                                                      //
  };                                                                                                                  //
};                                                                                                                    //
                                                                                                                      //
// TODO(cleanup): These methods should not be defined freestanding and should use collection                          //
//   objects created in SandstormDb's constructor rather than globals.                                                //
                                                                                                                      //
// Intentionally omitted:                                                                                             //
// - Migrations, since it's used only within this package.                                                            //
// - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.                       //
_.extend(SandstormDb.prototype, {                                                                                     // 956
  isDemoUser: isDemoUser,                                                                                             // 957
  isSignedUp: isSignedUp,                                                                                             // 958
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                                 // 959
  getUserQuota: getUserQuota,                                                                                         // 960
  isUserOverQuota: isUserOverQuota,                                                                                   // 961
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                                             // 962
  isAdmin: isAdmin,                                                                                                   // 963
  isAdminById: isAdminById,                                                                                           // 964
  findAdminUserForToken: findAdminUserForToken,                                                                       // 965
  matchWildcardHost: matchWildcardHost,                                                                               // 966
  makeWildcardHost: makeWildcardHost,                                                                                 // 967
  isApiHostId: isApiHostId,                                                                                           // 968
  isTokenSpecificHostId: isTokenSpecificHostId,                                                                       // 969
  apiHostIdHashForToken: apiHostIdHashForToken,                                                                       // 970
  apiHostIdForToken: apiHostIdForToken,                                                                               // 971
  makeApiHost: makeApiHost,                                                                                           // 972
  allowDevAccounts: allowDevAccounts,                                                                                 // 973
  roleAssignmentPattern: roleAssignmentPattern                                                                        // 974
});                                                                                                                   //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 977
  (function () {                                                                                                      //
    SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                                      // 978
                                                                                                                      //
    var Crypto = Npm.require("crypto");                                                                               // 980
    SandstormDb.prototype.removeApiTokens = function (query) {                                                        // 981
      // Remove all API tokens matching the query, making sure to clean up ApiHosts as well.                          //
                                                                                                                      //
      this.collections.apiTokens.find(query).forEach(function (token) {                                               // 984
        // Clean up ApiHosts for webkey tokens.                                                                       //
        if (token.hasApiHost) {                                                                                       // 986
          var hash2 = Crypto.createHash("sha256").update(token._id).digest("base64");                                 // 987
          ApiHosts.remove({ hash2: hash2 });                                                                          // 988
        }                                                                                                             //
      });                                                                                                             //
                                                                                                                      //
      this.collections.apiTokens.remove(query);                                                                       // 992
    };                                                                                                                //
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
// =======================================================================================                            //
// Below this point are newly-written or refactored functions.                                                        //
                                                                                                                      //
_.extend(SandstormDb.prototype, {                                                                                     // 999
  getUser: (function () {                                                                                             // 1000
    function getUser(userId) {                                                                                        // 1000
      check(userId, Match.OneOf(String, undefined, null));                                                            // 1001
      if (userId) {                                                                                                   // 1002
        return Meteor.users.findOne(userId);                                                                          // 1003
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return getUser;                                                                                                   //
  })(),                                                                                                               //
                                                                                                                      //
  getIdentity: (function () {                                                                                         // 1007
    function getIdentity(identityId) {                                                                                // 1007
      check(identityId, String);                                                                                      // 1008
      var identity = Meteor.users.findOne({ _id: identityId });                                                       // 1009
      if (identity) {                                                                                                 // 1010
        SandstormDb.fillInProfileDefaults(identity);                                                                  // 1011
        SandstormDb.fillInIntrinsicName(identity);                                                                    // 1012
        SandstormDb.fillInPictureUrl(identity);                                                                       // 1013
        return identity;                                                                                              // 1014
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return getIdentity;                                                                                               //
  })(),                                                                                                               //
                                                                                                                      //
  userHasIdentity: function (userId, identityId) {                                                                    // 1018
    check(userId, String);                                                                                            // 1019
    check(identityId, String);                                                                                        // 1020
                                                                                                                      //
    if (userId === identityId) return true;                                                                           // 1022
                                                                                                                      //
    var user = Meteor.users.findOne(userId);                                                                          // 1024
    return SandstormDb.getUserIdentityIds(user).indexOf(identityId) != -1;                                            // 1025
  },                                                                                                                  //
                                                                                                                      //
  userGrains: (function () {                                                                                          // 1028
    function userGrains(userId) {                                                                                     // 1028
      check(userId, Match.OneOf(String, undefined, null));                                                            // 1029
      return this.collections.grains.find({ userId: userId });                                                        // 1030
    }                                                                                                                 //
                                                                                                                      //
    return userGrains;                                                                                                //
  })(),                                                                                                               //
                                                                                                                      //
  currentUserGrains: (function () {                                                                                   // 1033
    function currentUserGrains() {                                                                                    // 1033
      return this.userGrains(Meteor.userId());                                                                        // 1034
    }                                                                                                                 //
                                                                                                                      //
    return currentUserGrains;                                                                                         //
  })(),                                                                                                               //
                                                                                                                      //
  getGrain: (function () {                                                                                            // 1037
    function getGrain(grainId) {                                                                                      // 1037
      check(grainId, String);                                                                                         // 1038
      return this.collections.grains.findOne(grainId);                                                                // 1039
    }                                                                                                                 //
                                                                                                                      //
    return getGrain;                                                                                                  //
  })(),                                                                                                               //
                                                                                                                      //
  userApiTokens: (function () {                                                                                       // 1042
    function userApiTokens(userId) {                                                                                  // 1042
      check(userId, Match.OneOf(String, undefined, null));                                                            // 1043
      var identityIds = SandstormDb.getUserIdentityIds(this.getUser(userId));                                         // 1044
      return this.collections.apiTokens.find({ "owner.user.identityId": { $in: identityIds } });                      // 1045
    }                                                                                                                 //
                                                                                                                      //
    return userApiTokens;                                                                                             //
  })(),                                                                                                               //
                                                                                                                      //
  currentUserApiTokens: (function () {                                                                                // 1048
    function currentUserApiTokens() {                                                                                 // 1048
      return this.userApiTokens(Meteor.userId());                                                                     // 1049
    }                                                                                                                 //
                                                                                                                      //
    return currentUserApiTokens;                                                                                      //
  })(),                                                                                                               //
                                                                                                                      //
  userActions: (function () {                                                                                         // 1052
    function userActions(user) {                                                                                      // 1052
      return this.collections.userActions.find({ userId: user });                                                     // 1053
    }                                                                                                                 //
                                                                                                                      //
    return userActions;                                                                                               //
  })(),                                                                                                               //
                                                                                                                      //
  currentUserActions: (function () {                                                                                  // 1056
    function currentUserActions() {                                                                                   // 1056
      return this.userActions(Meteor.userId());                                                                       // 1057
    }                                                                                                                 //
                                                                                                                      //
    return currentUserActions;                                                                                        //
  })(),                                                                                                               //
                                                                                                                      //
  iconSrcForPackage: (function () {                                                                                   // 1060
    function iconSrcForPackage(pkg, usage) {                                                                          // 1060
      return Identicon.iconSrcForPackage(pkg, usage, this.makeWildcardHost("static"));                                // 1061
    }                                                                                                                 //
                                                                                                                      //
    return iconSrcForPackage;                                                                                         //
  })(),                                                                                                               //
                                                                                                                      //
  getDenormalizedGrainInfo: (function () {                                                                            // 1064
    function getDenormalizedGrainInfo(grainId) {                                                                      // 1064
      var grain = this.getGrain(grainId);                                                                             // 1065
      var pkg = this.collections.packages.findOne(grain.packageId);                                                   // 1066
      var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };                             // 1067
      var grainInfo = { appTitle: appTitle };                                                                         // 1068
                                                                                                                      //
      if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {                              // 1070
        var icons = pkg.manifest.metadata.icons;                                                                      // 1071
        grainInfo.icon = icons.grain || icons.appGrid;                                                                // 1072
      }                                                                                                               //
                                                                                                                      //
      // Only provide an app ID if we have no icon asset to provide and need to offer an identicon.                   //
      if (!grainInfo.icon && pkg) {                                                                                   // 1076
        grainInfo.appId = pkg.appId;                                                                                  // 1077
      }                                                                                                               //
                                                                                                                      //
      return grainInfo;                                                                                               // 1080
    }                                                                                                                 //
                                                                                                                      //
    return getDenormalizedGrainInfo;                                                                                  //
  })(),                                                                                                               //
                                                                                                                      //
  getPlan: function (id) {                                                                                            // 1083
    check(id, String);                                                                                                // 1084
    var plan = Plans.findOne(id);                                                                                     // 1085
    if (!plan) {                                                                                                      // 1086
      throw new Error("no such plan: " + id);                                                                         // 1087
    }                                                                                                                 //
                                                                                                                      //
    return plan;                                                                                                      // 1090
  },                                                                                                                  //
                                                                                                                      //
  listPlans: function () {                                                                                            // 1093
    return Plans.find({}, { sort: { price: 1 } });                                                                    // 1094
  },                                                                                                                  //
                                                                                                                      //
  getMyPlan: function () {                                                                                            // 1097
    var user = Meteor.user();                                                                                         // 1098
    return user && Plans.findOne(user.plan || "free");                                                                // 1099
  },                                                                                                                  //
                                                                                                                      //
  getMyReferralBonus: function (user) {                                                                               // 1102
    // This function is called from the server and from the client, similar to getMyPlan().                           //
    //                                                                                                                //
    // The parameter may be omitted in which case the current user is assumed.                                        //
                                                                                                                      //
    return calculateReferralBonus(user || Meteor.user());                                                             // 1107
  },                                                                                                                  //
                                                                                                                      //
  getMyUsage: function (user) {                                                                                       // 1110
    user = user || Meteor.user();                                                                                     // 1111
    if (user && (Meteor.isServer || user.pseudoUsage)) {                                                              // 1112
      if (Meteor.isClient) {                                                                                          // 1113
        // Filled by pseudo-subscription to "getMyUsage". WARNING: The subscription is currenly                       //
        // not reactive.                                                                                              //
        return user.pseudoUsage;                                                                                      // 1116
      } else {                                                                                                        //
        return {                                                                                                      // 1118
          grains: Grains.find({ userId: user._id }).count(),                                                          // 1119
          storage: user.storageUsage || 0,                                                                            // 1120
          compute: 0 };                                                                                               // 1121
      }                                                                                                               //
    } else // not tracked yet                                                                                         //
      {                                                                                                               // 1124
        return { grains: 0, storage: 0, compute: 0 };                                                                 // 1125
      }                                                                                                               //
  },                                                                                                                  //
                                                                                                                      //
  isUninvitedFreeUser: function () {                                                                                  // 1129
    if (!Meteor.settings["public"].allowUninvited) return false;                                                      // 1130
                                                                                                                      //
    var user = Meteor.user();                                                                                         // 1132
    return user && !user.expires && (!user.plan || user.plan === "free");                                             // 1133
  },                                                                                                                  //
                                                                                                                      //
  getSetting: function (name) {                                                                                       // 1136
    var setting = Settings.findOne(name);                                                                             // 1137
    return setting && setting.value;                                                                                  // 1138
  },                                                                                                                  //
                                                                                                                      //
  addUserActions: function (packageId) {                                                                              // 1141
    //TODO(cleanup): implement this with meteor methods rather than client-side inserts/removes.                      //
    var pack = Packages.findOne(packageId);                                                                           // 1143
    if (pack) {                                                                                                       // 1144
      // Remove old versions.                                                                                         //
      UserActions.find({ userId: Meteor.userId(), appId: pack.appId }).forEach(function (action) {                    // 1146
        UserActions.remove(action._id);                                                                               // 1148
      });                                                                                                             //
                                                                                                                      //
      // Install new.                                                                                                 //
      var actions = pack.manifest.actions;                                                                            // 1152
      for (var i in babelHelpers.sanitizeForInObject(actions)) {                                                      // 1153
        var action = actions[i];                                                                                      // 1154
        if ("none" in action.input) {                                                                                 // 1155
          var userAction = {                                                                                          // 1156
            userId: Meteor.userId(),                                                                                  // 1157
            packageId: pack._id,                                                                                      // 1158
            appId: pack.appId,                                                                                        // 1159
            appTitle: pack.manifest.appTitle,                                                                         // 1160
            appMarketingVersion: pack.manifest.appMarketingVersion,                                                   // 1161
            appVersion: pack.manifest.appVersion,                                                                     // 1162
            title: action.title,                                                                                      // 1163
            nounPhrase: action.nounPhrase,                                                                            // 1164
            command: action.command                                                                                   // 1165
          };                                                                                                          //
          UserActions.insert(userAction);                                                                             // 1167
        } else {}                                                                                                     //
        // TODO(someday):  Implement actions with capability inputs.                                                  //
        //jscs:ignore disallowEmptyBlocks                                                                             //
      }                                                                                                               // 1170
                                                                                                                      //
      Meteor.call("deleteUnusedPackages", pack.appId);                                                                // 1173
    }                                                                                                                 //
  },                                                                                                                  //
                                                                                                                      //
  sendAdminNotification: function (message, link) {                                                                   // 1177
    Meteor.users.find({ isAdmin: true }, { fields: { _id: 1 } }).forEach(function (user) {                            // 1178
      Notifications.insert({                                                                                          // 1179
        admin: {                                                                                                      // 1180
          action: link,                                                                                               // 1181
          type: "reportStats"                                                                                         // 1182
        },                                                                                                            //
        userId: user._id,                                                                                             // 1184
        text: { defaultText: message },                                                                               // 1185
        timestamp: new Date(),                                                                                        // 1186
        isUnread: true                                                                                                // 1187
      });                                                                                                             //
    });                                                                                                               //
  },                                                                                                                  //
                                                                                                                      //
  getKeybaseProfile: function (keyFingerprint) {                                                                      // 1192
    return this.collections.keybaseProfiles.findOne(keyFingerprint) || {};                                            // 1193
  },                                                                                                                  //
                                                                                                                      //
  getServerTitle: function () {                                                                                       // 1196
    var setting = Settings.findOne({ _id: "serverTitle" });                                                           // 1197
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1198
  },                                                                                                                  //
                                                                                                                      //
  getReturnAddress: function () {                                                                                     // 1201
    var setting = Settings.findOne({ _id: "returnAddress" });                                                         // 1202
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1203
  },                                                                                                                  //
                                                                                                                      //
  isFeatureKeyValid: function () {                                                                                    // 1206
    return Meteor.settings["public"].isFeatureKeyValid;                                                               // 1207
  },                                                                                                                  //
                                                                                                                      //
  getLdapUrl: function () {                                                                                           // 1210
    var setting = Settings.findOne({ _id: "ldapUrl" });                                                               // 1211
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1212
  },                                                                                                                  //
                                                                                                                      //
  getLdapBase: function () {                                                                                          // 1215
    var setting = Settings.findOne({ _id: "ldapBase" });                                                              // 1216
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1217
  },                                                                                                                  //
                                                                                                                      //
  getLdapDnPattern: function () {                                                                                     // 1220
    var setting = Settings.findOne({ _id: "ldapDnPattern" });                                                         // 1221
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1222
  },                                                                                                                  //
                                                                                                                      //
  getLdapNameField: function () {                                                                                     // 1225
    var setting = Settings.findOne({ _id: "ldapNameField" });                                                         // 1226
    return setting ? setting.value : ""; // empty if subscription is not ready.                                       // 1227
  }                                                                                                                   //
});                                                                                                                   //
                                                                                                                      //
SandstormDb.escapeMongoKey = function (key) {                                                                         // 1231
  // This incredibly poor mechanism for escaping Mongo keys is recommended by the Mongo docs here:                    //
  //   https://docs.mongodb.org/manual/faq/developers/#dollar-sign-operator-escaping                                  //
  // and seems to be a de facto standard, for example:                                                                //
  //   https://www.npmjs.com/package/mongo-key-escape                                                                 //
  return key.replace(".", "．").replace("$", "＄");                                                                     // 1236
};                                                                                                                    //
                                                                                                                      //
var appNameFromPackage = function (packageObj) {                                                                      // 1239
  // This function takes a Package object from Mongo and returns an                                                   //
  // app title.                                                                                                       //
  var manifest = packageObj.manifest;                                                                                 // 1242
  if (!manifest) return packageObj.appId || packageObj._id || "unknown";                                              // 1243
  var action = manifest.actions[0];                                                                                   // 1244
  appName = manifest.appTitle && manifest.appTitle.defaultText || appNameFromActionName(action.title.defaultText);    // 1245
  return appName;                                                                                                     // 1247
};                                                                                                                    //
                                                                                                                      //
var appNameFromActionName = function (name) {                                                                         // 1250
  // Hack: Historically we only had action titles, like "New Etherpad Document", not app                              //
  //   titles. But for this UI we want app titles. As a transitionary measure, try to                                 //
  //   derive the app title from the action title.                                                                    //
  // TODO(cleanup): Get rid of this once apps have real titles.                                                       //
  if (!name) {                                                                                                        // 1255
    return "(unnamed)";                                                                                               // 1256
  }                                                                                                                   //
                                                                                                                      //
  if (name.lastIndexOf("New ", 0) === 0) {                                                                            // 1259
    name = name.slice(4);                                                                                             // 1260
  }                                                                                                                   //
                                                                                                                      //
  if (name.lastIndexOf("Hacker CMS", 0) === 0) {                                                                      // 1263
    name = "Hacker CMS";                                                                                              // 1264
  } else {                                                                                                            //
    var space = name.indexOf(" ");                                                                                    // 1266
    if (space > 0) {                                                                                                  // 1267
      name = name.slice(0, space);                                                                                    // 1268
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  return name;                                                                                                        // 1272
};                                                                                                                    //
                                                                                                                      //
var appShortDescriptionFromPackage = function (pkg) {                                                                 // 1275
  return pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.shortDescription && pkg.manifest.metadata.shortDescription.defaultText;
};                                                                                                                    //
                                                                                                                      //
var nounPhraseForActionAndAppTitle = function (action, appTitle) {                                                    // 1281
  // A hack to deal with legacy apps not including fields in their manifests.                                         //
  // I look forward to the day I can remove most of this code.                                                        //
  // Attempt to figure out the appropriate noun that this action will create.                                         //
  // Use an explicit noun phrase is one is available.  Apps should add these in the future.                           //
  if (action.nounPhrase) return action.nounPhrase.defaultText;                                                        // 1286
  // Otherwise, try to guess one from the structure of the action title field                                         //
  if (action.title && action.title.defaultText) {                                                                     // 1288
    var text = action.title.defaultText;                                                                              // 1289
    // Strip a leading "New "                                                                                         //
    if (text.lastIndexOf("New ", 0) === 0) {                                                                          // 1291
      var candidate = text.slice(4);                                                                                  // 1292
      // Strip a leading appname too, if provided                                                                     //
      if (candidate.lastIndexOf(appTitle, 0) === 0) {                                                                 // 1294
        var newCandidate = candidate.slice(appTitle.length);                                                          // 1295
        // Unless that leaves you with no noun, in which case, use "grain"                                            //
        if (newCandidate.length > 0) {                                                                                // 1297
          return newCandidate.toLowerCase();                                                                          // 1298
        } else {                                                                                                      //
          return "grain";                                                                                             // 1300
        }                                                                                                             //
      }                                                                                                               //
                                                                                                                      //
      return candidate.toLowerCase();                                                                                 // 1304
    }                                                                                                                 //
    // Some other verb phrase was given.  Just use it verbatim, and hope the app author updates                       //
    // the package soon.                                                                                              //
    return text;                                                                                                      // 1308
  } else {                                                                                                            //
    return "grain";                                                                                                   // 1310
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
// Static methods on SandstormDb that don't need an instance.                                                         //
// Largely things that deal with backwards-compatibility.                                                             //
_.extend(SandstormDb, {                                                                                               // 1316
  appNameFromActionName: appNameFromActionName,                                                                       // 1317
  appNameFromPackage: appNameFromPackage,                                                                             // 1318
  appShortDescriptionFromPackage: appShortDescriptionFromPackage,                                                     // 1319
  nounPhraseForActionAndAppTitle: nounPhraseForActionAndAppTitle                                                      // 1320
});                                                                                                                   //
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 1323
  (function () {                                                                                                      //
    var Crypto = Npm.require("crypto");                                                                               // 1324
    var ContentType = Npm.require("content-type");                                                                    // 1325
    var Zlib = Npm.require("zlib");                                                                                   // 1326
                                                                                                                      //
    var replicaNumber = Meteor.settings.replicaNumber || 0;                                                           // 1328
                                                                                                                      //
    var computeStagger = function (n) {                                                                               // 1330
      // Compute a fraction in the range [0, 1) such that, for any natural number k, the values                       //
      // of computeStagger(n) for all n in [1, 2^k) are uniformly distributed between 0 and 1.                        //
      // The sequence looks like:                                                                                     //
      //   0, 1/2, 1/4, 3/4, 1/8, 3/8, 5/8, 7/8, 1/16, ...                                                            //
      //                                                                                                              //
      // We use this to determine how we'll stagger periodic events performed by this replica.                        //
      // Notice that this allows us to compute a stagger which is independent of the number of                        //
      // front-end replicas present; we can add more replicas to the end without affecting how the                    //
      // earlier ones schedule their events.                                                                          //
      var denom = 1;                                                                                                  // 1340
      while (denom <= n) denom <<= 1;                                                                                 // 1341
      var num = n * 2 - denom + 1;                                                                                    // 1342
      return num / denom;                                                                                             // 1343
    };                                                                                                                //
                                                                                                                      //
    var stagger = computeStagger(replicaNumber);                                                                      // 1346
                                                                                                                      //
    SandstormDb.periodicCleanup = function (intervalMs, callback) {                                                   // 1348
      // Register a database cleanup function than should run periodically, roughly once every                        //
      // interval of the given length.                                                                                //
      //                                                                                                              //
      // In a blackrock deployment with multiple front-ends, the frequency of the cleanup will be                     //
      // scaled appropriately on the assumption that more data is being generated demanding more                      //
      // frequent cleanups.                                                                                           //
                                                                                                                      //
      check(intervalMs, Number);                                                                                      // 1356
      check(callback, Function);                                                                                      // 1357
                                                                                                                      //
      if (intervalMs < 120000) {                                                                                      // 1359
        throw new Error("less than 2-minute cleanup interval seems too fast; " + "are you using the right units?");   // 1360
      }                                                                                                               //
                                                                                                                      //
      // Schedule first cleanup to happen at the next intervalMs interval from the epoch, so that                     //
      // the schedule is independent of the exact startup time.                                                       //
      var first = intervalMs - Date.now() % intervalMs;                                                               // 1366
                                                                                                                      //
      // Stagger cleanups across replicas so that we don't have all replicas trying to clean the                      //
      // same data at the same time.                                                                                  //
      first += Math.floor(intervalMs * computeStagger(replicaNumber));                                                // 1370
                                                                                                                      //
      // If the stagger put us more than an interval away from now, back up.                                          //
      if (first > intervalMs) first -= intervalMs;                                                                    // 1373
                                                                                                                      //
      Meteor.setTimeout(function () {                                                                                 // 1375
        callback();                                                                                                   // 1376
        Meteor.setInterval(callback, intervalMs);                                                                     // 1377
      }, first);                                                                                                      //
    };                                                                                                                //
                                                                                                                      //
    // TODO(cleanup): Node 0.12 has a `gzipSync` but 0.10 (which Meteor still uses) does not.                         //
    var gzipSync = Meteor.wrapAsync(Zlib.gzip, Zlib);                                                                 // 1382
                                                                                                                      //
    var BufferSmallerThan = function (limit) {                                                                        // 1384
      return Match.Where(function (buf) {                                                                             // 1385
        check(buf, Buffer);                                                                                           // 1386
        return buf.length < limit;                                                                                    // 1387
      });                                                                                                             //
    };                                                                                                                //
                                                                                                                      //
    var DatabaseId = Match.Where(function (s) {                                                                       // 1391
      check(s, String);                                                                                               // 1392
      return !!s.match(/^[a-zA-Z0-9_]+$/);                                                                            // 1393
    });                                                                                                               //
                                                                                                                      //
    addStaticAsset = function (metadata, content) {                                                                   // 1396
      // Add a new static asset to the database. If `content` is a string rather than a buffer, it                    //
      // will be automatically gzipped before storage; do not specify metadata.encoding in this case.                 //
                                                                                                                      //
      if (typeof content === "string" && !metadata.encoding) {                                                        // 1400
        content = gzipSync(new Buffer(content, "utf8"));                                                              // 1401
        metadata.encoding = "gzip";                                                                                   // 1402
      }                                                                                                               //
                                                                                                                      //
      check(metadata, {                                                                                               // 1405
        mimeType: String,                                                                                             // 1406
        encoding: Match.Optional("gzip")                                                                              // 1407
      });                                                                                                             //
      check(content, BufferSmallerThan(1 << 20));                                                                     // 1409
                                                                                                                      //
      // Validate content type.                                                                                       //
      metadata.mimeType = ContentType.format(ContentType.parse(metadata.mimeType));                                   // 1412
                                                                                                                      //
      var hasher = Crypto.createHash("sha256");                                                                       // 1414
      hasher.update(metadata.mimeType + "\n" + metadata.encoding + "\n", "utf8");                                     // 1415
      hasher.update(content);                                                                                         // 1416
      var hash = hasher.digest("base64");                                                                             // 1417
                                                                                                                      //
      var existing = StaticAssets.findAndModify({                                                                     // 1419
        query: { hash: hash, refcount: { $gte: 1 } },                                                                 // 1420
        update: { $inc: { refcount: 1 } },                                                                            // 1421
        fields: { _id: 1, refcount: 1 }                                                                               // 1422
      });                                                                                                             //
      if (existing) {                                                                                                 // 1424
        return existing._id;                                                                                          // 1425
      }                                                                                                               //
                                                                                                                      //
      return StaticAssets.insert(_.extend({                                                                           // 1428
        hash: hash,                                                                                                   // 1429
        content: content,                                                                                             // 1430
        refcount: 1                                                                                                   // 1431
      }, metadata));                                                                                                  //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.addStaticAsset = addStaticAsset;                                                            // 1435
                                                                                                                      //
    SandstormDb.prototype.refStaticAsset = function (id) {                                                            // 1437
      // Increment the refcount on an existing static asset.                                                          //
      //                                                                                                              //
      // You must call this BEFORE adding the new reference to the DB, in case of failure between                     //
      // the two calls. (This way, the failure case is a storage leak, which is probably not a big                    //
      // deal and can be fixed by GC, rather than a mysteriously missing asset.)                                      //
                                                                                                                      //
      check(id, String);                                                                                              // 1444
                                                                                                                      //
      var existing = StaticAssets.findAndModify({                                                                     // 1446
        query: { hash: hash },                                                                                        // 1447
        update: { $inc: { refcount: 1 } },                                                                            // 1448
        fields: { _id: 1, refcount: 1 }                                                                               // 1449
      });                                                                                                             //
      if (!existing) {                                                                                                // 1451
        throw new Error("refStaticAsset() called on asset that doesn't exist");                                       // 1452
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.unrefStaticAsset = function (id) {                                                          // 1456
      // Decrement refcount on a static asset and delete if it has reached zero.                                      //
      //                                                                                                              //
      // You must call this AFTER removing the reference from the DB, in case of failure between                      //
      // the two calls. (This way, the failure case is a storage leak, which is probably not a big                    //
      // deal and can be fixed by GC, rather than a mysteriously missing asset.)                                      //
                                                                                                                      //
      check(id, String);                                                                                              // 1463
                                                                                                                      //
      var existing = StaticAssets.findAndModify({                                                                     // 1465
        query: { _id: id },                                                                                           // 1466
        update: { $inc: { refcount: -1 } },                                                                           // 1467
        fields: { _id: 1, refcount: 1 },                                                                              // 1468
        "new": true                                                                                                   // 1469
      });                                                                                                             //
      if (!existing) {                                                                                                // 1471
        console.error(new Error("unrefStaticAsset() called on asset that doesn't exist").stack);                      // 1472
      } else if (existing.refcount <= 0) {                                                                            //
        StaticAssets.remove({ _id: existing._id });                                                                   // 1474
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.getStaticAsset = function (id) {                                                            // 1478
      // Get a static asset's mimeType, encoding, and raw content.                                                    //
                                                                                                                      //
      check(id, String);                                                                                              // 1481
                                                                                                                      //
      var asset = StaticAssets.findOne(id, { fields: { _id: 0, mimeType: 1, encoding: 1, content: 1 } });             // 1483
      if (asset) {                                                                                                    // 1484
        // TODO(perf): Mongo converts buffers to something else. Figure out a way to avoid a copy                     //
        //   here.                                                                                                    //
        asset.content = new Buffer(asset.content);                                                                    // 1487
      }                                                                                                               //
                                                                                                                      //
      return asset;                                                                                                   // 1490
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.newAssetUpload = function (purpose) {                                                       // 1493
      check(purpose, { profilePicture: { userId: DatabaseId, identityId: DatabaseId } });                             // 1494
                                                                                                                      //
      return AssetUploadTokens.insert({                                                                               // 1496
        purpose: purpose,                                                                                             // 1497
        expires: new Date(Date.now() + 300000) });                                                                    // 1498
    };                                                                                                                //
                                                                                                                      //
    // in 5 minutes                                                                                                   //
    SandstormDb.prototype.fulfillAssetUpload = function (id) {                                                        // 1502
      // Indicates that the given asset upload has completed. It will be removed and its purpose                      //
      // returned. If no matching upload exists, returns undefined.                                                   //
                                                                                                                      //
      check(id, String);                                                                                              // 1506
                                                                                                                      //
      var upload = AssetUploadTokens.findAndModify({                                                                  // 1508
        query: { _id: id },                                                                                           // 1509
        remove: true                                                                                                  // 1510
      });                                                                                                             //
                                                                                                                      //
      if (upload.expires.valueOf() < Date.now()) {                                                                    // 1513
        return undefined; // already expired                                                                          // 1514
      } else {                                                                                                        //
          return upload.purpose;                                                                                      // 1516
        }                                                                                                             //
    };                                                                                                                //
                                                                                                                      //
    function cleanupExpiredAssetUploads() {                                                                           // 1520
      AssetUploadTokens.remove({ expires: { $lt: Date.now() } });                                                     // 1521
    }                                                                                                                 //
                                                                                                                      //
    // Cleanup tokens every hour.                                                                                     //
    SandstormDb.periodicCleanup(3600000, cleanupExpiredAssetUploads);                                                 // 1525
                                                                                                                      //
    var packageCache = {};                                                                                            // 1527
    // Package info is immutable. Let's cache to save on mongo queries.                                               //
                                                                                                                      //
    SandstormDb.prototype.getPackage = function (packageId) {                                                         // 1530
      // Get the given package record. Since package info is immutable, cache the data in the server                  //
      // to reduce mongo query overhead, since it turns out we have to fetch specific packages a                      //
      // lot.                                                                                                         //
                                                                                                                      //
      if (packageId in packageCache) {                                                                                // 1535
        return packageCache[packageId];                                                                               // 1536
      }                                                                                                               //
                                                                                                                      //
      var pkg = Packages.findOne(packageId);                                                                          // 1539
      if (pkg && pkg.status === "ready") {                                                                            // 1540
        packageCache[packageId] = pkg;                                                                                // 1541
      }                                                                                                               //
                                                                                                                      //
      return pkg;                                                                                                     // 1544
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.sendAppUpdateNotifications = function (appId, packageId, name, versionNumber, marketingVersion) {
      var _this = this;                                                                                               // 1549
      var actions = _this.collections.userActions.find({ appId: appId, appVersion: { $lt: versionNumber } }, { fields: { userId: 1 } });
      actions.forEach(function (action) {                                                                             // 1552
        var userId = action.userId;                                                                                   // 1553
        var updater = {                                                                                               // 1554
          userId: userId,                                                                                             // 1555
          timestamp: new Date(),                                                                                      // 1556
          isUnread: true                                                                                              // 1557
        };                                                                                                            //
                                                                                                                      //
        // Set only the appId that we care about. Use mongo's dot notation to specify only a single                   //
        // field inside of an object to update                                                                        //
        updater["appUpdates." + appId] = {                                                                            // 1562
          marketingVersion: marketingVersion,                                                                         // 1563
          packageId: packageId,                                                                                       // 1564
          name: name,                                                                                                 // 1565
          version: versionNumber                                                                                      // 1566
        };                                                                                                            //
        _this.collections.notifications.upsert({ userId: userId }, { $set: updater });                                // 1568
      });                                                                                                             //
                                                                                                                      //
      _this.collections.appIndex.update({ _id: appId }, { $set: { hasSentNotifications: true } });                    // 1571
                                                                                                                      //
      // In the case where we replaced a previous notification and that was the only reference to the                 //
      // package, we need to clean it up                                                                              //
      Meteor.call("deleteUnusedPackages", appId);                                                                     // 1575
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.sendReferralProgramNotification = sendReferralProgramNotification;                          // 1578
                                                                                                                      //
    SandstormDb.prototype.upgradeGrains = function (appId, version, packageId, backend) {                             // 1580
      check(appId, String);                                                                                           // 1581
      check(version, Match.Integer);                                                                                  // 1582
      check(packageId, String);                                                                                       // 1583
                                                                                                                      //
      var selector = {                                                                                                // 1585
        userId: Meteor.userId(),                                                                                      // 1586
        appId: appId,                                                                                                 // 1587
        appVersion: { $lte: version },                                                                                // 1588
        packageId: { $ne: packageId }                                                                                 // 1589
      };                                                                                                              //
                                                                                                                      //
      if (!this.isSimulation) {                                                                                       // 1592
        Grains.find(selector).forEach(function (grain) {                                                              // 1593
          backend.shutdownGrain(grain._id, grain.userId);                                                             // 1594
        });                                                                                                           //
      }                                                                                                               //
                                                                                                                      //
      Grains.update(selector, { $set: { appVersion: version, packageId: packageId } }, { multi: true });              // 1598
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.startInstall = function (packageId, url, retryFailed, isAutoUpdated) {                      // 1601
      // Mark package for possible installation.                                                                      //
                                                                                                                      //
      var fields = {                                                                                                  // 1604
        status: "download",                                                                                           // 1605
        progress: 0,                                                                                                  // 1606
        url: url,                                                                                                     // 1607
        isAutoUpdated: !!isAutoUpdated                                                                                // 1608
      };                                                                                                              //
                                                                                                                      //
      if (retryFailed) {                                                                                              // 1611
        Packages.update({ _id: packageId, status: "failed" }, { $set: fields });                                      // 1612
      } else {                                                                                                        //
        try {                                                                                                         // 1614
          fields._id = packageId;                                                                                     // 1615
          Packages.insert(fields);                                                                                    // 1616
        } catch (err) {                                                                                               //
          console.error("Simultaneous startInstall()s?", err.stack);                                                  // 1618
        }                                                                                                             //
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    var ValidKeyFingerprint = Match.Where(function (keyFingerprint) {                                                 // 1623
      check(keyFingerprint, String);                                                                                  // 1624
      return !!keyFingerprint.match(/^[0-9A-F]{40}$/);                                                                // 1625
    });                                                                                                               //
                                                                                                                      //
    SandstormDb.prototype.updateKeybaseProfileAsync = function (keyFingerprint) {                                     // 1628
      // Asynchronously fetch the given Keybase profile and populate the KeybaseProfiles collection.                  //
                                                                                                                      //
      check(keyFingerprint, ValidKeyFingerprint);                                                                     // 1631
                                                                                                                      //
      console.log("fetching keybase", keyFingerprint);                                                                // 1633
                                                                                                                      //
      HTTP.get("https://keybase.io/_/api/1.0/user/lookup.json?key_fingerprint=" + keyFingerprint + "&fields=basics,profile,proofs_summary", {
        timeout: 5000                                                                                                 // 1638
      }, function (err, keybaseResponse) {                                                                            //
        if (err) {                                                                                                    // 1640
          console.log("keybase lookup error:", err.stack);                                                            // 1641
          return;                                                                                                     // 1642
        }                                                                                                             //
                                                                                                                      //
        if (!keybaseResponse.data) {                                                                                  // 1645
          console.log("keybase didn't return JSON? Headers:", keybaseResponse.headers);                               // 1646
          return;                                                                                                     // 1647
        }                                                                                                             //
                                                                                                                      //
        var profile = (keybaseResponse.data.them || [])[0];                                                           // 1650
                                                                                                                      //
        if (profile) {                                                                                                // 1652
          // jscs:disable requireCamelCaseOrUpperCaseIdentifiers                                                      //
          var record = {                                                                                              // 1654
            displayName: (profile.profile || {}).full_name,                                                           // 1655
            handle: (profile.basics || {}).username,                                                                  // 1656
            proofs: (profile.proofs_summary || {}).all || []                                                          // 1657
          };                                                                                                          //
          // jscs:enable requireCamelCaseOrUpperCaseIdentifiers                                                       //
                                                                                                                      //
          record.proofs.forEach(function (proof) {                                                                    // 1661
            // Remove potentially Mongo-incompatible stuff. (Currently Keybase returns nothing that                   //
            // this would filter.)                                                                                    //
            for (var field in babelHelpers.sanitizeForInObject(proof)) {                                              // 1664
              // Don't allow field names containing '.' or '$'. Also don't allow sub-objects mainly                   //
              // because I'm too lazy to check the field names recursively (and Keybase doesn't                       //
              // return any objects anyway).                                                                          //
              if (field.match(/[.$]/) || typeof proof[field] === "object") {                                          // 1668
                delete proof[field];                                                                                  // 1669
              }                                                                                                       //
            }                                                                                                         //
                                                                                                                      //
            // Indicate not verified.                                                                                 //
            // TODO(security): Asynchronously verify proofs. Presumably we can borrow code from the                   //
            //   Keybase node-based CLI.                                                                              //
            proof.status = "unverified";                                                                              // 1676
          });                                                                                                         //
                                                                                                                      //
          KeybaseProfiles.update(keyFingerprint, { $set: record }, { upsert: true });                                 // 1679
        } else {                                                                                                      //
          // Keybase reports no match, so remove what we know of this user. We don't want to remove                   //
          // the item entirely from the cache as this will cause us to repeatedly re-fetch the data                   //
          // from Keybase.                                                                                            //
          //                                                                                                          //
          // TODO(someday): We could perhaps keep the proofs if we can still verify them directly,                    //
          //   but at present we don't have the ability to verify proofs.                                             //
          KeybaseProfiles.update(keyFingerprint, { $unset: { displayName: "", handle: "", proofs: "" } }, { upsert: true });
        }                                                                                                             //
      });                                                                                                             //
    };                                                                                                                //
                                                                                                                      //
    SandstormDb.prototype.deleteUnusedAccount = function (backend, identityId) {                                      // 1693
      // If there is an *unused* account that has `identityId` as a login identity, deletes it.                       //
                                                                                                                      //
      check(identityId, String);                                                                                      // 1696
      var account = Meteor.users.findOne({ "loginIdentities.id": identityId });                                       // 1697
      if (account && account.loginIdentities.length == 1 && account.nonloginIdentities.length == 0 && !Grains.findOne({ userId: account._id }) && !ApiTokens.findOne({ accountId: account._id }) && (!account.plan || account.plan === "free") && !(account.payments && account.payments.id) && !Contacts.findOne({ ownerId: account._id })) {
        Meteor.users.remove({ _id: account._id });                                                                    // 1706
        backend.deleteUser(account._id);                                                                              // 1707
      }                                                                                                               //
    };                                                                                                                //
                                                                                                                      //
    Meteor.publish("keybaseProfile", function (keyFingerprint) {                                                      // 1711
      check(keyFingerprint, ValidKeyFingerprint);                                                                     // 1712
      var db = this.connection.sandstormDb;                                                                           // 1713
                                                                                                                      //
      var cursor = db.collections.keybaseProfiles.find(keyFingerprint);                                               // 1715
      if (cursor.count() === 0) {                                                                                     // 1716
        // Fire off async update.                                                                                     //
        db.updateKeybaseProfileAsync(keyFingerprint);                                                                 // 1718
      }                                                                                                               //
                                                                                                                      //
      return cursor;                                                                                                  // 1721
    });                                                                                                               //
                                                                                                                      //
    Meteor.publish("appIndex", function (appId) {                                                                     // 1724
      check(appId, String);                                                                                           // 1725
      var db = this.connection.sandstormDb;                                                                           // 1726
      var cursor = db.collections.appIndex.find({ _id: appId });                                                      // 1727
      return cursor;                                                                                                  // 1728
    });                                                                                                               //
                                                                                                                      //
    Meteor.publish("userPackages", function () {                                                                      // 1731
      var _this2 = this;                                                                                              //
                                                                                                                      //
      // Users should be able to see packages that are either:                                                        //
      // 1. referenced by one of their userActions                                                                    //
      // 2. referenced by one of their grains                                                                         //
      var db = this.connection.sandstormDb;                                                                           // 1735
                                                                                                                      //
      // Note that package information, once it is in the database, is static. There's no need to                     //
      // reactively subscribe to changes to a package since they don't change. It's also unecessary                   //
      // to reactively remove a package from the client side when it is removed on the server, or                     //
      // when the client stops using it, because the worst case is the client has a small amount                      //
      // of extra info on a no-longer-used package held in memory until they refresh Sandstorm.                       //
      // So, we implement this as a cache: the first time each package ID shows up among the user's                   //
      // stuff, we push the package info to the client, and then we never update it.                                  //
      //                                                                                                              //
      // Alternatively, we could subscribe to each individual package query, but this would waste                     //
      // lots of server-side resources watching for events that will never happen or don't matter.                    //
      var hasPackage = {};                                                                                            // 1747
      var refPackage = function (packageId) {                                                                         // 1748
        // Ignore dev apps.                                                                                           //
        if (packageId.lastIndexOf("dev-", 0) === 0) return;                                                           // 1750
                                                                                                                      //
        if (!hasPackage[packageId]) {                                                                                 // 1752
          hasPackage[packageId] = true;                                                                               // 1753
          var pkg = db.getPackage(packageId);                                                                         // 1754
          if (pkg) {                                                                                                  // 1755
            _this2.added("packages", packageId, pkg);                                                                 // 1756
          } else {                                                                                                    //
            console.error("shouldn't happen: missing package referenced by user's stuff:", packageId);                // 1758
          }                                                                                                           //
        }                                                                                                             //
      };                                                                                                              //
                                                                                                                      //
      // package source 1: packages referred to by actions                                                            //
      var actions = db.userActions(this.userId);                                                                      // 1765
      var actionsHandle = actions.observe({                                                                           // 1766
        added: function (newAction) {                                                                                 // 1767
          refPackage(newAction.packageId);                                                                            // 1768
        },                                                                                                            //
                                                                                                                      //
        changed: function (oldAction, newAction) {                                                                    // 1771
          refPackage(newAction.packageId);                                                                            // 1772
        }                                                                                                             //
      });                                                                                                             //
                                                                                                                      //
      // package source 2: packages referred to by grains directly                                                    //
      var grains = db.userGrains(this.userId);                                                                        // 1777
      var grainsHandle = grains.observe({                                                                             // 1778
        added: function (newGrain) {                                                                                  // 1779
          refPackage(newGrain.packageId);                                                                             // 1780
        },                                                                                                            //
                                                                                                                      //
        changed: function (oldGrain, newGrain) {                                                                      // 1783
          refPackage(newGrain.packageId);                                                                             // 1784
        }                                                                                                             //
      });                                                                                                             //
                                                                                                                      //
      this.onStop(function () {                                                                                       // 1788
        actionsHandle.stop();                                                                                         // 1789
        grainsHandle.stop();                                                                                          // 1790
      });                                                                                                             //
                                                                                                                      //
      this.ready();                                                                                                   // 1793
    });                                                                                                               //
  })();                                                                                                               //
}                                                                                                                     //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/profile.js                                                                                   //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                                 //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                              //
// All rights reserved.                                                                                               //
//                                                                                                                    //
// Licensed under the Apache License, Version 2.0 (the "License");                                                    //
// you may not use this file except in compliance with the License.                                                   //
// You may obtain a copy of the License at                                                                            //
//                                                                                                                    //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                       //
//                                                                                                                    //
// Unless required by applicable law or agreed to in writing, software                                                //
// distributed under the License is distributed on an "AS IS" BASIS,                                                  //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                           //
// See the License for the specific language governing permissions and                                                //
// limitations under the License.                                                                                     //
                                                                                                                      //
var makeIdenticon = undefined;                                                                                        // 17
var httpProtocol = undefined;                                                                                         // 18
                                                                                                                      //
if (Meteor.isServer) {                                                                                                // 20
  SandstormDb.ensureSubscriberHasIdentity = function (publishHandler, identityId) {                                   // 21
    // Helper for publish functions that need to restrict access based on whether the subscriber                      //
    // has a given identity linked. Automatically stops the subscription if the user loses the                        //
    // identity. Returns a boolean indicating whether the user initially has the identity.                            //
                                                                                                                      //
    var userId = publishHandler.userId;                                                                               // 26
    if (userId === identityId) {                                                                                      // 27
      return true;                                                                                                    // 28
    } else {                                                                                                          //
      var _ret = (function () {                                                                                       //
        var hasIdentityCursor = Meteor.users.find({                                                                   // 30
          $or: [{                                                                                                     // 31
            _id: userId,                                                                                              // 33
            "loginIdentities.id": identityId                                                                          // 34
          }, {                                                                                                        //
            _id: userId,                                                                                              // 37
            "nonloginIdentities.id": identityId                                                                       // 38
          }]                                                                                                          //
        });                                                                                                           //
        if (hasIdentityCursor.count() == 0) {                                                                         // 42
          publishHandler.stop();                                                                                      // 43
          return {                                                                                                    // 44
            v: false                                                                                                  //
          };                                                                                                          //
        }                                                                                                             //
                                                                                                                      //
        var handle = hasIdentityCursor.observe({ removed: function () {                                               // 47
            publishHandler.stop();                                                                                    // 47
          } });                                                                                                       //
                                                                                                                      //
        publishHandler.onStop(function () {                                                                           // 49
          handle.stop();                                                                                              // 49
        });                                                                                                           //
                                                                                                                      //
        return {                                                                                                      // 51
          v: true                                                                                                     //
        };                                                                                                            //
      })();                                                                                                           //
                                                                                                                      //
      if (typeof _ret === "object") return _ret.v;                                                                    //
    }                                                                                                                 //
  };                                                                                                                  //
                                                                                                                      //
  Meteor.publish("identityProfile", function (identityId) {                                                           // 55
    check(identityId, String);                                                                                        // 56
    if (!SandstormDb.ensureSubscriberHasIdentity(this, identityId)) return;                                           // 57
                                                                                                                      //
    return Meteor.users.find({ _id: identityId }, { fields: {                                                         // 59
        profile: 1,                                                                                                   // 61
        unverifiedEmail: 1,                                                                                           // 62
        expires: 1,                                                                                                   // 63
        createdAt: 1,                                                                                                 // 64
                                                                                                                      //
        "services.dev.name": 1,                                                                                       // 66
                                                                                                                      //
        "services.google.id": 1,                                                                                      // 68
        "services.google.email": 1,                                                                                   // 69
        "services.google.verified_email": 1,                                                                          // 70
        "services.google.name": 1,                                                                                    // 71
        "services.google.picture": 1,                                                                                 // 72
        "services.google.gender": 1,                                                                                  // 73
                                                                                                                      //
        "services.github.id": 1,                                                                                      // 75
        "services.github.email": 1,                                                                                   // 76
        "services.github.emails": 1,                                                                                  // 77
        "services.github.username": 1,                                                                                // 78
                                                                                                                      //
        "services.email.email": 1,                                                                                    // 80
                                                                                                                      //
        "services.ldap.id": 1,                                                                                        // 82
        "services.ldap.rawAttrs": 1                                                                                   // 83
      } });                                                                                                           //
  }), Meteor.publish("accountIdentities", function () {                                                               //
    // Maybe this should be folded into the "credentials" subscription?                                               //
                                                                                                                      //
    if (!this.userId) return [];                                                                                      // 90
                                                                                                                      //
    return Meteor.users.find({ _id: this.userId }, { fields: {                                                        // 92
        profile: 1,                                                                                                   // 95
        verifiedEmail: 1,                                                                                             // 96
        loginIdentities: 1,                                                                                           // 97
        nonloginIdentities: 1,                                                                                        // 98
        expires: 1,                                                                                                   // 99
        primaryEmail: 1                                                                                               // 100
      } });                                                                                                           //
  });                                                                                                                 //
                                                                                                                      //
  makeIdenticon = function (id) {                                                                                     // 104
    // We only make identicons client-side.                                                                           //
    return undefined;                                                                                                 // 106
  };                                                                                                                  //
                                                                                                                      //
  var Url = Npm.require("url");                                                                                       // 109
  httpProtocol = Url.parse(process.env.ROOT_URL).protocol;                                                            // 110
} else {                                                                                                              //
  (function () {                                                                                                      //
    var identiconCache = {};                                                                                          // 112
                                                                                                                      //
    makeIdenticon = function (id) {                                                                                   // 114
      // Given a cryptographic hash as input, generate an identicon. We always use the user's                         //
      // identity ID as the input hash (the same thing passed to apps as the user ID).                                //
                                                                                                                      //
      // Pass to identicon.js exactly the hash that we'd pass to apps as X-Sandstorm-User-Id, so                      //
      // that apps can themselves use identicon.js to produce consistent identicons. As it turns out                  //
      // identicon.js doesn't use the second half of the hash even if we provide it, but slice it                     //
      // anyway to be safe.                                                                                           //
      id = id.slice(0, 32);                                                                                           // 122
                                                                                                                      //
      if (id in identiconCache) {                                                                                     // 124
        return identiconCache[id];                                                                                    // 125
      }                                                                                                               //
                                                                                                                      //
      // Unfortunately, Github's algorithm uses MD5. Whatever, we don't expect these to be secure.                    //
      var data = new Identicon(id, 512).toString();                                                                   // 129
      var result = "data:image/png;base64," + data;                                                                   // 130
      identiconCache[id] = result;                                                                                    // 131
      return result;                                                                                                  // 132
    };                                                                                                                //
                                                                                                                      //
    httpProtocol = window.location.protocol;                                                                          // 135
  })();                                                                                                               //
}                                                                                                                     //
                                                                                                                      //
var GENDERS = { male: "male", female: "female", neutral: "neutral", robot: "robot" };                                 // 138
                                                                                                                      //
function staticAssetUrl(id, staticHost) {                                                                             // 140
  if (id) {                                                                                                           // 141
    return staticHost + "/" + id;                                                                                     // 142
  } else {                                                                                                            //
    return undefined;                                                                                                 // 144
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function filterHandle(handle) {                                                                                       // 148
  // Convert disallowed letters into underscores, but no more than one underscore in a row.                           //
  return handle && handle.toLowerCase().split(/[^a-z0-9_]+/g).join("_");                                              // 150
}                                                                                                                     //
                                                                                                                      //
function emailToHandle(email) {                                                                                       // 153
  // Turn an email address into a handle.                                                                             //
                                                                                                                      //
  if (!email) return undefined;                                                                                       // 156
                                                                                                                      //
  // Use the stuff before the @. Ignore stuff after '+' because it's commonly used for filters                        //
  // on the same account.                                                                                             //
  var parts = email.split("@");                                                                                       // 160
  var base = filterHandle(parts[0].split("+")[0]);                                                                    // 161
                                                                                                                      //
  var domain = (parts[1] || "").split(".");                                                                           // 163
  if (domain[domain.length - 1] === "name") {                                                                         // 164
    // Oh, a .name domain. Let's use                                                                                  //
    base = filterHandle(domain.slice(0, domain.length - 1).join("."));                                                // 166
  } else if (_.contains(["me", "self", "contact", "admin", "administrator", "root", "info", "sandstorm", "sandstormio", "inbox", "indiegogo", "mail", "email"], base)) {
    // This is probably an address at a vanity domain. Use the domain itself as the handle.                           //
    base = filterHandle(domain[0]);                                                                                   // 174
  }                                                                                                                   //
                                                                                                                      //
  return filterHandle(base);                                                                                          // 177
}                                                                                                                     //
                                                                                                                      //
SandstormDb.fillInProfileDefaults = function (user) {                                                                 // 180
  var profile = user.profile;                                                                                         // 181
  if (profile.service === "github") {                                                                                 // 182
    profile.name = profile.name || user.services.github.username || "Name Unknown";                                   // 183
    profile.handle = profile.handle || filterHandle(user.services.github.username) || filterHandle(profile.name);     // 184
  } else if (profile.service === "google") {                                                                          //
    profile.name = profile.name || user.services.google.name || "Name Unknown";                                       // 187
    profile.handle = profile.handle || emailToHandle(user.services.google.email) || filterHandle(profile.name);       // 188
    profile.pronoun = profile.pronoun || GENDERS[user.services.google.gender] || "neutral";                           // 190
  } else if (profile.service === "email") {                                                                           //
    var email = user.services.email.email;                                                                            // 192
    profile.name = profile.name || email.split("@")[0];                                                               // 193
    profile.handle = profile.handle || emailToHandle(email);                                                          // 194
  } else if (profile.service === "dev") {                                                                             //
    var lowerCaseName = user.services.dev.name.split(" ")[0].toLowerCase();                                           // 196
    profile.name = profile.name || user.services.dev.name;                                                            // 197
    profile.handle = profile.handle || filterHandle(lowerCaseName);                                                   // 198
    profile.pronoun = profile.pronoun || (_.contains(["alice", "carol", "eve"], lowerCaseName) ? "female" : _.contains(["bob", "dave"], lowerCaseName) ? "male" : "neutral");
  } else if (profile.service === "demo") {                                                                            //
    profile.name = profile.name || "Demo User";                                                                       // 203
    profile.handle = profile.handle || "demo";                                                                        // 204
  } else if (profile.service === "ldap") {                                                                            //
    var setting = Settings.findOne({ _id: "ldapNameField" });                                                         // 206
    var key = setting ? setting.value : "";                                                                           // 207
    profile.name = profile.name || user.services.ldap.rawAttrs[key] || "Name Unknown";                                // 208
    profile.handle = profile.handle || filterHandle(profile.name);                                                    // 209
  } else {                                                                                                            //
    throw new Error("unrecognized identity service: ", profile.service);                                              // 211
  }                                                                                                                   //
                                                                                                                      //
  profile.pronoun = profile.pronoun || "neutral";                                                                     // 214
};                                                                                                                    //
                                                                                                                      //
SandstormDb.fillInIntrinsicName = function (user) {                                                                   // 217
  var profile = user.profile;                                                                                         // 218
  if (profile.service === "github") {                                                                                 // 219
    profile.intrinsicName = user.services.github.username;                                                            // 220
  } else if (profile.service === "google") {                                                                          //
    profile.intrinsicName = user.services.google.name;                                                                // 222
    user.privateIntrinsicName = user.services.google.email;                                                           // 223
  } else if (profile.service === "email") {                                                                           //
    profile.intrinsicName = user.services.email.email;                                                                // 225
  } else if (profile.service === "dev") {                                                                             //
    profile.intrinsicName = user.services.dev.name;                                                                   // 227
  } else if (profile.service === "demo") {                                                                            //
    profile.intrinsicName = "demo on " + user.createdAt.toISOString().substring(0, 10);                               // 229
  } else if (profile.service === "ldap") {                                                                            //
    profile.intrinsicName = user.services.ldap.username;                                                              // 231
  } else {                                                                                                            //
    throw new Error("unrecognized identity service: ", profile.service);                                              // 233
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
SandstormDb.getVerifiedEmails = function (identity) {                                                                 // 237
  if (identity.services.google && identity.services.google.email && identity.services.google.verified_email) {        // 238
    // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                             //
    return [{ email: identity.services.google.email, primary: true }];                                                // 240
  } else if (identity.services.email) {                                                                               //
    return [{ email: identity.services.email.email, primary: true }];                                                 // 242
  } else if (identity.services.github && identity.services.github.emails) {                                           //
    return _.chain(identity.services.github.emails).filter(function (email) {                                         // 244
      return email.verified;                                                                                          // 245
    }).map(function (email) {                                                                                         //
      return _.pick(email, "email", "primary");                                                                       //
    }).value();                                                                                                       //
  }                                                                                                                   //
                                                                                                                      //
  return [];                                                                                                          // 250
};                                                                                                                    //
                                                                                                                      //
SandstormDb.prototype.findIdentitiesByEmail = function (email) {                                                      // 253
  // Returns an array of identities which have the given email address as one of their verified                       //
  // addresses.                                                                                                       //
                                                                                                                      //
  check(email, String);                                                                                               // 257
                                                                                                                      //
  return Meteor.users.find({ $or: [{ "services.google.email": email }, { "services.email.email": email }, { "services.github.emails.email": email }] }).fetch().filter(function (identity) {
    // Verify that the email is verified, since our query doesn't technically do that.                                //
    return !!_findWhere(SandstormDb.getVerifiedEmails(identity), { email: email });                                   // 265
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
SandstormDb.prototype.findAccountsByEmail = function (email) {                                                        // 269
  var identityIds = _.pluck(this.findIdentitiesByEmail(email), "_id");                                                // 270
  return Meteor.users.find({ $or: [{ "loginIdentities.id": { $in: identityIds } }, { "nonloginIdentities.id": { $in: identityIds } }] }).fetch();
};                                                                                                                    //
                                                                                                                      //
SandstormDb.fillInPictureUrl = function (user) {                                                                      // 277
  var staticHost = httpProtocol + "//" + makeWildcardHost("static");                                                  // 278
  user.profile.pictureUrl = staticAssetUrl(user.profile.picture, staticHost) || makeIdenticon(user._id);              // 279
};                                                                                                                    //
                                                                                                                      //
SandstormDb.getUserIdentityIds = function (user) {                                                                    // 283
  // Given an account user object, returns an array containing the ID of each identity linked to the                  //
  // account. Always returns the most recently added login identity first.                                            //
  if (user && user.loginIdentities) {                                                                                 // 286
    return _.pluck(user.nonloginIdentities.concat(user.loginIdentities), "id").reverse();                             // 287
  } else {                                                                                                            //
    return [];                                                                                                        // 289
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
SandstormDb.getUserEmails = function (user) {                                                                         // 293
  // Given a user object, returns an array containing all email addresses associated with that user.                  //
  // Each entry in the array is an object of the form:                                                                //
  //     `{email: String, verified: Bool, primary: Optional(Bool)}`                                                   //
  //                                                                                                                  //
  // At most one entry in the result has `primary = true`.                                                            //
  //                                                                                                                  //
  // TODO(cleanup): This actually does need to query the database to fetch profile information                        //
  //   for linked identities, so it probably makes more sense for it to be a non-static method                        //
  //   on SandstormDb.                                                                                                //
                                                                                                                      //
  var identityIds = SandstormDb.getUserIdentityIds(user);                                                             // 304
  var verifiedEmails = {};                                                                                            // 305
  var unverifiedEmails = {};                                                                                          // 306
                                                                                                                      //
  identityIds.forEach(function (id) {                                                                                 // 308
    var identity = Meteor.users.findOne({ _id: id });                                                                 // 309
    if (identity && identity.services) {                                                                              // 310
      SandstormDb.getVerifiedEmails(identity).forEach(function (verifiedEmail) {                                      // 311
        if (verifiedEmail) {                                                                                          // 312
          verifiedEmails[verifiedEmail.email] = true;                                                                 // 313
        }                                                                                                             //
      });                                                                                                             //
    }                                                                                                                 //
                                                                                                                      //
    if (identity && identity.unverifiedEmail) {                                                                       // 318
      unverifiedEmails[identity.unverifiedEmail] = true;                                                              // 319
    }                                                                                                                 //
  });                                                                                                                 //
                                                                                                                      //
  var result = [];                                                                                                    // 323
  _.keys(verifiedEmails).map(function (email) {                                                                       // 324
    result.push({ email: email,                                                                                       // 325
      verified: true,                                                                                                 // 326
      primary: email === user.primaryEmail });                                                                        // 327
  });                                                                                                                 //
                                                                                                                      //
  _.keys(unverifiedEmails).map(function (email) {                                                                     // 330
    if (!(email in verifiedEmails)) {                                                                                 // 331
      result.push({ email: email, verified: false });                                                                 // 331
    }                                                                                                                 //
  });                                                                                                                 //
                                                                                                                      //
  // If `user.primaryEmail` is not among the verified emails, mark the first as primary.                              //
  if (!(user.primaryEmail in verifiedEmails) && result.length > 0) {                                                  // 335
    result[0].primary = true;                                                                                         // 336
  }                                                                                                                   //
                                                                                                                      //
  return result;                                                                                                      // 339
};                                                                                                                    //
                                                                                                                      //
SandstormDb.prototype.addContact = (function () {                                                                     // 342
  function addContact(ownerId, identityId) {                                                                          // 342
    check(ownerId, String);                                                                                           // 343
    check(identityId, String);                                                                                        // 344
    var profile = this.getIdentity(identityId).profile;                                                               // 345
    this.collections.contacts.upsert({ ownerId: ownerId, identityId: identityId }, {                                  // 346
      ownerId: ownerId,                                                                                               // 347
      petname: profile && profile.name,                                                                               // 348
      created: new Date(),                                                                                            // 349
      identityId: identityId                                                                                          // 350
    });                                                                                                               //
  }                                                                                                                   //
                                                                                                                      //
  return addContact;                                                                                                  //
})();                                                                                                                 //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/user.js                                                                                      //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                                 //
// Copyright (c) 2015 Sandstorm Development Group, Inc. and contributors                                              //
// All rights reserved.                                                                                               //
//                                                                                                                    //
// Licensed under the Apache License, Version 2.0 (the "License");                                                    //
// you may not use this file except in compliance with the License.                                                   //
// You may obtain a copy of the License at                                                                            //
//                                                                                                                    //
//   http://www.apache.org/licenses/LICENSE-2.0                                                                       //
//                                                                                                                    //
// Unless required by applicable law or agreed to in writing, software                                                //
// distributed under the License is distributed on an "AS IS" BASIS,                                                  //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                                           //
// See the License for the specific language governing permissions and                                                //
// limitations under the License.                                                                                     //
                                                                                                                      //
var Crypto = Npm.require("crypto");                                                                                   // 17
var Future = Npm.require("fibers/future");                                                                            // 18
                                                                                                                      //
userPictureUrl = function (user) {                                                                                    // 20
  if (user.services && !(user.profile && user.profile.picture)) {                                                     // 21
    // Try to determine user's avatar URL from login service.                                                         //
                                                                                                                      //
    var google = user.services.google;                                                                                // 24
    if (google && google.picture) {                                                                                   // 25
      return google.picture;                                                                                          // 26
    }                                                                                                                 //
                                                                                                                      //
    var github = user.services.github;                                                                                // 29
    if (github && github.id) {                                                                                        // 30
      return "https://avatars.githubusercontent.com/u/" + github.id;                                                  // 31
    }                                                                                                                 //
                                                                                                                      //
    // Note that we do NOT support Gravatar for email addresses because pinging Gravatar would be                     //
    // a data leak, revealing that the user has logged into this Sandstorm server. Google and                         //
    // Github are different because they are actually the identity providers, so they already know                    //
    // the user logged in.                                                                                            //
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
fetchPicture = function (url) {                                                                                       // 41
  try {                                                                                                               // 42
    var result = HTTP.get(url, {                                                                                      // 43
      npmRequestOptions: { encoding: null },                                                                          // 44
      timeout: 5000                                                                                                   // 45
    });                                                                                                               //
                                                                                                                      //
    var metadata = {};                                                                                                // 48
                                                                                                                      //
    metadata.mimeType = result.headers["content-type"];                                                               // 50
    if (metadata.mimeType.lastIndexOf("image/png", 0) === -1 && metadata.mimeType.lastIndexOf("image/jpeg", 0) === -1) {
      throw new Error("unexpected Content-Type:", metadata.mimeType);                                                 // 53
    }                                                                                                                 //
                                                                                                                      //
    var enc = result.headers["content-encoding"];                                                                     // 56
    if (enc && enc !== "identity") {                                                                                  // 57
      metadata.encoding = enc;                                                                                        // 58
    }                                                                                                                 //
                                                                                                                      //
    return addStaticAsset(metadata, result.content);                                                                  // 61
  } catch (err) {                                                                                                     //
    console.error("failed to fetch user profile picture:", url, err.stack);                                           // 63
  }                                                                                                                   //
};                                                                                                                    //
                                                                                                                      //
var ValidHandle = Match.Where(function (handle) {                                                                     // 67
  check(handle, String);                                                                                              // 68
  return !!handle.match(/^[a-z_][a-z0-9_]*$/);                                                                        // 69
});                                                                                                                   //
                                                                                                                      //
Accounts.onCreateUser(function (options, user) {                                                                      // 72
  if (user.loginIdentities) {                                                                                         // 73
    // it's an account                                                                                                //
    check(user, { _id: String,                                                                                        // 75
      createdAt: Date,                                                                                                // 76
      isAdmin: Match.Optional(Boolean),                                                                               // 77
      hasCompletedSignup: Match.Optional(Boolean),                                                                    // 78
      signupKey: Match.Optional(String),                                                                              // 79
      signupNote: Match.Optional(String),                                                                             // 80
      signupEmail: Match.Optional(String),                                                                            // 81
      expires: Match.Optional(Date),                                                                                  // 82
      appDemoId: Match.Optional(String),                                                                              // 83
      loginIdentities: [{ id: String }],                                                                              // 84
      nonloginIdentities: [{ id: String }] });                                                                        // 85
                                                                                                                      //
    if (Meteor.settings["public"].quotaEnabled) {                                                                     // 87
      user.experiments = user.experiments || {};                                                                      // 88
      user.experiments = {                                                                                            // 89
        firstTimeBillingPrompt: Math.random() < 0.5 ? "control" : "test"                                              // 90
      };                                                                                                              //
      if (!("expires" in user)) {                                                                                     // 92
        sendReferralProgramNotification(user._id);                                                                    // 93
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    return user;                                                                                                      // 97
  }                                                                                                                   //
                                                                                                                      //
  // Check profile.                                                                                                   //
  if (options.profile) {                                                                                              // 101
    // TODO(cleanup): This check also appears in accounts-ui-methods.js.                                              //
    check(options.profile, Match.ObjectIncluding({                                                                    // 103
      name: Match.OneOf(null, Match.Optional(String)),                                                                // 104
      handle: Match.Optional(ValidHandle),                                                                            // 105
      pronoun: Match.Optional(Match.OneOf("male", "female", "neutral", "robot"))                                      // 106
    }));                                                                                                              //
  }                                                                                                                   //
                                                                                                                      //
  check(options.unverifiedEmail, Match.Optional(String));                                                             // 110
                                                                                                                      //
  if (options.unverifiedEmail) {                                                                                      // 112
    user.unverifiedEmail = options.unverifiedEmail;                                                                   // 113
  }                                                                                                                   //
                                                                                                                      //
  user.profile = _.pick(options.profile || {}, "name", "handle", "pronouns");                                         // 116
                                                                                                                      //
  // Try downloading avatar.                                                                                          //
  var url = userPictureUrl(user);                                                                                     // 119
  if (url) {                                                                                                          // 120
    var assetId = fetchPicture(url);                                                                                  // 121
    if (assetId) {                                                                                                    // 122
      user.profile.picture = assetId;                                                                                 // 123
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  var serviceUserId = undefined;                                                                                      // 127
  if (user.services && user.services.dev) {                                                                           // 128
    check(user.services.dev, { name: String, isAdmin: Boolean, hasCompletedSignup: Boolean });                        // 129
    serviceUserId = user.services.dev.name;                                                                           // 130
    user.profile.service = "dev";                                                                                     // 131
  } else if ("expires" in user) {                                                                                     //
    serviceUserId = user._id;                                                                                         // 133
    user.profile.service = "demo";                                                                                    // 134
  } else if (user.services && user.services.email) {                                                                  //
    check(user.services.email, { email: String,                                                                       // 136
      tokens: [{ digest: String, algorithm: String, createdAt: Date }] });                                            // 138
    serviceUserId = user.services.email.email;                                                                        // 139
    user.profile.service = "email";                                                                                   // 140
  } else if (user.services && "google" in user.services) {                                                            //
    serviceUserId = user.services.google.id;                                                                          // 142
    user.profile.service = "google";                                                                                  // 143
  } else if (user.services && "github" in user.services) {                                                            //
    serviceUserId = user.services.github.id;                                                                          // 145
    user.profile.service = "github";                                                                                  // 146
  } else if (user.services && "ldap" in user.services) {                                                              //
    serviceUserId = user.services.ldap.id;                                                                            // 148
    user.profile.service = "ldap";                                                                                    // 149
  } else {                                                                                                            //
    throw new Meteor.Error(400, "user does not have a recognized identity provider: " + JSON.stringify(user));        // 151
  }                                                                                                                   //
                                                                                                                      //
  user._id = Crypto.createHash("sha256").update(user.profile.service + ":" + serviceUserId).digest("hex");            // 155
                                                                                                                      //
  return user;                                                                                                        // 158
});                                                                                                                   //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function(){

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                                //
//                                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                                   //
//                                                                                                                    //
// Because migrations can experience partial failure and likely have                                                  //
// side-effects, we should be careful to make sure all migrations are                                                 //
// idempotent and safe to accidentally run multiple times.                                                            //
                                                                                                                      //
var Future = Npm.require("fibers/future");                                                                            // 7
var Url = Npm.require("url");                                                                                         // 8
var Crypto = Npm.require("crypto");                                                                                   // 9
                                                                                                                      //
var updateLoginStyleToRedirect = function () {                                                                        // 11
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                          // 12
  ["google", "github"].forEach(function (serviceName) {                                                               // 13
    var config = configurations.findOne({ service: serviceName });                                                    // 14
    if (config && config.loginStyle !== "redirect") {                                                                 // 15
      configurations.update({ service: serviceName }, { $set: { loginStyle: "redirect" } });                          // 16
    }                                                                                                                 //
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
var enableLegacyOAuthProvidersIfNotInSettings = function () {                                                         // 21
  // In the before time, Google and Github login were enabled by default.                                             //
  //                                                                                                                  //
  // This actually didn't make much sense, required the first user to configure                                       //
  // OAuth, and had some trust-the-first-user properties that weren't totally                                         //
  // secure.                                                                                                          //
  //                                                                                                                  //
  // Now, we have admin-token, so we wish to disable all logins by default                                            //
  // (since they need to be configured anyway) but since users may have never                                         //
  // explicitly told Sandstorm that Google or Github login should be enabled,                                         //
  // we can't just use the value in the Settings collection, since it might                                           //
  // never have been set.                                                                                             //
  //                                                                                                                  //
  // Thus, if the service is configured but absent from Settings, we should                                           //
  // explicitly enable it in Settings, and then the rest of the logic can just                                        //
  // depend on what value is in Settings and default to false without breaking                                        //
  // user installations.                                                                                              //
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;                          // 38
  ["google", "github"].forEach(function (serviceName) {                                                               // 39
    var config = configurations.findOne({ service: serviceName });                                                    // 40
    var serviceConfig = Settings.findOne({ _id: serviceName });                                                       // 41
    if (config && !serviceConfig) {                                                                                   // 42
      // Only explicitly enable the login service if:                                                                 //
      // 1) the service is already configured                                                                         //
      // 2) there is no sandstorm configuration already present (the user was                                         //
      //    using the previous default behavior).                                                                     //
      Settings.insert({ _id: serviceName, value: true });                                                             // 47
    }                                                                                                                 //
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
var denormalizeInviteInfo = function () {                                                                             // 52
  // When a user is invited via a signup token, the `signupKey` field of their user table entry                       //
  // has always been populated to indicate the key they used. This points into the SignupKeys table                   //
  // which has more information about the key, namely a freeform note entered by the admin when                       //
  // they created the key. In the case that the email invite form was used, the note has the form                     //
  // "E-mail invite to <address>".                                                                                    //
  //                                                                                                                  //
  // Later, we decided it was useful to indicate in the users table visible to the admin                              //
  // information about the invite terms. Namely, for email invites we want to show the address                        //
  // and for others we want to show the note. To make this efficient, fields `signupNote` and                         //
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing                       //
  // from the SignupKeys table.                                                                                       //
                                                                                                                      //
  Meteor.users.find().forEach(function (user) {                                                                       // 65
    if (user.signupKey && typeof user.signupKey === "string" && user.signupKey !== "admin") {                         // 66
      var signupInfo = SignupKeys.findOne(user.signupKey);                                                            // 67
      if (signupInfo && signupInfo.note) {                                                                            // 68
        var newFields = { signupNote: signupInfo.note };                                                              // 69
                                                                                                                      //
        var prefix = "E-mail invite to ";                                                                             // 71
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                                              // 72
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                                               // 73
        }                                                                                                             //
                                                                                                                      //
        Meteor.users.update(user._id, { $set: newFields });                                                           // 76
      }                                                                                                               //
    }                                                                                                                 //
  });                                                                                                                 //
};                                                                                                                    //
                                                                                                                      //
function mergeRoleAssignmentsIntoApiTokens() {                                                                        // 82
  RoleAssignments.find().forEach(function (roleAssignment) {                                                          // 83
    ApiTokens.insert({                                                                                                // 84
      grainId: roleAssignment.grainId,                                                                                // 85
      userId: roleAssignment.sharer,                                                                                  // 86
      roleAssignment: roleAssignment.roleAssignment,                                                                  // 87
      petname: roleAssignment.petname,                                                                                // 88
      created: roleAssignment.created,                                                                                // 89
      owner: { user: { userId: roleAssignment.recipient,                                                              // 90
          title: roleAssignment.title } }                                                                             // 91
    });                                                                                                               //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function fixOasisStorageUsageStats() {}                                                                               // 96
// This migration only pertained to Oasis and it was successfully applied there. Since it referred                    //
// to some global variables that we later wanted to remove and/or rename, we've since replaced it                     //
// with a no-op.                                                                                                      //
                                                                                                                      //
function fetchProfilePictures() {                                                                                     // 101
  Meteor.users.find({}).forEach(function (user) {                                                                     // 102
    var url = userPictureUrl(user);                                                                                   // 103
    if (url) {                                                                                                        // 104
      console.log("Fetching user picture:", url);                                                                     // 105
      var assetId = fetchPicture(url);                                                                                // 106
      if (assetId) {                                                                                                  // 107
        Meteor.users.update(user._id, { $set: { "profile.picture": assetId } });                                      // 108
      }                                                                                                               //
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function assignPlans() {                                                                                              // 114
  if (Meteor.settings["public"].quotaEnabled && SandstormDb.paymentsMigrationHook) {                                  // 115
    SandstormDb.paymentsMigrationHook(SignupKeys, Plans.find().fetch());                                              // 116
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function removeKeyrings() {                                                                                           // 120
  // These blobs full of public keys were not intended to find their way into mongo and while                         //
  // harmless they slow things down because they're huge. Remove them.                                                //
  Packages.update({ "manifest.metadata.pgpKeyring": { $exists: true } }, { $unset: { "manifest.metadata.pgpKeyring": "" } }, { multi: true });
}                                                                                                                     //
                                                                                                                      //
function useLocalizedTextInUserActions() {                                                                            // 128
  function toLocalizedText(newObj, oldObj, field) {                                                                   // 129
    if (field in oldObj) {                                                                                            // 130
      if (typeof oldObj[field] === "string") {                                                                        // 131
        newObj[field] = { defaultText: oldObj[field] };                                                               // 132
      } else {                                                                                                        //
        newObj[field] = oldObj[field];                                                                                // 134
      }                                                                                                               //
    }                                                                                                                 //
  }                                                                                                                   //
                                                                                                                      //
  UserActions.find({}).forEach(function (userAction) {                                                                // 139
    var fields = {};                                                                                                  // 140
    toLocalizedText(fields, userAction, "appTitle");                                                                  // 141
    toLocalizedText(fields, userAction, "title");                                                                     // 142
    toLocalizedText(fields, userAction, "nounPhrase");                                                                // 143
    UserActions.update(userAction._id, { $set: fields });                                                             // 144
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function verifyAllPgpSignatures() {                                                                                   // 148
  Packages.find({}).forEach(function (pkg) {                                                                          // 149
    try {                                                                                                             // 150
      console.log("checking PGP signature for package:", pkg._id);                                                    // 151
      var info = waitPromise(globalBackend.cap().tryGetPackage(pkg._id));                                             // 152
      if (info.authorPgpKeyFingerprint) {                                                                             // 153
        console.log("  " + info.authorPgpKeyFingerprint);                                                             // 154
        Packages.update(pkg._id, { $set: { authorPgpKeyFingerprint: info.authorPgpKeyFingerprint } });                // 155
      } else {                                                                                                        //
        console.log("  no signature");                                                                                // 158
      }                                                                                                               //
    } catch (err) {                                                                                                   //
      console.error(err.stack);                                                                                       // 161
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function splitUserIdsIntoAccountIdsAndIdentityIds() {                                                                 // 166
  Meteor.users.find().forEach(function (user) {                                                                       // 167
    var identity = {};                                                                                                // 168
    var serviceUserId = undefined;                                                                                    // 169
    if ("devName" in user) {                                                                                          // 170
      identity.service = "dev";                                                                                       // 171
      serviceUserId = user.devName;                                                                                   // 172
    } else if ("expires" in user) {                                                                                   //
      identity.service = "demo";                                                                                      // 174
      serviceUserId = user._id;                                                                                       // 175
    } else if (user.services && "google" in user.services) {                                                          //
      identity.service = "google";                                                                                    // 177
      if (user.services.google.email && user.services.google.verified_email) {                                        // 178
        // jscs:ignore requireCamelCaseOrUpperCaseIdentifiers                                                         //
        identity.verifiedEmail = user.services.google.email;                                                          // 179
      }                                                                                                               //
                                                                                                                      //
      serviceUserId = user.services.google.id;                                                                        // 182
    } else if (user.services && "github" in user.services) {                                                          //
      identity.service = "github";                                                                                    // 184
      identity.unverifiedEmail = user.services.github.email;                                                          // 185
      serviceUserId = user.services.github.id;                                                                        // 186
    } else if (user.services && "emailToken" in user.services) {                                                      //
      identity.service = "emailToken";                                                                                // 188
      identity.verifiedEmail = user.services.emailToken.email;                                                        // 189
      serviceUserId = user.services.emailToken.email;                                                                 // 190
    }                                                                                                                 //
                                                                                                                      //
    identity.id = Crypto.createHash("sha256").update(identity.service + ":" + serviceUserId).digest("hex");           // 193
                                                                                                                      //
    if (user.profile) {                                                                                               // 196
      if (user.profile.name) {                                                                                        // 197
        identity.name = user.profile.name;                                                                            // 198
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.handle) {                                                                                      // 201
        identity.handle = user.profile.handle;                                                                        // 202
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.picture) {                                                                                     // 205
        identity.picture = user.profile.picture;                                                                      // 206
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.pronoun) {                                                                                     // 209
        identity.pronoun = user.profile.pronoun;                                                                      // 210
      }                                                                                                               //
                                                                                                                      //
      if (user.profile.email) {                                                                                       // 213
        identity.unverifiedEmail = user.profile.email;                                                                // 214
      }                                                                                                               //
    }                                                                                                                 //
                                                                                                                      //
    identity.main = true;                                                                                             // 218
                                                                                                                      //
    Meteor.users.update(user._id, { $set: { identities: [identity] } });                                              // 220
                                                                                                                      //
    Grains.update({ userId: user._id }, { $set: { identityId: identity.id } }, { multi: true });                      // 222
    Sessions.update({ userId: user._id }, { $set: { identityId: identity.id } }, { multi: true });                    // 223
    ApiTokens.update({ userId: user._id }, { $set: { identityId: identity.id } }, { multi: true });                   // 224
    ApiTokens.update({ "owner.user.userId": user._id }, { $set: { "owner.user.identityId": identity.id } }, { multi: true });
    ApiTokens.update({ "owner.grain.introducerUser": user._id }, { $set: { "owner.grain.introducerIdentity": identity.id } }, { multi: true });
                                                                                                                      //
    while (ApiTokens.update({ "requirements.permissionsHeld.userId": user._id }, { $set: { "requirements.$.permissionsHeld.identityId": identity.id },
      $unset: { "requirements.$.permissionsHeld.userId": 1 } }, { multi: true }) > 0);                                // 236
    // The `$` operatorer modifies the first element in the array that matches the query. Since                       //
    // there may be many matches, we need to repeat until no documents are modified.                                  //
  });                                                                                                                 //
                                                                                                                      //
  ApiTokens.remove({ userInfo: { $exists: true } });                                                                  // 243
  // We've renamed `Grain.UserInfo.userId` to `Grain.userInfo.identityId`. The only place                             //
  // that this field could show up in the database was in this deprecated, no-longer-functional                       //
  // form of API token.                                                                                               //
}                                                                                                                     //
                                                                                                                      //
function appUpdateSettings() {                                                                                        // 249
  Settings.insert({ _id: "appMarketUrl", value: "https://apps.sandstorm.io" });                                       // 250
  Settings.insert({ _id: "appIndexUrl", value: "https://app-index.sandstorm.io" });                                   // 251
  Settings.insert({ _id: "appUpdatesEnabled", value: true });                                                         // 252
}                                                                                                                     //
                                                                                                                      //
function moveDevAndEmailLoginDataIntoIdentities() {                                                                   // 255
  Meteor.users.find().forEach(function (user) {                                                                       // 256
    if (user.identities.length != 1) {                                                                                // 257
      throw new Error("User does not have exactly one identity: ", user);                                             // 258
    }                                                                                                                 //
                                                                                                                      //
    var identity = user.identities[0];                                                                                // 261
    if (Match.test(identity.service, Object)) {                                                                       // 262
      return;                                                                                                         // 262
    } // Already migrated.                                                                                            //
                                                                                                                      //
    var newIdentity = _.pick(identity, "id", "main", "noLogin", "verifiedEmail", "unverifiedEmail");                  // 264
    newIdentity.profile = _.pick(identity, "name", "handle", "picture", "pronoun");                                   // 265
                                                                                                                      //
    var serviceObject = {};                                                                                           // 267
    var fieldsToUnset = {};                                                                                           // 268
                                                                                                                      //
    if (identity.service === "dev") {                                                                                 // 270
      serviceObject.name = user.devName;                                                                              // 271
      fieldsToUnset.devName = 1;                                                                                      // 272
    } else if (identity.service === "emailToken") {                                                                   //
      serviceObject.tokens = user.services.emailToken.tokens;                                                         // 274
      serviceObject.email = user.services.emailToken.email;                                                           // 275
      fieldsToUnset["services.emailToken"] = 1;                                                                       // 276
    }                                                                                                                 //
                                                                                                                      //
    newIdentity.service = {};                                                                                         // 279
    newIdentity.service[identity.service] = serviceObject;                                                            // 280
                                                                                                                      //
    var modifier = { $set: { identities: [newIdentity] } };                                                           // 282
    if (Object.keys(fieldsToUnset).length > 0) {                                                                      // 283
      modifier.$unset = fieldsToUnset;                                                                                // 284
    }                                                                                                                 //
                                                                                                                      //
    Meteor.users.update({ _id: user._id }, modifier);                                                                 // 287
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function repairEmailIdentityIds() {                                                                                   // 291
  Meteor.users.find({ "identities.service.emailToken": { $exists: 1 } }).forEach(function (user) {                    // 292
    if (user.identities.length != 1) {                                                                                // 293
      throw new Error("User does not have exactly one identity: ", user);                                             // 294
    }                                                                                                                 //
                                                                                                                      //
    var identity = user.identities[0];                                                                                // 297
    var newIdentity = _.pick(identity, "main", "noLogin", "verifiedEmail", "unverifiedMail", "profile");              // 298
    newIdentity.service = { email: identity.service.emailToken };                                                     // 300
    newIdentity.id = Crypto.createHash("sha256").update("email:" + identity.service.emailToken.email).digest("hex");  // 301
                                                                                                                      //
    Grains.update({ identityId: identity.id }, { $set: { identityId: newIdentity.id } }, { multi: true });            // 304
    Sessions.update({ identityId: identity.id }, { $set: { identityId: newIdentity.id } }, { multi: true });          // 305
    ApiTokens.update({ identityId: identity.id }, { $set: { identityId: newIdentity.id } }, { multi: true });         // 306
    ApiTokens.update({ "owner.user.identityId": identity.id }, { $set: { "owner.user.identityId": newIdentity.id } }, { multi: true });
    ApiTokens.update({ "owner.grain.introducerIdentity": identity.id }, { $set: { "owner.grain.introducerIdentity": newIdentity.id } }, { multi: true });
                                                                                                                      //
    while (ApiTokens.update({ "requirements.permissionsHeld.identityId": identity.id }, { $set: { "requirements.$.permissionsHeld.identityId": newIdentity.id } }, { multi: true }) > 0);
                                                                                                                      //
    Meteor.users.update({ _id: user._id }, { $set: { identities: [newIdentity] } });                                  // 320
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function splitAccountUsersAndIdentityUsers() {                                                                        // 324
  Meteor.users.find({ identities: { $exists: true } }).forEach(function (user) {                                      // 325
    if (user.identities.length != 1) {                                                                                // 326
      throw new Error("User does not have exactly one identity: ", user);                                             // 327
    }                                                                                                                 //
                                                                                                                      //
    var identity = user.identities[0];                                                                                // 330
    var identityUser = _.pick(user, "createdAt", "lastActive", "expires");                                            // 331
    identityUser._id = identity.id;                                                                                   // 332
    identityUser.profile = identity.profile;                                                                          // 333
    _.extend(identityUser, _.pick(identity, "unverifiedEmail"));                                                      // 334
    identityUser.profile.service = Object.keys(identity.service)[0];                                                  // 335
                                                                                                                      //
    // Updating this user needs to be a two step process because the `services` field typically                       //
    // contains subfields that are constrained to be unique by Mongo indices.                                         //
    identityUser.stagedServices = _.omit(user.services, "resume");                                                    // 339
    if (identity.service.dev) {                                                                                       // 340
      identityUser.stagedServices.dev = identity.service.dev;                                                         // 341
    } else if (identity.service.email) {                                                                              //
      identityUser.stagedServices.email = identity.service.email;                                                     // 343
    }                                                                                                                 //
                                                                                                                      //
    var accountUser = _.pick(user, "_id", "createdAt", "lastActive", "expires", "isAdmin", "signupKey", "signupNote", "signupEmail", "plan", "storageUsage", "isAppDemoUser", "appDemoId", "payments", "dailySentMailCount", "hasCompletedSignup");
    accountUser.loginIdentities = [_.pick(identity, "id")];                                                           // 350
    accountUser.nonloginIdentities = [];                                                                              // 351
    if (user.services && user.services.resume) {                                                                      // 352
      accountUser.services = { resume: user.services.resume };                                                        // 353
    }                                                                                                                 //
                                                                                                                      //
    accountUser.stashedOldUser = user;                                                                                // 356
                                                                                                                      //
    ApiTokens.update({ identityId: identityUser._id }, { $set: { accountId: user._id } }, { multi: true });           // 358
                                                                                                                      //
    Meteor.users.upsert({ _id: identityUser._id }, identityUser);                                                     // 361
    Meteor.users.update({ _id: user._id }, accountUser);                                                              // 362
  });                                                                                                                 //
                                                                                                                      //
  Meteor.users.find({ stagedServices: { $exists: true } }).forEach(function (identity) {                              // 365
    Meteor.users.update({ _id: identity._id }, { $unset: { stagedServices: 1 },                                       // 366
      $set: { services: identity.stagedServices } });                                                                 // 367
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function populateContactsFromApiTokens() {                                                                            // 371
  ApiTokens.find({ "owner.user.identityId": { $exists: 1 },                                                           // 372
    accountId: { $exists: 1 } }).forEach(function (token) {                                                           // 373
    var identityId = token.owner.user.identityId;                                                                     // 374
    var identity = SandstormDb.prototype.getIdentity(identityId);                                                     // 375
    if (identity) {                                                                                                   // 376
      var profile = identity.profile;                                                                                 // 377
      Contacts.upsert({ ownerId: token.accountId, identityId: identityId }, {                                         // 378
        ownerId: token.accountId,                                                                                     // 379
        petname: profile && profile.name,                                                                             // 380
        created: new Date(),                                                                                          // 381
        identityId: identityId,                                                                                       // 382
        profile: profile                                                                                              // 383
      });                                                                                                             //
    }                                                                                                                 //
  });                                                                                                                 //
}                                                                                                                     //
                                                                                                                      //
function cleanUpApiTokens() {                                                                                         // 389
  // The `splitUserIdsIntoAccountIdsAndIdentityIds()` migration only added `identityId` in cases                      //
  // where the user still existed in the database.                                                                    //
  ApiTokens.remove({ userId: { $exists: true }, identityId: { $exists: false } });                                    // 392
  ApiTokens.remove({ "owner.user.userId": { $exists: true },                                                          // 393
    "owner.user.identityId": { $exists: false } });                                                                   // 394
                                                                                                                      //
  // For a while we were accidentally setting `appIcon` instead of `icon`.                                            //
  ApiTokens.find({ "owner.user.denormalizedGrainMetadata.appIcon": { $exists: true } }).forEach(function (apiToken) {
    var icon = apiToken.owner.user.denormalizedGrainMetadata.appIcon;                                                 // 399
    ApiTokens.update({ _id: apiToken._id }, { $set: { "owner.user.denormalizedGrainMetadata.icon": icon },            // 400
      $unset: { "owner.user.denormalizedGrainMetadata.appIcon": true } });                                            // 402
  });                                                                                                                 //
                                                                                                                      //
  // For a while the `identityId` field of child UiView tokens was not getting set.                                   //
  function repairChain(parentToken) {                                                                                 // 406
    ApiTokens.find({ parentToken: parentToken._id, grainId: { $exists: true },                                        // 407
      identityId: { $exists: false } }).forEach(function (childToken) {                                               // 408
      ApiTokens.update({ _id: childToken._id }, { $set: { identityId: parentToken.identityId } });                    // 409
      repairChain(childToken);                                                                                        // 410
    });                                                                                                               //
  }                                                                                                                   //
                                                                                                                      //
  ApiTokens.find({ grainId: { $exists: true }, identityId: { $exists: true },                                         // 414
    parentToken: { $exists: false } }).forEach(repairChain);                                                          // 415
}                                                                                                                     //
                                                                                                                      //
function initServerTitleAndReturnAddress() {                                                                          // 418
  var hostname = Url.parse(process.env.ROOT_URL).hostname;                                                            // 419
  Settings.insert({ _id: "serverTitle", value: hostname });                                                           // 420
  Settings.insert({ _id: "returnAddress", value: "no-reply@" + hostname });                                           // 421
}                                                                                                                     //
                                                                                                                      //
function sendReferralNotifications() {                                                                                // 424
  if (Meteor.settings["public"].quotaEnabled) {                                                                       // 425
    Meteor.users.find({                                                                                               // 426
      loginIdentities: { $exists: true },                                                                             // 427
      expires: { $exists: false }                                                                                     // 428
    }, { fields: { _id: 1 } }).forEach(function (user) {                                                              //
      sendReferralProgramNotification(user._id);                                                                      // 430
    });                                                                                                               //
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
function assignBonuses() {                                                                                            // 435
  if (Meteor.settings["public"].quotaEnabled && SandstormDb.bonusesMigrationHook) {                                   // 436
    SandstormDb.bonusesMigrationHook();                                                                               // 437
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
// This must come after all the functions named within are defined.                                                   //
// Only append to this list!  Do not modify or remove list entries;                                                   //
// doing so is likely change the meaning and semantics of user databases.                                             //
var MIGRATIONS = [updateLoginStyleToRedirect, enableLegacyOAuthProvidersIfNotInSettings, denormalizeInviteInfo, mergeRoleAssignmentsIntoApiTokens, fixOasisStorageUsageStats, fetchProfilePictures, assignPlans, removeKeyrings, useLocalizedTextInUserActions, verifyAllPgpSignatures, splitUserIdsIntoAccountIdsAndIdentityIds, appUpdateSettings, moveDevAndEmailLoginDataIntoIdentities, repairEmailIdentityIds, splitAccountUsersAndIdentityUsers, populateContactsFromApiTokens, cleanUpApiTokens, initServerTitleAndReturnAddress, sendReferralNotifications, assignBonuses];
                                                                                                                      //
function migrateToLatest() {                                                                                          // 467
  if (Meteor.settings.replicaNumber) {                                                                                // 468
    (function () {                                                                                                    //
      // This is a replica. Wait for the first replica to perform migrations.                                         //
                                                                                                                      //
      console.log("Waiting for migrations on replica zero...");                                                       // 471
                                                                                                                      //
      var done = new Future();                                                                                        // 473
      var change = function (doc) {                                                                                   // 474
        console.log("Migrations applied elsewhere: " + doc.value + "/" + MIGRATIONS.length);                          // 475
        if (doc.value >= MIGRATIONS.length) done["return"]();                                                         // 476
      };                                                                                                              //
                                                                                                                      //
      var observer = Migrations.find({ _id: "migrations_applied" }).observe({                                         // 479
        added: change,                                                                                                // 480
        changed: change                                                                                               // 481
      });                                                                                                             //
                                                                                                                      //
      done.wait();                                                                                                    // 484
      observer.stop();                                                                                                // 485
      console.log("Migrations have completed on replica zero.");                                                      // 486
    })();                                                                                                             //
  } else {                                                                                                            //
    var applied = Migrations.findOne({ _id: "migrations_applied" });                                                  // 489
    var start = undefined;                                                                                            // 490
    if (!applied) {                                                                                                   // 491
      // Migrations table is not yet seeded with a value.  This means it has                                          //
      // applied 0 migrations.  Persist this.                                                                         //
      Migrations.insert({ _id: "migrations_applied", value: 0 });                                                     // 494
      start = 0;                                                                                                      // 495
    } else {                                                                                                          //
      start = applied.value;                                                                                          // 497
    }                                                                                                                 //
                                                                                                                      //
    console.log("Migrations already applied: " + start + "/" + MIGRATIONS.length);                                    // 500
                                                                                                                      //
    for (var i = start; i < MIGRATIONS.length; i++) {                                                                 // 502
      // Apply migration i, then record that migration i was successfully run.                                        //
      console.log("Applying migration " + (i + 1));                                                                   // 504
      MIGRATIONS[i]();                                                                                                // 505
      Migrations.update({ _id: "migrations_applied" }, { $set: { value: i + 1 } });                                   // 506
      console.log("Applied migration " + (i + 1));                                                                    // 507
    }                                                                                                                 //
  }                                                                                                                   //
}                                                                                                                     //
                                                                                                                      //
SandstormDb.prototype.migrateToLatest = migrateToLatest;                                                              // 512
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
